import pickle as pk
import numpy as np

# %% 
# MARK: Simulation parameters

def simulation_parameters(NUMBER_ATOMIC_AGENTS):
    # Simulation horizon
    # SIMULATION_HORIZON = 50 +1
    SIMULATION_HORIZON = 60 +1
    PREDICTION_HORIZON = 5

    # Bounds variables
    MAX_u = 0.5
    MIN_u = -0.5
    MAX_x = 0.9
    MIN_x = -0.9

    # Weighting matrices
    Q = 100*np.identity(NUMBER_ATOMIC_AGENTS)
    R = 1*np.identity(NUMBER_ATOMIC_AGENTS)

    epsilon = 1e-12

    PARAMETERS_DICT = {'SIMULATION_HORIZON': SIMULATION_HORIZON, 
                       'PREDICTION_HORIZON': PREDICTION_HORIZON, 
                       'NUMBER_ATOMIC_AGENTS': NUMBER_ATOMIC_AGENTS, 
                       'MAX_u': MAX_u, 
                       'MIN_u': MIN_u, 
                       'MAX_x': MAX_x, 
                       'MIN_x': MIN_x, 
                       'Q': Q, 
                       'R': R,
                       'epsilon': epsilon}

    return PARAMETERS_DICT

# %% 
# MARK: State update equation
def state_update_equation_linear(A, B, x_k, u_k):
    return A @ x_k + B @ u_k

# %% 
# MARK: List creation

def create_empty_list(N):
    return [[] for _ in range(N)]

def create_empty_list_2D(x, y):
    empty_list = [[] for _ in range(x)]
    for i in range(0,x):
        empty_list[i] = [[] for _ in range(y)]
    return empty_list

def load_linear_system(System_File):

    with open("Input_Data/" + System_File, 'rb') as f:
        System_List = pk.load(f)

    A = System_List[0].astype(np.float64)
    B = System_List[1].astype(np.float64)
    Weighted_Adjacency_Matrix = System_List[2]
    NUMBER_ATOMIC_AGENTS = System_List[3]

    return A, B, Weighted_Adjacency_Matrix, NUMBER_ATOMIC_AGENTS


# %% Get points from a circle
def generate_circle_points(center, radius, num_points):
    """
    Generate points along the circumference of a circle.

    :param center: Tuple (h, k) as the center of the circle.
    :param radius: Radius of the circle.
    :param num_points: Number of points to generate.
    :return: Two numpy arrays containing the x and y coordinates of the points.
    """
    theta = np.linspace(0, 2 * np.pi, num_points)  # Angle from 0 to 2*pi
    x = center[0] + radius * np.cos(theta)  # x = h + r*cos(theta)
    y = center[1] + radius * np.sin(theta)  # y = k + r*sin(theta)
    
    return x, y

# %% Old support functions to be updated

#  Function to create empty lists        
        
def init_list_of_objects(size):
    list_of_objects = list()
    for i in range(0,size):
        list_of_objects.append( list() ) #different object reference each time
    return list_of_objects      


#  Function to get indexes of elements from lists

def index_2d(my_list, v):
    for i in range(len(my_list)):
        for j in range(len(my_list[i])):
            if v == my_list[i][j]:
                return (i, j)
    
            
#  Function to check if an element belongs to a 2D list

def check_2d(my_list, v):
    condition = False
    for i in range(len(my_list)):
        for j in range(len(my_list[i])):
            if v == my_list[i][j]:
                condition = True
    return condition
 
    
 #  Function to check if an element belongs to a list

def check_list(my_list, v):
    condition = False
    for i in range(len(my_list)):
        if v == my_list[i]:
            condition = True
    return condition

#  Function to get indexes of elements from list one dimensional

def index_list(my_list, v):
    for i in range(len(my_list)):
        if v == my_list[i]:
                return i