# %% MARK: Preliminary Operations

from IPython import get_ipython
get_ipython().run_line_magic('reset', '-sf')

# For linear algebra
import numpy as np
import scipy as sp
from scipy.spatial import ConvexHull

import matplotlib.colors as mcolors

# For plots
import matplotlib.pyplot as plt

# For graphs plots
import networkx as nx

# For random numbers
import random as rn

# For results reproducibility
rn.seed(0)

# Math module
import math as mt

# Data module
import pandas as pd
import pickle

# For color Bar of Graphs
from matplotlib.colors import LinearSegmentedColormap, to_rgba 
from matplotlib.cm import ScalarMappable
from matplotlib import rc
# Enable LaTeX rendering
rc('text', usetex=True)

# Set the font to Times New Roman or the closest available font in LaTeX
rc('font', family='serif')
rc('font', serif='Times New Roman')  # This line might need adjustment based on LaTeX distribution

# For matrix operations
import numpy as np
# For file storage
import pickle as pk
# For time tracking
import time as tm
# For data location

OUTPUT_FOLDER = "A01_Output_Selection_Atomic_Agents/"
import os
if not os.path.exists(OUTPUT_FOLDER):
    os.mkdir(OUTPUT_FOLDER)



#  Random Network Generation
from A00_Random_Network_Generation import *

# Lists Manipulation
from A00_Support_Funcitons import *
     
# %% MARK: Options

# number_of_states = 12
# number_of_inputs = 4
# number_of_elemnts_per_column = 1
# FONT_SIZE_ATOMIC = 12
# radius_circle_atomic = 0.35
# radius_atomic = 1 
# FONT_SIZE_NETWORK = 12
# radius_circle_network = 0.4

# number_of_states = 50
# number_of_inputs = 10 
# number_of_elemnts_per_column = 1
# FONT_SIZE_NETWORK = 6
# radius_circle_network = 0.35
# FONT_SIZE_ATOMIC = 6
# radius_circle_atomic = 0.35
# radius_atomic = 3

number_of_states = 100
number_of_inputs = 20 
number_of_elemnts_per_column = 2
FONT_SIZE_NETWORK = 3
radius_circle_network = 0.35
FONT_SIZE_ATOMIC = 3
radius_circle_atomic = 0.35
radius_atomic = 6

# number_of_states = 10
# number_of_inputs = 3 
# number_of_elemnts_per_column = 3
# FONT_SIZE_NETWORK = 6
# radius_circle_network = 0.35
# FONT_SIZE_ATOMIC = 6
# radius_circle_atomic = 0.35
# radius_atomic = 3
 
FILE_NAME = "Network_States_" + str(number_of_states) + "_Inputs_" + str(number_of_inputs)

SAVE_DATA = True

# %% MARK: State Dynamics Generation

# Define the number of states and inputs


 
A, B = random_network_generator(number_of_states, number_of_inputs, number_of_elemnts_per_column)
             
# %% MARK: Initialization of empty lists and graphs

# Arrays of input and state nodes
Input_Nodes = np.arange(1,number_of_inputs+1)
State_Nodes = np.arange(number_of_inputs+1,number_of_inputs+number_of_states+1)

# List of the agents
Atomic_Agents = init_list_of_objects(number_of_inputs)
# List of the trees associated with the agents during their creation
Atomic_Agents_Trees = init_list_of_objects(number_of_inputs)
# List of the sub-graphs associated with the agents
Atomic_Agents_Graphs = init_list_of_objects(number_of_inputs)
# Create empty list of atomic agents matrices
Atomic_Agents_Matrices = init_list_of_objects(number_of_inputs)

# Initialization of Atomic Agents Graphs
for i in range(0,len(Atomic_Agents_Graphs)):
    Atomic_Agents_Trees[i] = nx.Graph()
    Atomic_Agents_Graphs[i] = nx.Graph()


# List of the nodes that still have to be assigned
Unassigned_Nodes = []
for i in range(1,number_of_inputs+number_of_states+1):
    Unassigned_Nodes.append(i)


# Graph of the entire network
G_complete = nx.Graph()

# Creation of the network graph
for i in range(0,number_of_inputs):
    G_complete.add_node(Input_Nodes[i])
    
for i in range(0,number_of_states):
    G_complete.add_node(State_Nodes[i])
    
for j in range(0,number_of_inputs):
    for i in range(0,number_of_states):
        if B[i,j] != 0:
            G_complete.add_edge(Input_Nodes[j], State_Nodes[i], weight=B[i,j])   

for i in range(0,number_of_states):
    for j in range(0,number_of_states):
        if A[i,j] != 0:
            G_complete.add_edge(State_Nodes[j], State_Nodes[i], weight=A[i,j]) 


# %% MARK: Identifications of the Roots of Atomic Control Agents

for i in range(0,number_of_inputs):
    
    for j in range(0,number_of_states):
        
        if B[j,i] != 0:
            
            if check_2d(Atomic_Agents, Input_Nodes[i]) == False:    
                
                # New Atomic Agent one input node and one state node           
                if check_2d(Atomic_Agents, State_Nodes[j]) == False:
                    
                    Atomic_Agents[i].append(Input_Nodes[i])
                    Atomic_Agents[i].append(State_Nodes[j]) 
                                  
                    Unassigned_Nodes.remove(Input_Nodes[i])
                    Unassigned_Nodes.remove(State_Nodes[j])
                
                # Merge Atomic Agents through a couplig with a state ndoe  
                elif check_2d(Atomic_Agents, State_Nodes[j]) == True:
                    
                    idx_agent = index_2d(Atomic_Agents, State_Nodes[j])
                    
                    Atomic_Agents[idx_agent[0]].append(Input_Nodes[i])
                       
                    Unassigned_Nodes.remove(Input_Nodes[i])
            
            elif check_2d(Atomic_Agents, Input_Nodes[i]) == True:     
                
                # Add State Nodes To existing Agent
                if check_2d(Atomic_Agents, State_Nodes[j]) == False:
                    
                    idx_agent = index_2d(Atomic_Agents, Input_Nodes[i])
                    
                    Atomic_Agents[idx_agent[0]].append(State_Nodes[j])
                    
                    Unassigned_Nodes.remove(State_Nodes[j])


# %% MARK: Sort Agents and remove empty agents

# Sort
Atomic_Agents.sort()
for k in range(0, len(Atomic_Agents)):
    Atomic_Agents[k].sort()

# Remove Empty Agents
while [] in Atomic_Agents:
    Atomic_Agents.remove([])


# %% MARK: Update Atomic Agents Trees and Graphs

for k in range(0, len(Atomic_Agents)):
    
    for i in Atomic_Agents[k]:
        
        Atomic_Agents_Trees[k].add_node(i)
        Atomic_Agents_Graphs[k].add_node(i)
        
        if i in Input_Nodes: 
            
            for j in Atomic_Agents[k]:
                
                if j in State_Nodes:                
                    
                    if B[j-number_of_inputs-1,i-1] != 0:
                        
                        Atomic_Agents_Trees[k].add_edge(i, j, weight=B[j-number_of_inputs-1,i-1])
                        Atomic_Agents_Graphs[k].add_edge(i, j, weight=B[j-number_of_inputs-1,i-1])   
            

# %% MARK: Assignments of the leafs of atomic control agents through forward exploration

reiterate = True

while reiterate == True:
    
    reiterate = False
    
    for i in range(0, number_of_states):
        
        if State_Nodes[i] in Unassigned_Nodes:
            
            A_max = 0
            j_max = 0
            
            for k in range(0,len(Atomic_Agents)):
                for j in range(0, number_of_states):    
                    if State_Nodes[j] in Atomic_Agents[k]:
                        if A[i,j] > A_max:
                            A_max = A[i,j]
                            j_max = j
            
            if A_max != 0:
                    
                idx_agent = index_2d(Atomic_Agents, State_Nodes[j_max])
                k = idx_agent[0]
                
                Atomic_Agents[k].append(State_Nodes[i])
                
                Unassigned_Nodes.remove(State_Nodes[i])
                
                reiterate = True

            
# %% MARK: Assignments of the unassignad nodes 

i = 0
reiterate = False

while len(Unassigned_Nodes) > 0:
    
    reiterate = False
    
    if State_Nodes[i] in Unassigned_Nodes:
        
        j = np.argmax(A[:,i])
        
        if check_2d(Atomic_Agents, State_Nodes[j]) == True and np.max(A[:,i]) != 0:
            
            idx_agent = index_2d(Atomic_Agents, State_Nodes[j])
            k = idx_agent[0]
            
            Atomic_Agents[k].append(State_Nodes[i])    
                        
            Atomic_Agents_Graphs[k].add_node(State_Nodes[i]) 
            Atomic_Agents_Graphs[k].add_edge(State_Nodes[i], State_Nodes[j], weight=A[j,i])            
            
            Atomic_Agents_Trees[k].add_node(State_Nodes[i]) 
            Atomic_Agents_Trees[k].add_edge(State_Nodes[i], State_Nodes[j], weight=A[j,i])

            Unassigned_Nodes.remove(State_Nodes[i])
            
            reiterate = True     
            
    i += 1
    if i == number_of_states:
        break
    if reiterate == True:
        i = 0


# %% MARK: Assignment of the edges to the Atomic Agents Graphs

for k in range(0, len(Atomic_Agents)):
    
    for i in range(0, number_of_states):
        
        if State_Nodes[i] in Atomic_Agents[k]:
        
            for j in range(0, number_of_inputs):
                
                if Input_Nodes[j] in Atomic_Agents[k]:
                    
                    if B[i,j] != 0:
                    
                        Atomic_Agents_Graphs[k].add_edge(Input_Nodes[j], State_Nodes[i], weight=B[i,j])
                        
        
            for j in range(0, number_of_states):
                
                if State_Nodes[j] in Atomic_Agents[k]:
                    
                    if A[i,j] != 0:
                        
                        Atomic_Agents_Graphs[k].add_edge(State_Nodes[j], State_Nodes[i], weight=A[i,j])

    
# %% MARK: Build Atomic Control Agents state and input matrices

for k in range(0, len(Atomic_Agents)):
    
    # Detect the number of states and inputs
    nxi = 0
    nui = 0
        
    for i in range(0, len(Atomic_Agents[k])):
            
        if Atomic_Agents[k][i] <= number_of_inputs:              
            nui = nui + 1 
                
        elif Atomic_Agents[k][i] > number_of_inputs:          
            nxi = nxi + 1 
                
    Ai = np.full((nxi,nxi), 0, dtype='float64')
    Aij = np.full((nxi,number_of_states), 0, dtype='float64')
    Bi = np.full((nxi,nui), 0, dtype='float64')
        
    raw_xi = 0  
    col_ui = 0

    for i in Atomic_Agents[k]:
 
        if i > number_of_inputs:
            
            coli = 0
            
            for j in range(0, number_of_states):     
                    
                if State_Nodes[j] in Atomic_Agents[k]:

                    Ai[raw_xi,coli] = A[i-number_of_inputs-1,j]

                    coli = coli+1
                            
                if State_Nodes[j] not in Atomic_Agents[k]:
                    
                    if A[i-number_of_inputs-1,j] != 0:
                        
                        Aij[raw_xi,j] = A[i-number_of_inputs-1,j] 
                            
            raw_xi = raw_xi + 1
                
        elif i <= number_of_inputs:    
            
            raw_ui = 0   

            for j in range(0, number_of_states):
                    
                if State_Nodes[j] in Atomic_Agents[k]:

                    Bi[raw_ui,col_ui] = B[j,i-1]                       
                        
                    raw_ui = raw_ui + 1
                    
            col_ui = col_ui + 1    
                
    Atomic_Agents_Matrices[k] = [[Ai],[Bi],[Aij]]
        

# Remove Empty Agents' Matrices
while [] in Atomic_Agents_Matrices:
    Atomic_Agents_Matrices.remove([])

Number_Atomic_Agents = len(Atomic_Agents)

# %% MARK: Export Data

Export_Data = {'A': A,
               'B': B,
               'Input_Nodes': Input_Nodes,
               'State_Nodes': State_Nodes,
               'Atomic_Agents': Atomic_Agents,
               'Number_Atomic_Agents': Number_Atomic_Agents,
               'Atomic_Agents_Matrices': Atomic_Agents_Matrices}

if SAVE_DATA == True:
    with open(OUTPUT_FOLDER + FILE_NAME + '_data.pkl', 'wb') as file:
        pk.dump(Export_Data, file)

# %% MARK: Construction of the graphical graph

positions_G_Complete = init_list_of_objects(number_of_states + number_of_inputs)
degree_list = init_list_of_objects(number_of_states + number_of_inputs)

degree_list_raw = list(G_complete.degree())

# Determination of themaximum  radius of the network
for k in range(0, len(degree_list)):
    degree_list[k] = degree_list_raw[k][1]

max_radius = np.amax(degree_list)


# Basic radius for the node of the network
# radius_circle = 0.15


# Positon index 
idx = 0

# Assignment of the positions of the nodes for input nodes
for i in range(0, number_of_inputs):
        
    positions_G_Complete[idx] = [Input_Nodes[i], max_radius * mt.cos((2*mt.pi*i)/(number_of_inputs)),
                                                 max_radius * mt.sin((2*mt.pi*i)/(number_of_inputs))]
        
    idx += 1


states_pos_indx = np.full(np.amax(degree_list) + 1, 0, dtype='float64')

# Assignment of the positions of the nodes for state nodes
for i in range(0, number_of_states):
    
    node_degree = degree_list[i + number_of_inputs]
    
    
    number_of_sections = degree_list.count(node_degree)
    
    positions_G_Complete[idx] = [State_Nodes[i], (max_radius + 1 - node_degree) * mt.cos((2*mt.pi*states_pos_indx[node_degree])/(number_of_sections)),
                                                 (max_radius + 1 - node_degree) * mt.sin((2*mt.pi*states_pos_indx[node_degree])/(number_of_sections))]
        
    states_pos_indx[node_degree] += 1    
    idx += 1


# %% MARK: Plot Network Graph Degree-based

plot_network_map = True
save_network_map = True

FONT_SIZE = 12

if plot_network_map == True:

    # plt.figure(figsize=(50, 50), dpi=100)
    # plt.figure(dpi=100)
    Figure_Network_Map = plt.figure(figsize=(8, 8), dpi=100)
    
    # Plot Arrows
    sin_val_A = np.linalg.svd(A)
    max_sin_val_A = np.amax(sin_val_A[1])
    
    sin_val_B = np.linalg.svd(B)
    max_sin_val_B = np.amax(sin_val_B[1])
    


    for i in range(0, number_of_states):
        for j in range(0, number_of_inputs):
            if B[i,j] != 0:
                
                x  = positions_G_Complete[j][1]
                y  = positions_G_Complete[j][2]
                dx = positions_G_Complete[i + number_of_inputs][1] - positions_G_Complete[j][1]
                dy = positions_G_Complete[i + number_of_inputs][2] - positions_G_Complete[j][2]
                
                
                angle = mt.atan2(dy,dx)
                x_shift = radius_circle_network*mt.cos(angle)
                y_shift = radius_circle_network*mt.sin(angle)
                
                # plt.arrow(x, y, dx-x_shift, dy-y_shift, width=0.04, length_includes_head=True, color="cyan", alpha=abs(B[i,j])/max_sin_val_B, linestyle=' ')
                plt.arrow(x, y, dx-x_shift, dy-y_shift, width=0.04, length_includes_head=True, color="black", alpha=abs(B[i,j])/max_sin_val_B, linestyle=' ')
    
    for i in range(0, number_of_states):
        for j in range(0, number_of_states):
            if A[i,j] != 0 and i != j:
                
                x  = positions_G_Complete[j + number_of_inputs][1]
                y  = positions_G_Complete[j + number_of_inputs][2]
                dx = positions_G_Complete[i + number_of_inputs][1] - positions_G_Complete[j + number_of_inputs][1]
                dy = positions_G_Complete[i + number_of_inputs][2] - positions_G_Complete[j + number_of_inputs][2]
                
                
                angle = mt.atan2(dy,dx)
                x_shift = radius_circle_network*mt.cos(angle)
                y_shift = radius_circle_network*mt.sin(angle)
                
                # plt.arrow(x, y, dx-x_shift, dy-y_shift, width=0.04, length_includes_head=True, color="cyan", alpha=abs(A[i,j])/max_sin_val_A, linestyle=' ')
                plt.arrow(x, y, dx-x_shift, dy-y_shift, width=0.04, length_includes_head=True, color="black", alpha=abs(A[i,j])/max_sin_val_A, linestyle=' ')
    
    
    
    ax = plt.gca()

    # Plot circles
    for radius in range(1, max_radius):
        circle = plt.Circle((0,0), radius, facecolor='None', edgecolor='#40e0d0',alpha=0.05)
        ax.add_patch(circle)

    # Plot Nodes
    for k in range(0, len(positions_G_Complete)):
        
        if k < number_of_inputs:
            
            circle = plt.Circle((positions_G_Complete[k][1], positions_G_Complete[k][2]), radius_circle_network, color='black', alpha=1)
            ax.add_patch(circle)
            circle = plt.Circle((positions_G_Complete[k][1], positions_G_Complete[k][2]), radius_circle_network*0.9, color='white',  alpha=1)
            ax.add_patch(circle)
            circle = plt.Circle((positions_G_Complete[k][1], positions_G_Complete[k][2]), radius_circle_network*0.9, color='r', alpha=0.9)
            ax.add_patch(circle)
            # To add labels
            # plt.text(positions_G_Complete[k][1], positions_G_Complete[k][2],  f'u$_{{{positions_G_Complete[k][0]}}}$', color="black", fontsize=FONT_SIZE_NETWORK, horizontalalignment="center", verticalalignment="center")
            
            # plt.text(positions_G_Complete[k][1], positions_G_Complete[k][2], str(positions_G_Complete[k][0]), color="black", fontsize=FONT_SIZE, horizontalalignment="center", verticalalignment="center")
    
        else:
            circle = plt.Circle((positions_G_Complete[k][1], positions_G_Complete[k][2]), radius_circle_network, color='black', alpha=1)
            ax.add_patch(circle)
            circle = plt.Circle((positions_G_Complete[k][1], positions_G_Complete[k][2]), radius_circle_network*0.9, color='white',  alpha=1)
            ax.add_patch(circle)
            circle = plt.Circle((positions_G_Complete[k][1], positions_G_Complete[k][2]), radius_circle_network*0.9, color='cyan', alpha=0.9)
            ax.add_patch(circle)
            # To add labels
            # plt.text(positions_G_Complete[k][1], positions_G_Complete[k][2],  f'x$_{{{positions_G_Complete[k][0]-number_of_inputs}}}$', color="black", fontsize=FONT_SIZE_NETWORK, horizontalalignment="center", verticalalignment="center")

            # plt.text(positions_G_Complete[k][1], positions_G_Complete[k][2], str(positions_G_Complete[k][0]), color="black", fontsize=FONT_SIZE, horizontalalignment="center", verticalalignment="center")
    

    
    # Color Bar
    plot_color_map = False
    
    if plot_color_map == True:
        my_cmap = LinearSegmentedColormap.from_list(None, [(0, 1, 1, 0.0), (0, 0, 255, 100.0)])
        cbar = plt.colorbar(ScalarMappable(cmap=my_cmap), ticks=[0, 1], pad=0.02)
        cbar.ax.set_yticklabels(["Min", "Max"], fontsize=FONT_SIZE_NETWORK)
    
    plt.tight_layout()
    
        
    plt.axis("equal")
    plt.axis('off') 
    Figure_Network_Map.tight_layout()
    fig = plt.gcf()   

    fig.patch.set_linewidth(1)
    fig.patch.set_edgecolor('black')

    plt.show() 
    if save_network_map == True:
        Figure_Network_Map.savefig(OUTPUT_FOLDER + FILE_NAME + "_network_map.pdf", dpi=300)
        
      
# %% MARK: Plot Network Graph agent-based


# List of the positions of the nodes
positions = init_list_of_objects(number_of_states + number_of_inputs)

# %%% Determination of the number of rings of the circular graph

# Number of sections of the external radius
number_of_external_elements = len(Atomic_Agents) 

pos_idx = 0

# Basic radius unit used throughout the networks

rad_idx = 0
longest_path = 0

# Determination of the number of rings of the circular graph
for k in range(0,len(Atomic_Agents)): 
    
    for j in range(0, len(Atomic_Agents[k])):

        if Atomic_Agents[k][j] in State_Nodes:

            shortest_lenght = len(nx.shortest_path(Atomic_Agents_Graphs[k], source=Atomic_Agents[k][0], target=Atomic_Agents[k][j], weight=1))
    
            if shortest_lenght > longest_path:
                
                longest_path = shortest_lenght


# %%% Assignment of the Position of nodes
                    
for k in range(0,len(Atomic_Agents)): 
    
    # input nodes
    
    if Atomic_Agents[k] != []:

        n_ui = 0
        
        for j in range(0, len(Atomic_Agents[k])):
            
            if Atomic_Agents[k][j] in Input_Nodes:
                
                n_ui += 1
        
        
        k_ui = 1       
        
        for j in range(0, len(Atomic_Agents[k])):
            
            # index internal division
            if Atomic_Agents[k][j] in Input_Nodes:
                
                positions[pos_idx] = [Atomic_Agents[k][j],  (longest_path)*radius_atomic * mt.cos(((2 * mt.pi * rad_idx)/(number_of_external_elements)) + (((2*mt.pi)/(number_of_external_elements))/(n_ui+1))*k_ui), (longest_path)*radius_atomic * mt.sin(((2 * mt.pi * rad_idx)/(number_of_external_elements)) + (((2*mt.pi)/(number_of_external_elements))/(n_ui+1))*k_ui)]
                
                k_ui += 1 
                
                pos_idx += 1
                
        
        # State Nodes
        n_xi = 0
        
        max_eccentricity_tree = 0
        
        for j in range(0, len(Atomic_Agents[k])):
            
            if Atomic_Agents[k][j] in State_Nodes:

                if max_eccentricity_tree < len(nx.shortest_path(Atomic_Agents_Graphs[k], source=Atomic_Agents[k][0], target=Atomic_Agents[k][j],weight=1)):
                    
                    max_eccentricity_tree = len(nx.shortest_path(Atomic_Agents_Graphs[k], source=Atomic_Agents[k][0], target=Atomic_Agents[k][j],weight=1))  

                
        number_states_eccentricity = np.full(max_eccentricity_tree, 0, dtype='float64')
 
        for j in range(0, len(Atomic_Agents[k])):
            
            if Atomic_Agents[k][j] in State_Nodes:
                
                shortest_lenght = len(nx.shortest_path(Atomic_Agents_Graphs[k], source=Atomic_Agents[k][0], target=Atomic_Agents[k][j], weight=1))

                for l in range(0, len(Atomic_Agents[k])):
                    
                    if Atomic_Agents[k][l] in Input_Nodes:
                        
                        if len(nx.shortest_path(Atomic_Agents_Graphs[k], Atomic_Agents[k][l], Atomic_Agents[k][j])) < shortest_lenght:
                            
                            shortest_lenght = len(nx.shortest_path(Atomic_Agents_Graphs[k], Atomic_Agents[k][l], Atomic_Agents[k][j]))    
                
                
                eccentricity_state = shortest_lenght

                number_states_eccentricity[eccentricity_state-1] = number_states_eccentricity[eccentricity_state-1] + 1
                
        
        for m in range(0, max_eccentricity_tree):
            
            int_rad_idx = 1
            
            for j in range(0, len(Atomic_Agents[k])):
                
                if Atomic_Agents[k][j] in State_Nodes:
                    
                    shortest_lenght = len(nx.shortest_path(Atomic_Agents_Graphs[k], Atomic_Agents[k][0], Atomic_Agents[k][j]))
                    
                    for l in range(0, len(Atomic_Agents[k])):
                        
                        if Atomic_Agents[k][l] in Input_Nodes:
                            
                            if len(nx.shortest_path(Atomic_Agents_Graphs[k], Atomic_Agents[k][l], Atomic_Agents[k][j])) < shortest_lenght:
                                
                                shortest_lenght = len(nx.shortest_path(Atomic_Agents_Graphs[k], Atomic_Agents[k][l], Atomic_Agents[k][j]))


                    if shortest_lenght == m + 1:
                        
                        n_ui = number_states_eccentricity[m]
                        
                        positions[pos_idx] = [Atomic_Agents[k][j], ( longest_path-m) * radius_atomic * mt.cos(((2 * mt.pi * rad_idx)/(number_of_external_elements)) + (((2*mt.pi)/(number_of_external_elements))/(n_ui+1))*int_rad_idx), ( longest_path-m) * radius_atomic * mt.sin(((2 * mt.pi * rad_idx)/(number_of_external_elements)) + (((2*mt.pi)/(number_of_external_elements))/(n_ui+1))*int_rad_idx)]
                        
                        int_rad_idx += 1
                        
                        pos_idx += 1
  
        
        rad_idx +=1

    
# %%% Plot Radial Atomic Agents Graph


# !!!! This have to be modified w.r.t. the one used for the entire network graph to ensure a good graphical result


sin_val_A = np.linalg.svd(A)
max_sin_val_A = np.amax(sin_val_A[1])

sin_val_B = np.linalg.svd(B)
max_sin_val_B = np.amax(sin_val_B[1])

plot_radial_atomic_agents_graph = True
save_radial_atomic_agents_graph = True

if plot_radial_atomic_agents_graph == True:

    # plt.figure(figsize=(50, 50), dpi=100)
    Figure_Atomic_Agents = plt.figure(figsize=(8, 8), dpi=100)
    
    ax = plt.gca()
    
    for k in range(0,len(Atomic_Agents)):
        
        if Atomic_Agents_Matrices[k] != []:
            
            for i in Atomic_Agents[k]:
                
                if i in State_Nodes:                   
                    
                    for j in State_Nodes:
                        
                        if j not in Atomic_Agents[k]:
                            
                            if A[i-number_of_inputs-1,j-number_of_inputs-1] != 0 and i != 0:
                                
                                for m in range(0,len(positions)): 
                                    
                                    if positions[m][0] == j:
                                        tail_idx = m
                                    if positions[m][0] == i:
                                        head_idx = m    
                                        
                                angle = mt.atan2(positions[head_idx][2]-positions[tail_idx][2],positions[head_idx][1]-positions[tail_idx][1])
                                x_shift = radius_circle_atomic*mt.cos(angle)
                                y_shift = radius_circle_atomic*mt.sin(angle)
                                # plt.arrow(positions[tail_idx][1], positions[tail_idx][2], positions[head_idx][1]-positions[tail_idx][1]-x_shift, positions[head_idx][2]-positions[tail_idx][2]-y_shift, width=0.05, length_includes_head=True, color="black", alpha=abs(A[i-number_of_inputs-1,j-number_of_inputs-1])/max_sin_val_A, linestyle=' ')
                                
                                plt.arrow(positions[tail_idx][1], positions[tail_idx][2], positions[head_idx][1]-positions[tail_idx][1]-x_shift, positions[head_idx][2]-positions[tail_idx][2]-y_shift, width=0.05, length_includes_head=True, color="#40e0d0", alpha=abs(A[i-number_of_inputs-1,j-number_of_inputs-1])/max_sin_val_A, linestyle=' ', zorder=2)
                                # plt.arrow(positions[tail_idx][1], positions[tail_idx][2], positions[head_idx][1]-positions[tail_idx][1]-x_shift, positions[head_idx][2]-positions[tail_idx][2]-y_shift, width=0.05, length_includes_head=True, color="cyan", alpha=abs(A[i-number_of_inputs-1,j-number_of_inputs-1])/max_sin_val_A, linestyle=' ', zorder=2)

        
    for k in range(0,len(Atomic_Agents)):
        
        if Atomic_Agents[k] != []:
            
            for i in Atomic_Agents[k]:
                
                if i in State_Nodes:                   
                    
                    for j in Atomic_Agents[k]:
                        
                        if j in State_Nodes:
                            
                            if A[i-number_of_inputs-1,j-number_of_inputs-1] != 0:
                                
                                for m in range(0,len(positions)): 
                                    
                                    if positions[m][0] == j:
                                        tail_idx = m
                                    if positions[m][0] == i:
                                        head_idx = m    
                                        
                                angle = mt.atan2(positions[head_idx][2]-positions[tail_idx][2],positions[head_idx][1]-positions[tail_idx][1])
                                x_shift = radius_circle_atomic*mt.cos(angle)
                                y_shift = radius_circle_atomic*mt.sin(angle)
                                plt.arrow(positions[tail_idx][1], positions[tail_idx][2], positions[head_idx][1]-positions[tail_idx][1]-x_shift, positions[head_idx][2]-positions[tail_idx][2]-y_shift, width=0.05, length_includes_head=True, color="black", alpha=abs(A[i-number_of_inputs-1,j-number_of_inputs-1])/max_sin_val_A, linestyle=' ', zorder=2)
     
    
    for k in range(0,len(Atomic_Agents)):
        
        if Atomic_Agents_Matrices[k] != []:
            
            for j in Atomic_Agents[k]:
                
                if j in Input_Nodes:                   
                    
                    for i in Atomic_Agents[k]:
                        
                        if i in State_Nodes:
                            
                            if B[i-number_of_inputs-1,j-1] != 0:
                                
                                for m in range(0,len(positions)): 
                                    
                                    if positions[m][0] == j:
                                        tail_idx = m
                                    if positions[m][0] == i:
                                        head_idx = m
                                        
                                angle = mt.atan2(positions[head_idx][2]-positions[tail_idx][2],positions[head_idx][1]-positions[tail_idx][1])
                                x_shift = radius_circle_atomic*mt.cos(angle)
                                y_shift = radius_circle_atomic*mt.sin(angle)
                                # plt.arrow(positions[tail_idx][1], positions[tail_idx][2], positions[head_idx][1]-positions[tail_idx][1]-x_shift, positions[head_idx][2]-positions[tail_idx][2]-y_shift, width=0.05, length_includes_head=True, color="cyan", alpha=abs(B[i-number_of_inputs-1,j-1])/max_sin_val_B, linestyle=' ')
                                plt.arrow(positions[tail_idx][1], positions[tail_idx][2], positions[head_idx][1]-positions[tail_idx][1]-x_shift, positions[head_idx][2]-positions[tail_idx][2]-y_shift, width=0.05, length_includes_head=True, color="black", alpha=abs(B[i-number_of_inputs-1,j-1])/max_sin_val_B, linestyle=' ', zorder=2)
                                
    
    for k in range(0,len(positions)): 
       
        if positions[k] != []:
            
            if positions[k][0] in Input_Nodes:
                
                circle = plt.Circle((positions[k][1], positions[k][2]), radius_circle_atomic, color='black', alpha=1, zorder=3)
                ax.add_patch(circle)
                circle = plt.Circle((positions[k][1], positions[k][2]), radius_circle_atomic*0.9, color='w', alpha=1, zorder=3)
                ax.add_patch(circle)
                circle = plt.Circle((positions[k][1], positions[k][2]), radius_circle_atomic*0.9, color='r', alpha=0.7, zorder=3)
                ax.add_patch(circle)
                # To add labels
                # plt.text(positions[k][1], positions[k][2],  f'u$_{{{positions[k][0]}}}$', color="black", fontsize=FONT_SIZE_ATOMIC, horizontalalignment="center", verticalalignment="center", zorder=4)
                
                # plt.text(positions[k][1], positions[k][2], str(positions[k][0]), color="black", fontsize=FONT_SIZE, horizontalalignment="center", verticalalignment="center")
            
            
            if positions[k][0] in State_Nodes:
                
                circle = plt.Circle((positions[k][1], positions[k][2]), radius_circle_atomic, color='black', alpha=1, zorder=3)
                ax.add_patch(circle)
                circle = plt.Circle((positions[k][1], positions[k][2]), radius_circle_atomic*0.9, color='w', alpha=1, zorder=3)
                ax.add_patch(circle)
                circle = plt.Circle((positions[k][1], positions[k][2]), radius_circle_atomic*0.9, color='cyan', alpha=0.9, zorder=3)
                ax.add_patch(circle)
                # plt.text(positions[k][1], positions[k][2], "x" + str(positions[k][0]-number_of_inputs), color="black", fontsize=FONT_SIZE, horizontalalignment="center", verticalalignment="center")
                # To add labels
                # plt.text(positions[k][1], positions[k][2],  f'x$_{{{positions[k][0]-number_of_inputs}}}$', color="black", fontsize=FONT_SIZE_ATOMIC, horizontalalignment="center", verticalalignment="center", zorder=4)
            
    # Plot convex hull
    # Add function generate_circle_points
    # Create perpetual color list
    # Import libraries to do color list and convex hull

    Table_Colors = mcolors.TABLEAU_COLORS
    NUMBER_ATOMIC_AGENTS = len(Atomic_Agents)
    Color_List_Base = list(Table_Colors.keys())
    Color_List = create_empty_list(NUMBER_ATOMIC_AGENTS)
    external_idx = 0
    for idx in range(NUMBER_ATOMIC_AGENTS):
        Color_List[idx] = Color_List_Base[external_idx]
        external_idx += 1
        if external_idx >= len(Color_List_Base):
            external_idx = 0
    number_points_circumference = 100
    global_idx2 = 0
    for k in range(0,len(Atomic_Agents)): 
        Points_Convex_Hull = np.zeros((number_points_circumference*len(Atomic_Agents[k]),2))
        global_idx = 0
        
        for i in range(len(Atomic_Agents[k])):
            agent_idx_i = i
            Circle_Points = generate_circle_points([positions[global_idx2][1],positions[global_idx2][2]], radius_circle_atomic*1.5, number_points_circumference)
            # radius_circle_atomic 
            for idx_circle in range(number_points_circumference):
                Points_Convex_Hull[global_idx,0] = Circle_Points[0][idx_circle]
                Points_Convex_Hull[global_idx,1] = Circle_Points[1][idx_circle]
                global_idx+= 1 

            # Points_Convex_Hull[i,0] = positions[global_idx2][1]
            # Points_Convex_Hull[i,1] = positions[global_idx2][2]
            global_idx2 += 1
        hull = ConvexHull(Points_Convex_Hull)

        # plt.fill(Points_Convex_Hull[hull.vertices,0], Points_Convex_Hull[hull.vertices,1], color=Color_List[k], alpha=0.3, zorder=1)

        plt.fill(Points_Convex_Hull[hull.vertices,0], Points_Convex_Hull[hull.vertices,1], color='#B5E6A2', alpha=0.3, zorder=1)

        

        
    # for idx in range(len(Atomic_Agents[k])):
        
            



    plt.axis("equal")
    plt.axis('off') 

    fig = plt.gcf()

    fig.patch.set_linewidth(1)
    fig.patch.set_edgecolor('black')

    Figure_Atomic_Agents.tight_layout()   
    plt.show() 
    if save_radial_atomic_agents_graph == True:
        Figure_Atomic_Agents.savefig(OUTPUT_FOLDER + FILE_NAME + "_atomic_control_agents.pdf", dpi=300)
        





    
















