# %% MARK: Import Libraries

import numpy as np 
import random as rnd
import pickle as pk
import networkx as nx
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib import rc
import matplotlib.colors as mcolors
import pandas as pd 

# Enable LaTeX rendering
rc('text', usetex=True)

# Set the font to Times New Roman or the closest available font in LaTeX
rc('font', family='serif')
rc('font', serif='Times New Roman')  # This line might need adjustment based on LaTeX distribution


# # Support functions
# import sys
# from pathlib import Path
# sys.path.append(str(Path(__file__).resolve().parents[1]))
# from Support_Functions.

from A00_Support_Funcitons import *

import os
if not os.path.exists("Input_Data"):
    os.mkdir("Input_Data")
INPUT_FOLDER = "Input_Data"



# %% MARK: Options

EXPORT_DATA = True
DRAW_PLOTS = True
EXPORT_PLOTS = False

# %% MARK: Random 10

File_Name = "Random_10"

rnd.seed(1)
NUMBER_OF_ATOMIC_AGENTS =  10
AGENT_DYNAMICS = 0.2

Weighted_Adjacency_Matrix = np.zeros((NUMBER_OF_ATOMIC_AGENTS,NUMBER_OF_ATOMIC_AGENTS))

Row_Index = 0
Col_Index = 0

# Ensure connectedness
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    Assignement_Performed = False
    while Assignement_Performed == False:
        j = rnd.randint(0, NUMBER_OF_ATOMIC_AGENTS-1)
        if i != j:
            Weight = rnd.randrange(100,800,1)/1000
            Weighted_Adjacency_Matrix[i,j] = Weight
            Assignement_Performed = True

# Connectedness cap
Connectedness_Cap = np.zeros((NUMBER_OF_ATOMIC_AGENTS,1))
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    Connectedness_Cap[i] = rnd.randrange(50,80,1)/100

# Assign self-edges
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    for j in range(i,NUMBER_OF_ATOMIC_AGENTS):
        if i == j:
            Weighted_Adjacency_Matrix[i,j] = AGENT_DYNAMICS

# Random topology assignment
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    for j in range(i,NUMBER_OF_ATOMIC_AGENTS):
        if i != j:
            Assign_Chance = rnd.random()
            if Assign_Chance > 0.5:
                Weight = rnd.random()
                if Weighted_Adjacency_Matrix[i,j] == 0:
                    if np.sum(Weighted_Adjacency_Matrix[i,:] + Weight) <= Connectedness_Cap[i]:
                        Weighted_Adjacency_Matrix[i,j] = Weight
                        Weighted_Adjacency_Matrix[j,i] = Weight
        

A = Weighted_Adjacency_Matrix
B = np.identity(NUMBER_OF_ATOMIC_AGENTS) 

System_List = {"A": A,
               "B": B,
               "Weighted_Adjacency_Matrix": Weighted_Adjacency_Matrix,
               "NUMBER_OF_ATOMIC_AGENTS": NUMBER_OF_ATOMIC_AGENTS}

if EXPORT_DATA == True:
    with open(INPUT_FOLDER + "/" + File_Name, 'wb') as Write_File:
        pk.dump(System_List, Write_File)

if DRAW_PLOTS == True:
    # Build system graph
    Network_Graph = nx.Graph()
    for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
        for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
            if i != j:
                if Weighted_Adjacency_Matrix[i,j] != 0:
                    Network_Graph.add_edge(i, j, weight=Weighted_Adjacency_Matrix[i,j])
    
    degree_sequence = sorted((d for n, d in Network_Graph.degree()), reverse=True)
    dmax = max(degree_sequence)
    fig = plt.figure("Random graph topology", figsize=(8, 8))
    ax0 = plt.gca()
    Gcc = Network_Graph.subgraph(sorted(nx.connected_components(Network_Graph), key=len, reverse=True)[0])
    pos = nx.spring_layout(Network_Graph)
    # pos = nx.rescale_layout(Network_Graph)
    
    plot_colors = dict(Network_Graph.degree())

    Edges_Alpha = np.zeros((NUMBER_OF_ATOMIC_AGENTS*NUMBER_OF_ATOMIC_AGENTS,1))
    external_idx = 0
    for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
        for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
            if i != j:
                if Weighted_Adjacency_Matrix[i,j] != 0:
                    Edges_Alpha[external_idx] = Weighted_Adjacency_Matrix[i,j]
                    external_idx += 1
    non_zero_mask = Edges_Alpha != 0
    Edges_Alpha_Filtered = Edges_Alpha[non_zero_mask]

    nx.draw_networkx_nodes(Network_Graph, pos, ax=ax0,  node_size=50,cmap= "Blues", node_color=list(plot_colors.values()), edgecolors='cyan', linewidths=0.4)

    nx.draw_networkx_edges(Network_Graph, pos, ax=ax0, alpha=0.3, edge_color="green", width=Edges_Alpha_Filtered*3)
    fig.tight_layout()
    plt.title("Random graph with 10 nodes", fontsize=18)
    plt.show()

    if EXPORT_PLOTS == True:
        fig.savefig(INPUT_FOLDER + "/" + File_Name + "_GraphPlot.pdf", dpi=300, bbox_inches='tight')       
        fig.savefig(INPUT_FOLDER + "/" + File_Name + "_GraphPlot.png", dpi=300, bbox_inches='tight')       


# %% MARK: Random 100

File_Name = "Random_100"

rnd.seed(1)
NUMBER_OF_ATOMIC_AGENTS =  100
AGENT_DYNAMICS = 0.2
SIMULATION_HORIZON = 1000


X_INITIAL = np.zeros((NUMBER_OF_ATOMIC_AGENTS,1))
rnd.seed(0)
for i in range(NUMBER_OF_ATOMIC_AGENTS):
    X_INITIAL[i] = rnd.randint(-9000,9000)/10000

Reference_Signal = np.zeros((NUMBER_OF_ATOMIC_AGENTS,SIMULATION_HORIZON))# Sinusoid 

rnd.seed(0)
for i in range(0,len(Reference_Signal)):
    Frequency = 2*np.pi/(rnd.randrange(20,100,1))
    Amplitude = rnd.randrange(10,100,1)/100
    Phase = 2*np.pi*rnd.randrange(100,1000,1)/1000
    for k in range(0,len(Reference_Signal[i])):
        Reference_Signal[i,k] = Amplitude*np.sin((Frequency * k) + Phase)


Weighted_Adjacency_Matrix = np.zeros((NUMBER_OF_ATOMIC_AGENTS,NUMBER_OF_ATOMIC_AGENTS))

Row_Index = 0
Col_Index = 0

# Ensure connectedness
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    Assignement_Performed = False
    while Assignement_Performed == False:
        j = rnd.randint(0, NUMBER_OF_ATOMIC_AGENTS-1)
        if i != j:
            Weight = rnd.randrange(100,800,1)/1000
            Weighted_Adjacency_Matrix[i,j] = Weight
            Assignement_Performed = True

# Connectedness cap
Connectedness_Cap = np.zeros((NUMBER_OF_ATOMIC_AGENTS,1))
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    Connectedness_Cap[i] = rnd.randrange(50,80,1)/100

# Assign self-edges
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    for j in range(i,NUMBER_OF_ATOMIC_AGENTS):
        if i == j:
            Weighted_Adjacency_Matrix[i,j] = AGENT_DYNAMICS

# Random topology assignment
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    for j in range(i,NUMBER_OF_ATOMIC_AGENTS):
        if i != j:
            Assign_Chance = rnd.random()
            if Assign_Chance > 0.5:
                Weight = rnd.random()
                if Weighted_Adjacency_Matrix[i,j] == 0:
                    if np.sum(Weighted_Adjacency_Matrix[i,:] + Weight) <= Connectedness_Cap[i]:
                        Weighted_Adjacency_Matrix[i,j] = Weight
                        Weighted_Adjacency_Matrix[j,i] = Weight
        

A = Weighted_Adjacency_Matrix
B = np.identity(NUMBER_OF_ATOMIC_AGENTS) 

System_List = {"A": A,
               "B": B,
               "Weighted_Adjacency_Matrix": Weighted_Adjacency_Matrix,
               "NUMBER_OF_ATOMIC_AGENTS": NUMBER_OF_ATOMIC_AGENTS,
               "Reference_Signal": Reference_Signal,
               "X_INITIAL": X_INITIAL}

if EXPORT_DATA == True:
    with open(INPUT_FOLDER + "/" + File_Name, 'wb') as Write_File:
        pk.dump(System_List, Write_File)
    df = pd.DataFrame(Weighted_Adjacency_Matrix)
    df.to_csv(INPUT_FOLDER + "/" + File_Name + ".csv", header=False, index=False)

if DRAW_PLOTS == True:
    # Build system graph
    Network_Graph = nx.Graph()
    for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
        for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
            if i != j:
                if Weighted_Adjacency_Matrix[i,j] != 0:
                    Network_Graph.add_edge(i, j, weight=Weighted_Adjacency_Matrix[i,j])
    
    degree_sequence = sorted((d for n, d in Network_Graph.degree()), reverse=True)
    dmax = max(degree_sequence)
    fig = plt.figure("Random graph topology", figsize=(8, 8))
    ax0 = plt.gca()
    Gcc = Network_Graph.subgraph(sorted(nx.connected_components(Network_Graph), key=len, reverse=True)[0])
    # pos = nx.spring_layout(Network_Graph)
    pos = nx.arf_layout(Network_Graph, seed=0)
    # pos = nx.rescale_layout(Network_Graph)
    
    plot_colors = dict(Network_Graph.degree())

    Edges_Alpha = np.zeros((NUMBER_OF_ATOMIC_AGENTS*NUMBER_OF_ATOMIC_AGENTS,1))
    external_idx = 0
    for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
        for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
            if i != j:
                if Weighted_Adjacency_Matrix[i,j] != 0:
                    Edges_Alpha[external_idx] = Weighted_Adjacency_Matrix[i,j]
                    external_idx += 1
    non_zero_mask = Edges_Alpha != 0
    Edges_Alpha_Filtered = Edges_Alpha[non_zero_mask]

    nx.draw_networkx_nodes(Network_Graph, pos, ax=ax0,  node_size=50,cmap= "Blues", node_color=list(plot_colors.values()), edgecolors='cyan', linewidths=0.4)

    nx.draw_networkx_edges(Network_Graph, pos, ax=ax0, alpha=0.3, edge_color="green", width=Edges_Alpha_Filtered*3)
    fig.tight_layout()
    plt.title("Random graph with 100 nodes", fontsize=18)
    plt.show()

    if EXPORT_PLOTS == True:
        fig.savefig(INPUT_FOLDER + "/" + File_Name + "_GraphPlot.pdf", dpi=300, bbox_inches='tight')       
        fig.savefig(INPUT_FOLDER + "/" + File_Name + "_GraphPlot.png", dpi=300, bbox_inches='tight')       

# %% MARK: Random 50

File_Name = "Random_50"

rnd.seed(0)
NUMBER_OF_ATOMIC_AGENTS =  50
AGENT_DYNAMICS = 0.2

Weighted_Adjacency_Matrix = np.zeros((NUMBER_OF_ATOMIC_AGENTS,NUMBER_OF_ATOMIC_AGENTS))

Row_Index = 0
Col_Index = 0

# Ensure connectedness
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    Assignement_Performed = False
    while Assignement_Performed == False:
        j = rnd.randint(0, NUMBER_OF_ATOMIC_AGENTS-1)
        if i != j:
            Weight = rnd.randrange(100,800,1)/1000
            Weighted_Adjacency_Matrix[i,j] = Weight
            Assignement_Performed = True

# Connectedness cap
Connectedness_Cap = np.zeros((NUMBER_OF_ATOMIC_AGENTS,1))
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    Connectedness_Cap[i] = rnd.randrange(50,80,1)/100

# Assign self-edges
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    for j in range(i,NUMBER_OF_ATOMIC_AGENTS):
        if i == j:
            Weighted_Adjacency_Matrix[i,j] = AGENT_DYNAMICS

# Random topology assignment
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    for j in range(i,NUMBER_OF_ATOMIC_AGENTS):
        if i != j:
            Assign_Chance = rnd.random()
            if Assign_Chance > 0.5:
                Weight = rnd.random()
                if Weighted_Adjacency_Matrix[i,j] == 0:
                    if np.sum(Weighted_Adjacency_Matrix[i,:] + Weight) <= Connectedness_Cap[i]:
                        Weighted_Adjacency_Matrix[i,j] = Weight
                        Weighted_Adjacency_Matrix[j,i] = Weight
        

X_INITIAL = np.zeros((NUMBER_OF_ATOMIC_AGENTS,1))


rnd.seed(0)
for i in range(NUMBER_OF_ATOMIC_AGENTS):
    X_INITIAL[i] = rnd.randint(-4000,4000)/10000

Reference_Signal = np.zeros((NUMBER_OF_ATOMIC_AGENTS,SIMULATION_HORIZON))# Sinusoid 
rnd.seed(0)
for i in range(0,len(Reference_Signal)):
    Frequency = 2*np.pi/(rnd.randrange(20,100,1))
    Amplitude = rnd.randrange(1,100,1)/100
    Phase = 2*np.pi*rnd.randrange(100,1000,1)/1000
    for k in range(0,len(Reference_Signal[i])):
        Reference_Signal[i,k] = Amplitude*np.sin((Frequency * k) + Phase)
        
A = Weighted_Adjacency_Matrix
B = np.identity(NUMBER_OF_ATOMIC_AGENTS) 

System_List = {"A": A,
               "B": B,
               "Weighted_Adjacency_Matrix": Weighted_Adjacency_Matrix,
               "NUMBER_OF_ATOMIC_AGENTS": NUMBER_OF_ATOMIC_AGENTS,
               "X_INITIAL": X_INITIAL,
               "Reference_Signal": Reference_Signal}


if EXPORT_DATA == True:
    with open(INPUT_FOLDER + "/" + File_Name, 'wb') as Write_File:
        pk.dump(System_List, Write_File)
    df = pd.DataFrame(Weighted_Adjacency_Matrix)
    df.to_csv(INPUT_FOLDER + "/" + File_Name + ".csv", header=False, index=False)

        

if DRAW_PLOTS == True:
    # Build system graph
    Network_Graph = nx.Graph()
    for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
        for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
            if i != j:
                if Weighted_Adjacency_Matrix[i,j] != 0:
                    Network_Graph.add_edge(i, j, weight=Weighted_Adjacency_Matrix[i,j])
    
    degree_sequence = sorted((d for n, d in Network_Graph.degree()), reverse=True)
    dmax = max(degree_sequence)
    fig = plt.figure("Random graph topology", figsize=(8, 8))
    ax0 = plt.gca()
    Gcc = Network_Graph.subgraph(sorted(nx.connected_components(Network_Graph), key=len, reverse=True)[0])
    # pos = nx.spring_layout(Network_Graph)
    pos = nx.arf_layout(Network_Graph, seed=0)
    # pos = nx.rescale_layout(Network_Graph)
    
    plot_colors = dict(Network_Graph.degree())

    Edges_Alpha = np.zeros((NUMBER_OF_ATOMIC_AGENTS*NUMBER_OF_ATOMIC_AGENTS,1))
    external_idx = 0
    for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
        for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
            if i != j:
                if Weighted_Adjacency_Matrix[i,j] != 0:
                    Edges_Alpha[external_idx] = Weighted_Adjacency_Matrix[i,j]
                    external_idx += 1
    non_zero_mask = Edges_Alpha != 0
    Edges_Alpha_Filtered = Edges_Alpha[non_zero_mask]

    nx.draw_networkx_nodes(Network_Graph, pos, ax=ax0,  node_size=50,cmap= "Blues", node_color=list(plot_colors.values()), edgecolors='cyan', linewidths=0.4)

    nx.draw_networkx_edges(Network_Graph, pos, ax=ax0, alpha=0.3, edge_color="green", width=Edges_Alpha_Filtered*3)
    fig.tight_layout()
    plt.title("Random graph with 50 nodes", fontsize=18)
    plt.show()

    if EXPORT_PLOTS == True:
        fig.savefig(INPUT_FOLDER + "/" + File_Name + "_GraphPlot.pdf", dpi=300, bbox_inches='tight')       
        fig.savefig(INPUT_FOLDER + "/" + File_Name + "_GraphPlot.png", dpi=300, bbox_inches='tight')     


# %% MARK: Random 50 circular plot

# Build system graph
Network_Graph = nx.Graph()
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
        if i != j:
            if Weighted_Adjacency_Matrix[i,j] != 0:
                Network_Graph.add_edge(i, j, weight=Weighted_Adjacency_Matrix[i,j])

degree_sequence = dict(Network_Graph.degree())
Number_Rings = max(degree_sequence.values()) + 1

Ring_Radius = 10
Total_Circumference_Radius = Ring_Radius*Number_Rings
Agents_Sorted = create_empty_list(Number_Rings)

for idx in range(len(Agents_Sorted)):
    for i in range(NUMBER_OF_ATOMIC_AGENTS):
        if degree_sequence[i] == idx:
            Agents_Sorted[idx].append(i) 

Nodes_Positions = {}
Circumference_Radius = Total_Circumference_Radius
# External_Idx = 0
for idx in range(1,len(Agents_Sorted)):
    External_Idx = 0
    for i in range(len(Agents_Sorted[idx])):
        Agent_Idx = Agents_Sorted[idx][i]
        Nodes_Positions[Agent_Idx] = [Circumference_Radius*np.sin(((2*np.pi)/len(Agents_Sorted[idx]))*External_Idx),
                                      Circumference_Radius*np.cos(((2*np.pi)/len(Agents_Sorted[idx]))*External_Idx) ]
        External_Idx += 1
    
    Circumference_Radius = Circumference_Radius - Ring_Radius

# PLOT

Graph_Figure = plt.figure("Random graph topology", figsize=(8, 8))
# fig = plt.figure("Random graph topology")
Graph_Axis = plt.gca()
# Gcc = Network_Graph.subgraph(sorted(nx.connected_components(Network_Graph), key=len, reverse=True)[0])
# pos = nx.spring_layout(Network_Graph)

# plt.autoscale(False)
plot_colors = dict(Network_Graph.degree())

Edges_Alpha = np.zeros((NUMBER_OF_ATOMIC_AGENTS*NUMBER_OF_ATOMIC_AGENTS,1))
external_idx = 0
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
        if i != j:
            if Weighted_Adjacency_Matrix[i,j] != 0:
                Edges_Alpha[external_idx] = Weighted_Adjacency_Matrix[i,j]
                external_idx += 1
non_zero_mask = Edges_Alpha != 0
Edges_Alpha_Filtered = Edges_Alpha[non_zero_mask]

# nx.draw_networkx_nodes(Network_Graph, pos, ax=ax0,  node_size=50,cmap= "Blues", node_color=list(plot_colors.values()), edgecolors='cyan', linewidths=0.4)

radius_circle = 3.5
number_points_circumference = 100

Node_Degrees_List = Network_Graph.degree()
Table_Colors = mcolors.TABLEAU_COLORS

Color_List_Base = list(Table_Colors.keys())
Color_List = create_empty_list(NUMBER_OF_ATOMIC_AGENTS)
external_idx = 0
for idx in range(NUMBER_OF_ATOMIC_AGENTS):
    Color_List[idx] = Color_List_Base[external_idx]
    external_idx += 1
    if external_idx >= len(Color_List_Base):
        external_idx = 0

pos = Nodes_Positions
for idx in range(NUMBER_OF_ATOMIC_AGENTS):

    agent_idx_i = idx
    
    circle = plt.Circle((pos[agent_idx_i][0], pos[agent_idx_i][1]), radius_circle, color='black', alpha=1, zorder=2)
    Graph_Axis.add_patch(circle)
    circle = plt.Circle((pos[agent_idx_i][0], pos[agent_idx_i][1]), radius_circle*0.9, color="white", alpha=1 , zorder=2)
    Graph_Axis.add_patch(circle)
    circle = plt.Circle((pos[agent_idx_i][0], pos[agent_idx_i][1]), radius_circle*0.9, color="cyan", alpha= (Node_Degrees_List[idx]/(dmax)) , zorder=2)

    Graph_Axis.add_patch(circle)
    plt.text(pos[agent_idx_i][0], pos[agent_idx_i][1], f"$\mathcal{{A}}_{{{str(agent_idx_i)}}}$", color="black", fontsize=12, horizontalalignment="center", verticalalignment="center")

for idx in range(NUMBER_OF_ATOMIC_AGENTS):
    agent_idx_i = idx

    for j in range(NUMBER_OF_ATOMIC_AGENTS):
        agent_idx_j = j

        if Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j] != 0:
            plt.plot([pos[agent_idx_i][0], pos[agent_idx_j][0]], [pos[agent_idx_i][1], pos[agent_idx_j][1]], color="green", zorder=1,linewidth=3.5*Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j], alpha=0.5)

plt.xlim(-1, 8)
plt.ylim(-1, 8)
plt.axis("equal")
plt.axis('off') 
Graph_Figure.patch.set_linewidth(1)
Graph_Figure.patch.set_edgecolor('black')
Graph_Figure.tight_layout()  
plt.tick_params(axis='both',        # Apply to both x and y axes
                which='both',       # Apply to both major and minor ticks
                bottom=False,       # Turn off bottom ticks
                top=False,          # Turn off top ticks
                left=False,         # Turn off left ticks
                right=False,        # Turn off right ticks
                labelbottom=False,  # Turn off bottom labels
                labelleft=False)    # Turn off left labels
# plt.title("Random graph with 50 nodes", fontsize=18)
plt.show()

Graph_Figure.savefig(INPUT_FOLDER + "/" + File_Name + "_Graph_Figure.pdf", dpi=300, bbox_inches=None)
Graph_Figure.savefig(INPUT_FOLDER + "/" + File_Name + "_Graph_Figure.jpeg", dpi=300, bbox_inches=None)

# %% MARK: Random 256

File_Name = "Random_256"

rnd.seed(1)
NUMBER_OF_ATOMIC_AGENTS =  256
AGENT_DYNAMICS = 0.2

Weighted_Adjacency_Matrix = np.zeros((NUMBER_OF_ATOMIC_AGENTS,NUMBER_OF_ATOMIC_AGENTS))

Row_Index = 0
Col_Index = 0

# Ensure connectedness
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    Assignement_Performed = False
    while Assignement_Performed == False:
        j = rnd.randint(0, NUMBER_OF_ATOMIC_AGENTS-1)
        if i != j:
            Weight = rnd.randrange(100,800,1)/1000
            Weighted_Adjacency_Matrix[i,j] = Weight
            Assignement_Performed = True

# Connectedness cap
Connectedness_Cap = np.zeros((NUMBER_OF_ATOMIC_AGENTS,1))
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    Connectedness_Cap[i] = rnd.randrange(50,80,1)/100

# Assign self-edges
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    for j in range(i,NUMBER_OF_ATOMIC_AGENTS):
        if i == j:
            Weighted_Adjacency_Matrix[i,j] = AGENT_DYNAMICS

# Random topology assignment
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    for j in range(i,NUMBER_OF_ATOMIC_AGENTS):
        if i != j:
            Assign_Chance = rnd.random()
            if Assign_Chance > 0.5:
                Weight = rnd.random()
                if Weighted_Adjacency_Matrix[i,j] == 0:
                    if np.sum(Weighted_Adjacency_Matrix[i,:] + Weight) <= Connectedness_Cap[i]:
                        Weighted_Adjacency_Matrix[i,j] = Weight
                        Weighted_Adjacency_Matrix[j,i] = Weight
        

A = Weighted_Adjacency_Matrix
B = np.identity(NUMBER_OF_ATOMIC_AGENTS) 

System_List = {"A": A,
               "B": B,
               "Weighted_Adjacency_Matrix": Weighted_Adjacency_Matrix,
               "NUMBER_OF_ATOMIC_AGENTS": NUMBER_OF_ATOMIC_AGENTS}

if EXPORT_DATA == True:
    with open(INPUT_FOLDER + "/" + File_Name, 'wb') as Write_File:
        pk.dump(System_List, Write_File)

if DRAW_PLOTS == True:
    # Build system graph
    Network_Graph = nx.Graph()
    for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
        for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
            if i != j:
                if Weighted_Adjacency_Matrix[i,j] != 0:
                    Network_Graph.add_edge(i, j, weight=Weighted_Adjacency_Matrix[i,j])
    
    degree_sequence = sorted((d for n, d in Network_Graph.degree()), reverse=True)
    dmax = max(degree_sequence)
    fig = plt.figure("Random graph topology", figsize=(8, 8))
    ax0 = plt.gca()
    Gcc = Network_Graph.subgraph(sorted(nx.connected_components(Network_Graph), key=len, reverse=True)[0])
    pos = nx.spring_layout(Network_Graph)
    # pos = nx.arf_layout(Network_Graph, seed=0)
    # pos = nx.spiral_layout(Network_Graph)
    
    
    plot_colors = dict(Network_Graph.degree())

    Edges_Alpha = np.zeros((NUMBER_OF_ATOMIC_AGENTS*NUMBER_OF_ATOMIC_AGENTS,1))
    external_idx = 0
    for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
        for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
            if i != j:
                if Weighted_Adjacency_Matrix[i,j] != 0:
                    Edges_Alpha[external_idx] = Weighted_Adjacency_Matrix[i,j]
                    external_idx += 1
    non_zero_mask = Edges_Alpha != 0
    Edges_Alpha_Filtered = Edges_Alpha[non_zero_mask]

    nx.draw_networkx_nodes(Network_Graph, pos, ax=ax0,  node_size=50,cmap= "Blues", node_color=list(plot_colors.values()), edgecolors='cyan', linewidths=0.4)

    nx.draw_networkx_edges(Network_Graph, pos, ax=ax0, alpha=0.3, edge_color="green", width=Edges_Alpha_Filtered*3)
    fig.tight_layout()
    plt.title("Random graph with 256 nodes", fontsize=18)
    plt.show()

    if EXPORT_PLOTS == True:
        fig.savefig(INPUT_FOLDER + "/" + File_Name + "_GraphPlot.pdf", dpi=300, bbox_inches='tight')       
        fig.savefig(INPUT_FOLDER + "/" + File_Name + "_GraphPlot.png", dpi=300, bbox_inches='tight')       


# %% MARK: Random 256 test new layout
# Build system graph
Network_Graph = nx.Graph()
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
        if i != j:
            if Weighted_Adjacency_Matrix[i,j] != 0:
                Network_Graph.add_edge(i, j, weight=Weighted_Adjacency_Matrix[i,j])

degree_sequence = dict(Network_Graph.degree())
Number_Rings = max(degree_sequence.values()) + 1

Ring_Radius = 10
Total_Circumference_Radius = Ring_Radius*Number_Rings
Agents_Sorted = create_empty_list(Number_Rings)

for idx in range(len(Agents_Sorted)):
    for i in range(NUMBER_OF_ATOMIC_AGENTS):
        if degree_sequence[i] == idx:
            Agents_Sorted[idx].append(i) 

Nodes_Positions = {}
Circumference_Radius = Total_Circumference_Radius
# External_Idx = 0
for idx in range(1,len(Agents_Sorted)):
    External_Idx = 0
    for i in range(len(Agents_Sorted[idx])):
        Agent_Idx = Agents_Sorted[idx][i]
        Nodes_Positions[Agent_Idx] = [Circumference_Radius*np.sin(((2*np.pi)/len(Agents_Sorted[idx]))*External_Idx),
                                      Circumference_Radius*np.cos(((2*np.pi)/len(Agents_Sorted[idx]))*External_Idx) ]
        External_Idx += 1
    
    Circumference_Radius = Circumference_Radius - Ring_Radius

# PLOT

Graph_Figure = plt.figure("Random graph topology", figsize=(8, 8))
# fig = plt.figure("Random graph topology")
Graph_Axis = plt.gca()
# Gcc = Network_Graph.subgraph(sorted(nx.connected_components(Network_Graph), key=len, reverse=True)[0])
# pos = nx.spring_layout(Network_Graph)

# plt.autoscale(False)
plot_colors = dict(Network_Graph.degree())

Edges_Alpha = np.zeros((NUMBER_OF_ATOMIC_AGENTS*NUMBER_OF_ATOMIC_AGENTS,1))
external_idx = 0
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
        if i != j:
            if Weighted_Adjacency_Matrix[i,j] != 0:
                Edges_Alpha[external_idx] = Weighted_Adjacency_Matrix[i,j]
                external_idx += 1
non_zero_mask = Edges_Alpha != 0
Edges_Alpha_Filtered = Edges_Alpha[non_zero_mask]

# nx.draw_networkx_nodes(Network_Graph, pos, ax=ax0,  node_size=50,cmap= "Blues", node_color=list(plot_colors.values()), edgecolors='cyan', linewidths=0.4)

radius_circle = 0.5
number_points_circumference = 100

Node_Degrees_List = Network_Graph.degree()
Table_Colors = mcolors.TABLEAU_COLORS

Color_List_Base = list(Table_Colors.keys())
Color_List = create_empty_list(NUMBER_OF_ATOMIC_AGENTS)
external_idx = 0
for idx in range(NUMBER_OF_ATOMIC_AGENTS):
    Color_List[idx] = Color_List_Base[external_idx]
    external_idx += 1
    if external_idx >= len(Color_List_Base):
        external_idx = 0

pos = Nodes_Positions
for idx in range(NUMBER_OF_ATOMIC_AGENTS):

    agent_idx_i = idx
    
    circle = plt.Circle((pos[agent_idx_i][0], pos[agent_idx_i][1]), radius_circle, color='black', alpha=1, zorder=2)
    Graph_Axis.add_patch(circle)
    circle = plt.Circle((pos[agent_idx_i][0], pos[agent_idx_i][1]), radius_circle*0.9, color="white", alpha=1 , zorder=2)
    Graph_Axis.add_patch(circle)
    circle = plt.Circle((pos[agent_idx_i][0], pos[agent_idx_i][1]), radius_circle*0.9, color="Green", alpha= (Node_Degrees_List[idx]/(dmax)) , zorder=2)

    Graph_Axis.add_patch(circle)
    # plt.text(pos[agent_idx_i][0], pos[agent_idx_i][1], str(agent_idx_i), color="black", fontsize=12, horizontalalignment="center", verticalalignment="center")

for idx in range(NUMBER_OF_ATOMIC_AGENTS):
    agent_idx_i = idx

    for j in range(NUMBER_OF_ATOMIC_AGENTS):
        agent_idx_j = j

        if Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j] != 0:
            plt.plot([pos[agent_idx_i][0], pos[agent_idx_j][0]], [pos[agent_idx_i][1], pos[agent_idx_j][1]], color="cyan", zorder=1,linewidth=2.5*Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j], alpha=0.5)

plt.tick_params(axis='both',        # Apply to both x and y axes
                which='both',       # Apply to both major and minor ticks
                bottom=False,       # Turn off bottom ticks
                top=False,          # Turn off top ticks
                left=False,         # Turn off left ticks
                right=False,        # Turn off right ticks
                labelbottom=False,  # Turn off bottom labels
                labelleft=False)    # Turn off left labels

# %% MARK: Random 1024

File_Name = "Random_1024"

rnd.seed(1)
NUMBER_OF_ATOMIC_AGENTS =  1024
AGENT_DYNAMICS = 0.2

Weighted_Adjacency_Matrix = np.zeros((NUMBER_OF_ATOMIC_AGENTS,NUMBER_OF_ATOMIC_AGENTS))

Row_Index = 0
Col_Index = 0

# Ensure connectedness
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    Assignement_Performed = False
    while Assignement_Performed == False:
        j = rnd.randint(0, NUMBER_OF_ATOMIC_AGENTS-1)
        if i != j:
            Weight = rnd.randrange(100,800,1)/1000
            Weighted_Adjacency_Matrix[i,j] = Weight
            Assignement_Performed = True

# Connectedness cap
Connectedness_Cap = np.zeros((NUMBER_OF_ATOMIC_AGENTS,1))
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    Connectedness_Cap[i] = rnd.randrange(50,80,1)/100

# Assign self-edges
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    for j in range(i,NUMBER_OF_ATOMIC_AGENTS):
        if i == j:
            Weighted_Adjacency_Matrix[i,j] = AGENT_DYNAMICS

# Random topology assignment
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    for j in range(i,NUMBER_OF_ATOMIC_AGENTS):
        if i != j:
            Assign_Chance = rnd.random()
            if Assign_Chance > 0.5:
                Weight = rnd.random()
                if Weighted_Adjacency_Matrix[i,j] == 0:
                    if np.sum(Weighted_Adjacency_Matrix[i,:] + Weight) <= Connectedness_Cap[i]:
                        Weighted_Adjacency_Matrix[i,j] = Weight
                        Weighted_Adjacency_Matrix[j,i] = Weight
        

A = Weighted_Adjacency_Matrix
B = np.identity(NUMBER_OF_ATOMIC_AGENTS) 

System_List = {"A": A,
               "B": B,
               "Weighted_Adjacency_Matrix": Weighted_Adjacency_Matrix,
               "NUMBER_OF_ATOMIC_AGENTS": NUMBER_OF_ATOMIC_AGENTS}

if EXPORT_DATA == True:
    with open(INPUT_FOLDER + "/" + File_Name, 'wb') as Write_File:
        pk.dump(System_List, Write_File)

if DRAW_PLOTS == True:
    # Build system graph
    Network_Graph = nx.Graph()
    for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
        for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
            if i != j:
                if Weighted_Adjacency_Matrix[i,j] != 0:
                    Network_Graph.add_edge(i, j, weight=Weighted_Adjacency_Matrix[i,j])
    
    degree_sequence = sorted((d for n, d in Network_Graph.degree()), reverse=True)
    dmax = max(degree_sequence)
    fig = plt.figure("Random graph topology", figsize=(8, 8))
    ax0 = plt.gca()
    Gcc = Network_Graph.subgraph(sorted(nx.connected_components(Network_Graph), key=len, reverse=True)[0])
    # pos = nx.spring_layout(Network_Graph)
    pos = nx.forceatlas2_layout(Network_Graph)
    
    plot_colors = dict(Network_Graph.degree())

    Edges_Alpha = np.zeros((NUMBER_OF_ATOMIC_AGENTS*NUMBER_OF_ATOMIC_AGENTS,1))
    external_idx = 0
    for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
        for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
            if i != j:
                if Weighted_Adjacency_Matrix[i,j] != 0:
                    Edges_Alpha[external_idx] = Weighted_Adjacency_Matrix[i,j]
                    external_idx += 1
    non_zero_mask = Edges_Alpha != 0
    Edges_Alpha_Filtered = Edges_Alpha[non_zero_mask]

    nx.draw_networkx_nodes(Network_Graph, pos, ax=ax0,  node_size=50,cmap= "Blues", node_color=list(plot_colors.values()), edgecolors='cyan', linewidths=0.4)

    nx.draw_networkx_edges(Network_Graph, pos, ax=ax0, alpha=0.3, edge_color="green", width=Edges_Alpha_Filtered*3)
    fig.tight_layout()
    plt.title("Random graph with 1024 nodes", fontsize=18)
    plt.show()

    if EXPORT_PLOTS == True:
        fig.savefig(INPUT_FOLDER + "/" + File_Name + "_GraphPlot.pdf", dpi=300, bbox_inches='tight')       
        fig.savefig(INPUT_FOLDER + "/" + File_Name + "_GraphPlot.png", dpi=300, bbox_inches='tight')   


# %% MARK: Modular 64 

File_Name = "Modular_64"

NUMBER_OF_ATOMIC_AGENTS =  64

# Creation of the list of atomic agents
List_Atomic_Agents = []
for idx in range(NUMBER_OF_ATOMIC_AGENTS):
    List_Atomic_Agents.append(idx)

COUPLING_STRENGTH_0 = 0.5
COUPLING_STRENGTH_1 = 0.1
COUPLING_STRENGTH_2 = 0.01
COUPLING_STRENGTH_3 = 0.001

Weighted_Adjacency_Matrix = np.zeros((NUMBER_OF_ATOMIC_AGENTS,NUMBER_OF_ATOMIC_AGENTS))
for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
    Weighted_Adjacency_Matrix[i,i] = COUPLING_STRENGTH_0

###################################################### Level 1
# Group 1
Weighted_Adjacency_Matrix[0,1] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[0,2] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[0,3] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[1,0] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[1,2] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[1,3] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[2,0] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[2,1] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[2,3] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[3,0] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[3,1] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[3,2] = COUPLING_STRENGTH_1

# Group 2
Weighted_Adjacency_Matrix[4,5] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[4,6] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[4,7] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[5,4] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[5,6] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[5,7] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[6,4] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[6,5] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[6,7] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[7,4] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[7,5] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[7,6] = COUPLING_STRENGTH_1

# Group 3
Weighted_Adjacency_Matrix[8,9] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[8,10] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[8,11] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[9,8] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[9,10] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[9,11] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[10,8] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[10,9] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[10,11] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[11,8] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[11,9] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[11,10] = COUPLING_STRENGTH_1

# Group 4
Weighted_Adjacency_Matrix[12,13] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[12,14] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[12,15] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[13,12] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[13,14] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[13,15] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[14,12] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[14,13] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[14,15] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[15,12] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[15,13] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[15,14] = COUPLING_STRENGTH_1

# Group 5
Weighted_Adjacency_Matrix[16,17] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[16,18] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[16,19] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[17,16] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[17,18] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[17,19] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[18,16] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[18,17] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[18,19] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[19,16] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[19,17] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[19,18] = COUPLING_STRENGTH_1

# Group 6
Weighted_Adjacency_Matrix[20,21] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[20,22] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[20,23] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[21,20] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[21,22] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[21,23] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[22,20] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[22,21] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[22,23] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[23,20] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[23,21] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[23,22] = COUPLING_STRENGTH_1

# Group 7
Weighted_Adjacency_Matrix[24,25] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[24,26] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[24,27] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[25,24] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[25,26] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[25,27] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[26,24] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[26,25] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[26,27] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[27,24] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[27,25] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[27,26] = COUPLING_STRENGTH_1

# Group 8
Weighted_Adjacency_Matrix[28,29] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[28,30] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[28,31] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[29,28] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[29,30] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[29,31] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[30,28] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[30,29] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[30,31] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[31,28] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[31,29] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[31,30] = COUPLING_STRENGTH_1

# Group 9
Weighted_Adjacency_Matrix[32,33] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[32,34] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[32,35] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[33,32] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[33,34] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[33,35] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[34,32] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[34,33] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[34,35] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[35,32] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[35,33] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[35,34] = COUPLING_STRENGTH_1

# Group 10
Weighted_Adjacency_Matrix[36,37] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[36,38] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[36,39] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[37,36] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[37,38] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[37,39] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[38,36] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[38,37] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[38,39] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[39,36] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[39,37] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[39,38] = COUPLING_STRENGTH_1

# Group 11
Weighted_Adjacency_Matrix[40,41] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[40,42] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[40,43] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[41,40] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[41,42] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[41,43] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[42,40] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[42,41] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[42,43] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[43,40] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[43,41] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[43,42] = COUPLING_STRENGTH_1

# Group 12
Weighted_Adjacency_Matrix[44,45] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[44,46] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[44,47] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[45,44] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[45,46] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[45,47] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[46,44] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[46,45] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[46,47] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[47,44] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[47,45] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[47,46] = COUPLING_STRENGTH_1

# Group 13
Weighted_Adjacency_Matrix[48,49] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[48,50] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[48,51] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[49,48] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[49,50] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[49,51] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[50,48] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[50,49] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[50,51] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[51,48] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[51,49] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[51,50] = COUPLING_STRENGTH_1

# Group 14
Weighted_Adjacency_Matrix[52,53] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[52,54] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[52,55] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[53,52] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[53,54] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[53,55] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[54,52] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[54,53] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[54,55] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[55,52] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[55,53] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[55,54] = COUPLING_STRENGTH_1

# Group 15
Weighted_Adjacency_Matrix[56,57] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[56,58] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[56,59] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[57,56] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[57,58] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[57,59] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[58,56] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[58,57] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[58,59] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[59,56] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[59,57] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[59,58] = COUPLING_STRENGTH_1

# Group 16
Weighted_Adjacency_Matrix[60,61] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[60,62] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[60,63] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[61,60] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[61,62] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[61,63] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[62,60] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[62,61] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[62,63] = COUPLING_STRENGTH_1

Weighted_Adjacency_Matrix[63,60] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[63,61] = COUPLING_STRENGTH_1
Weighted_Adjacency_Matrix[63,62] = COUPLING_STRENGTH_1

###################################################### Level 2
# Group 1
Weighted_Adjacency_Matrix[3,6] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[3,9] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[3,12] = COUPLING_STRENGTH_2

Weighted_Adjacency_Matrix[6,3] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[6,9] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[6,12] = COUPLING_STRENGTH_2

Weighted_Adjacency_Matrix[9,3] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[9,6] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[9,12] = COUPLING_STRENGTH_2

Weighted_Adjacency_Matrix[12,3] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[12,6] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[12,9] = COUPLING_STRENGTH_2

# Group 2
Weighted_Adjacency_Matrix[19,22] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[19,25] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[19,28] = COUPLING_STRENGTH_2

Weighted_Adjacency_Matrix[22,19] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[22,25] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[22,28] = COUPLING_STRENGTH_2

Weighted_Adjacency_Matrix[25,19] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[25,22] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[25,28] = COUPLING_STRENGTH_2

Weighted_Adjacency_Matrix[28,19] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[28,22] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[28,25] = COUPLING_STRENGTH_2

# Group 3
Weighted_Adjacency_Matrix[35,38] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[35,41] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[35,44] = COUPLING_STRENGTH_2

Weighted_Adjacency_Matrix[38,35] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[38,41] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[38,44] = COUPLING_STRENGTH_2

Weighted_Adjacency_Matrix[41,35] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[41,38] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[41,44] = COUPLING_STRENGTH_2

Weighted_Adjacency_Matrix[44,35] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[44,38] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[44,41] = COUPLING_STRENGTH_2

# Group 4
Weighted_Adjacency_Matrix[51,54] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[51,57] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[51,60] = COUPLING_STRENGTH_2

Weighted_Adjacency_Matrix[54,51] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[54,57] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[54,60] = COUPLING_STRENGTH_2

Weighted_Adjacency_Matrix[57,51] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[57,54] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[57,60] = COUPLING_STRENGTH_2

Weighted_Adjacency_Matrix[60,51] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[60,54] = COUPLING_STRENGTH_2
Weighted_Adjacency_Matrix[60,57] = COUPLING_STRENGTH_2

###################################################### Level 3
Weighted_Adjacency_Matrix[15,26] = COUPLING_STRENGTH_3
Weighted_Adjacency_Matrix[15,37] = COUPLING_STRENGTH_3
Weighted_Adjacency_Matrix[15,48] = COUPLING_STRENGTH_3

Weighted_Adjacency_Matrix[26,15] = COUPLING_STRENGTH_3
Weighted_Adjacency_Matrix[26,37] = COUPLING_STRENGTH_3
Weighted_Adjacency_Matrix[26,48] = COUPLING_STRENGTH_3

Weighted_Adjacency_Matrix[37,15] = COUPLING_STRENGTH_3
Weighted_Adjacency_Matrix[37,26] = COUPLING_STRENGTH_3
Weighted_Adjacency_Matrix[37,48] = COUPLING_STRENGTH_3

Weighted_Adjacency_Matrix[48,15] = COUPLING_STRENGTH_3
Weighted_Adjacency_Matrix[48,26] = COUPLING_STRENGTH_3
Weighted_Adjacency_Matrix[48,37] = COUPLING_STRENGTH_3

######################################################

A = Weighted_Adjacency_Matrix
B = np.identity(NUMBER_OF_ATOMIC_AGENTS) 

X_INITIAL = np.zeros((NUMBER_OF_ATOMIC_AGENTS,1))


rnd.seed(0)
for i in range(NUMBER_OF_ATOMIC_AGENTS):
    X_INITIAL[i] = rnd.randint(-4000,4000)/10000

System_List = {"A": A,
               "B": B,
               "Weighted_Adjacency_Matrix": Weighted_Adjacency_Matrix,
               "NUMBER_OF_ATOMIC_AGENTS": NUMBER_OF_ATOMIC_AGENTS,
               "X_INITIAL": X_INITIAL}

if EXPORT_DATA == True:
    with open(INPUT_FOLDER + "/" + File_Name, 'wb') as Write_File:
        pk.dump(System_List, Write_File)
    df = pd.DataFrame(Weighted_Adjacency_Matrix)
    df.to_csv(INPUT_FOLDER + "/" + File_Name + ".csv", header=False, index=False)
# %%
