# %% 
# MARK: Reset Operations
# from IPython import get_ipython
# get_ipython().run_line_magic('reset', '-f')
# print("\n" * 100)

# %%
# MARK: Import libraries


# For matrix operations
import numpy as np
# For file storage
import pickle as pk
# For time tracking
import time as tm
# For data location
INPUT_FOLDER = "Input_Data"
OUTPUT_FOLDER = "A03_Output_Data_Algorithmic_Partitioning_Mixed"
import os
if not os.path.exists(OUTPUT_FOLDER):
    os.mkdir(OUTPUT_FOLDER)


import networkx as nx
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib import rc
# Enable LaTeX rendering
rc('text', usetex=True)

# Set the font to Times New Roman or the closest available font in LaTeX
rc('font', family='serif')
rc('font', serif='Times New Roman')  # This line might need adjustment based on LaTeX distribution

from scipy.spatial import ConvexHull
import matplotlib.colors as mcolors

# Support functions
from A00_Support_Funcitons import *
from A02_Weighting_Function import*

# Settings
EXPORT_DATA = True
DRAW_PLOTS = False
EXPORT_PLOTS = False

# %% 
# MARK: Parameters

# Modular_64 
# ALPHA = 25 # 64 Modules of 1 agent
# Partition_Number = "Partition_01"
# ALPHA = 1 # 16 Modules of 4 agents 
# Partition_Number = "Partition_02"
# ALPHA = 1e-6 # Mixed results
# Partition_Number = "Partition_03" 

# Random_50 
# ALPHA = 100  # 50 agents
# Partition_Number = "Partition_01"
# ALPHA = 30 # 11 agents
# Partition_Number = "Partition_02"
# ALPHA = 20  # 11 agents
# Partition_Number = "Partition_03"
# ALPHA = 10 # 5 agents
# Partition_Number = "Partition_04"
# ALPHA = 1 # 4 agents
# Partition_Number = "Partition_05"
# ALPHA = 0.1 # 4 agents
# Partition_Number = "Partition_06"
# ALPHA = 0.001 # 4 agents
# Partition_Number = "Partition_07"
# ALPHA = 0.00001 # 4 agents
# Partition_Number = "Partition_08"
# ALPHA = 0.0000000000001 # 1 agents
# Partition_Number = "Partition_09"
# ALPHA = 0.000000000001 # 3 agents
# Partition_Number = "Partition_10"
ALPHA = 0.000000000005 # 4 agents
Partition_Number = "Partition_11"


# %%
# MARK: Select network

# Network_Dict = modular64()
# Network_File = "Random_100"
Network_File = "Random_50"
# Network_File = "Modular_64"
# Network_File = "Random_10"

# Network_File = "Random_1024"

with open("Input_Data/" + Network_File, 'rb') as f:
    Network_Dict = pk.load(f)

A = Network_Dict['A']
B = Network_Dict['B']
Weighted_Adjacency_Matrix = Network_Dict['Weighted_Adjacency_Matrix']
NUMBER_OF_ATOMIC_AGENTS = Network_Dict['NUMBER_OF_ATOMIC_AGENTS']
List_Atomic_Agents = []
for i in range(NUMBER_OF_ATOMIC_AGENTS):
    List_Atomic_Agents.append(i)
# print("test")

Partition_ID = Network_File + "_" + Partition_Number + "_ALPHA_" + str(ALPHA)


# %% MARK: Partition refinement function

def Partition_Refinement(Network_Dict, List_Atomic_Agents, Partition, ALPHA):
    

    W_Partition = weighting_function(Network_Dict, List_Atomic_Agents, Partition, ALPHA)

    Delta_Partition_Max = 0

    while Delta_Partition_Max >= 0:

        Next_Agent = None
        Next_Set = None
        Prev_Set = None
        for i in range(len(Partition)):
            for j in range(len(Partition[i])):
                Test_Agent = Partition[i][j]
                Partition[i].remove(Test_Agent)
                for k in range(len(Partition)):
                    if k != i:
                        Partition[k].append(Test_Agent)
                        W_Partition_Test = weighting_function(Network_Dict, List_Atomic_Agents, Partition, ALPHA)
                        Delta_Partition =  W_Partition_Test - W_Partition 
                        if Delta_Partition > Delta_Partition_Max:
                            Prev_Set = i
                            Next_Agent = Test_Agent
                            Next_Set = k
                        Partition[k].remove(Test_Agent)
                Partition[i].append(Test_Agent)
        
        if Next_Agent != None:
            Partition[Prev_Set].remove(Next_Agent)
            Partition[Next_Set].append(Next_Agent)
            W_Partition_New = weighting_function(Network_Dict, List_Atomic_Agents, Partition, ALPHA)
            Delta_Partition_Print = W_Partition_New - W_Partition
            W_Partition = W_Partition_New
            print("Local Partition refinement")
            print("Assign agent ", Next_Agent, " to set ", Next_Set)
            print("Delta Partition:", Delta_Partition_Print )
            print("Partition weight ", W_Partition)           
        else:
            Delta_Partition_Max = -1       
    
    return Partition

# %%
# MARK: Partitioning algorithm

Agents_Left = List_Atomic_Agents.copy()

# Allocation of the empty partitioning
Partition = create_empty_list(NUMBER_OF_ATOMIC_AGENTS)

# %% 

Computation_Time_Start = tm.time()
# MARK: Greedy algorithm partitioning
W_Partition = 0
while len(Agents_Left) > 0:
    print("Computation percentage:" + str(100*((NUMBER_OF_ATOMIC_AGENTS-len(Agents_Left))/NUMBER_OF_ATOMIC_AGENTS)))
    # Candidate_Partition = Partition.copy()
    # W_Partition_Max = W_Partition
    Delta_Partition_Max = float('-inf')
    # Delta_Partition_Max = 0
    Next_Agent = None
    Next_Set = None
    for i in range(len(Agents_Left)):
        Agent_i = Agents_Left[i]
        for j in range(len(Partition)):
            # Candidate_Partition[j].append(Agent_i)
            Partition[j].append(Agent_i)
            # W_Partition_New = weighting_function(Network_Dict, Candidate_Partition, ALPHA)
            W_Partition_New = weighting_function(Network_Dict, List_Atomic_Agents, Partition, ALPHA)
            Delta_Partition = W_Partition_New - W_Partition
            if Delta_Partition  > Delta_Partition_Max:
                Next_Agent = Agent_i
                Next_Set = j
                Delta_Partition_Max = Delta_Partition
                # W_Partition_Max = W_Partition_New 
            # Candidate_Partition[j].remove(Agent_i)
            Partition[j].remove(Agent_i)
    
    Partition[Next_Set].append(Next_Agent)
    Agents_Left.remove(Next_Agent)
    W_Partition = weighting_function(Network_Dict, List_Atomic_Agents, Partition, ALPHA)
    print("Assign agent ", Next_Agent, " to set ", Next_Set)
    print("Partition weight ", W_Partition)

    Partition = Partition_Refinement(Network_Dict, List_Atomic_Agents, Partition, ALPHA)
    Simulation_Timer = tm.time() - Computation_Time_Start
    print("Time enlapsed:", round(Simulation_Timer, 2), "[s]")
    print("Expected time to finish:", round((Simulation_Timer*len(Agents_Left))/(NUMBER_OF_ATOMIC_AGENTS - len(Agents_Left)), 2), "[s]")
    # print("Expected time to finish:", round(((Simulation_Timer_End*NUMBER_OF_ATOMIC_AGENTS)/(len(Agents_Left)+1)) - Simulation_Timer_End, 2), "[s]")
    print("######################################")
                
Computation_Time_Total = tm.time() - Computation_Time_Start


# Remove empty sets
Partition = [item for item in Partition if item != []]
NUMBER_CONTROL_AGENTS = len(Partition)

# Sort list
for idx in range(NUMBER_CONTROL_AGENTS):
    Partition[idx].sort()

# %% Display result
print("Computation time: ", Computation_Time_Total)
print("Number atomic control agents: ", NUMBER_OF_ATOMIC_AGENTS)
print("Number control agents: ", NUMBER_CONTROL_AGENTS)
for idx in range(NUMBER_CONTROL_AGENTS):
    print("Control agent ", idx, ": ", Partition[idx])

# %%
# MARK: Store results

Output_Dict = {"A": A, 
               "B": B, 
               "Weighted_Adjacency_Matrix": Weighted_Adjacency_Matrix,
               "ALPHA": ALPHA, 
               "NUMBER_OF_ATOMIC_AGENTS": NUMBER_OF_ATOMIC_AGENTS,
               "List_Atomic_Agents": List_Atomic_Agents,
               "Partition": Partition,
               "NUMBER_CONTROL_AGENTS": NUMBER_CONTROL_AGENTS,
               "Computation_Time_Total": Computation_Time_Total}

if EXPORT_DATA == True:
    with open(OUTPUT_FOLDER + "/" + Partition_ID , 'wb') as Write_File:
        pk.dump(Output_Dict, Write_File)
print(Partition)
Partition_Holder = Partition.copy()
# %% MARK: Draw plot
if DRAW_PLOTS == True:
    # Build system graph
    Network_Graph = nx.Graph()
    for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
        for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
            if i != j:
                if Weighted_Adjacency_Matrix[i,j] != 0:
                    Network_Graph.add_edge(i, j, weight=Weighted_Adjacency_Matrix[i,j])

    degree_sequence = sorted((d for n, d in Network_Graph.degree()), reverse=True)
    dmax = max(degree_sequence)
    # pos = nx.forceatlas2_layout(Network_Graph, scaling_ratio=1, linlog=True)
    # pos = nx.kamada_kawai_layout(Network_Graph, scale=10)
    # For Random_10
    # pos = nx.spring_layout(Network_Graph, scale=10, weight=None, center=[0,0], seed=4)
    # For Modular_64
    pos = nx.spring_layout(Network_Graph, scale=50, weight=None, center=[0,0])

    # Build subgraph from partitioning
    Subgraphs_Partitioning = create_empty_list(NUMBER_CONTROL_AGENTS)

    for idx in range(NUMBER_CONTROL_AGENTS):
        Subgraphs_Partitioning[idx] = nx.Graph()
        for i in range(len(Partition[idx])):
            for j in range(len(Partition[idx])):
                if i != j:
                    if Weighted_Adjacency_Matrix[i,j] != 0:
                        Subgraphs_Partitioning[idx].add_edge(i, j, weight=Weighted_Adjacency_Matrix[i,j])


    Table_Colors = mcolors.TABLEAU_COLORS

    Color_List_Base = list(Table_Colors.keys())
    Color_List = create_empty_list(NUMBER_CONTROL_AGENTS)
    external_idx = 0
    for idx in range(NUMBER_CONTROL_AGENTS):
        Color_List[idx] = Color_List_Base[external_idx]
        external_idx += 1
        if external_idx >= len(Color_List_Base):
            external_idx = 0
    Node_Degrees_List = Network_Graph.degree()

    if DRAW_PLOTS == True:
        
        Graph_Figure = plt.figure("Random graph topology", figsize=(8, 8))
        # fig = plt.figure("Random graph topology")
        Graph_Axis = plt.gca()
        # Gcc = Network_Graph.subgraph(sorted(nx.connected_components(Network_Graph), key=len, reverse=True)[0])
        # pos = nx.spring_layout(Network_Graph)
        
        # plt.autoscale(False)
        plot_colors = dict(Network_Graph.degree())

        Edges_Alpha = np.zeros((NUMBER_OF_ATOMIC_AGENTS*NUMBER_OF_ATOMIC_AGENTS,1))
        external_idx = 0
        for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
            for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
                if i != j:
                    if Weighted_Adjacency_Matrix[i,j] != 0:
                        Edges_Alpha[external_idx] = Weighted_Adjacency_Matrix[i,j]
                        external_idx += 1
        non_zero_mask = Edges_Alpha != 0
        Edges_Alpha_Filtered = Edges_Alpha[non_zero_mask]

        # nx.draw_networkx_nodes(Network_Graph, pos, ax=ax0,  node_size=50,cmap= "Blues", node_color=list(plot_colors.values()), edgecolors='cyan', linewidths=0.4)

        radius_circle = 0.5
        number_points_circumference = 100

        for idx in range(NUMBER_CONTROL_AGENTS):
            # if len(Partition[idx]) > 1:
                # Points_Convex_Hull = np.zeros((len(Partition[idx]),2))
                # for i in range(len(Partition[idx])):
                #     agent_idx_i = Partition[idx][i]
                #     Points_Convex_Hull[i,0] = pos[agent_idx_i][0]
                #     Points_Convex_Hull[i,1] = pos[agent_idx_i][1]
            Points_Convex_Hull = np.zeros((number_points_circumference*len(Partition[idx]),2))
            global_idx = 0
            for i in range(len(Partition[idx])):
                agent_idx_i = Partition[idx][i]
                Circle_Points = generate_circle_points([pos[agent_idx_i][0],pos[agent_idx_i][1]], radius_circle*1.8, number_points_circumference)
                for idx_circle in range(number_points_circumference):
                    Points_Convex_Hull[global_idx,0] = Circle_Points[0][idx_circle]
                    Points_Convex_Hull[global_idx,1] = Circle_Points[1][idx_circle]
                    global_idx+= 1 

                Points_Convex_Hull[i,0] = pos[agent_idx_i][0]
                Points_Convex_Hull[i,1] = pos[agent_idx_i][1]
                
            hull = ConvexHull(Points_Convex_Hull)
            plt.fill(Points_Convex_Hull[hull.vertices,0], Points_Convex_Hull[hull.vertices,1], color=Color_List[idx], alpha=0.3, zorder=3)


        for idx in range(NUMBER_CONTROL_AGENTS):
            for i in range(len(Partition[idx])):
                agent_idx_i = Partition[idx][i]
                
                circle = plt.Circle((pos[agent_idx_i][0], pos[agent_idx_i][1]), radius_circle, color='black', alpha=1, zorder=2)
                Graph_Axis.add_patch(circle)
                circle = plt.Circle((pos[agent_idx_i][0], pos[agent_idx_i][1]), radius_circle*0.9, color="white", alpha=1 , zorder=2)
                Graph_Axis.add_patch(circle)
                circle = plt.Circle((pos[agent_idx_i][0], pos[agent_idx_i][1]), radius_circle*0.9, color=Color_List[idx], alpha=(Node_Degrees_List[i]/(dmax)) , zorder=2)
                Graph_Axis.add_patch(circle)
                # plt.text(pos[agent_idx_i][0], pos[agent_idx_i][1], str(agent_idx_i), color="black", fontsize=12, horizontalalignment="center", verticalalignment="center")
        
        for idx in range(NUMBER_CONTROL_AGENTS):
            for i in range(len(Partition[idx])):
                agent_idx_i = Partition[idx][i]

                for j in range(NUMBER_OF_ATOMIC_AGENTS):
                    agent_idx_j = j
                    if agent_idx_j in Partition[idx]:
                        if agent_idx_i != agent_idx_j:
                            if Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j] != 0:
                                plt.plot([pos[agent_idx_i][0], pos[agent_idx_j][0]], [pos[agent_idx_i][1], pos[agent_idx_j][1]], color=Color_List[idx], zorder=1, linewidth=5*Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j], alpha=0.5)
                    else:
                        if Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j] != 0:
                            plt.plot([pos[agent_idx_i][0], pos[agent_idx_j][0]], [pos[agent_idx_i][1], pos[agent_idx_j][1]], color="black", zorder=1,linewidth=5*Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j], alpha=0.5, linestyle='dashed')

        

        # nx.draw_networkx_edges(Network_Graph, pos, ax=ax0, alpha=0.3, edge_color="green", width=Edges_Alpha_Filtered*3)
        # Graph_Figure.tight_layout(h_pad=2, w_pad=2)
        # plt.title("Random graph with 10 nodes", fontsize=18)
        # Remove axis markers
        plt.tick_params(axis='both',        # Apply to both x and y axes
                        which='both',       # Apply to both major and minor ticks
                        bottom=False,       # Turn off bottom ticks
                        top=False,          # Turn off top ticks
                        left=False,         # Turn off left ticks
                        right=False,        # Turn off right ticks
                        labelbottom=False,  # Turn off bottom labels
                        labelleft=False)    # Turn off left labels
        Graph_Figure.show()

    if (DRAW_PLOTS == True) and (EXPORT_PLOTS == True):
        Graph_Figure.savefig(OUTPUT_FOLDER + "/" + Partition_ID + ".pdf", dpi=300)
        Graph_Figure.savefig(OUTPUT_FOLDER + "/" + Partition_ID + ".jpeg", dpi=300)

    # %%

    # Build system graph
    Network_Graph = nx.Graph()
    for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
        for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
            if i != j:
                if Weighted_Adjacency_Matrix[i,j] != 0:
                    Network_Graph.add_edge(i, j, weight=Weighted_Adjacency_Matrix[i,j])

    degree_sequence = dict(Network_Graph.degree())
    Number_Rings = max(degree_sequence.values()) + 1

    from A00_Support_Funcitons import*
    Ring_Radius = 10
    Total_Circumference_Radius = Ring_Radius*Number_Rings
    Agents_Sorted = create_empty_list(Number_Rings)

    for idx in range(len(Agents_Sorted)):
        for i in range(NUMBER_OF_ATOMIC_AGENTS):
            if degree_sequence[i] == idx:
                Agents_Sorted[idx].append(i) 

    Nodes_Positions = {}
    Circumference_Radius = Total_Circumference_Radius
    # External_Idx = 0
    for idx in range(1,len(Agents_Sorted)):
        External_Idx = 0
        for i in range(len(Agents_Sorted[idx])):
            Agent_Idx = Agents_Sorted[idx][i]
            Nodes_Positions[Agent_Idx] = [Circumference_Radius*np.sin(((2*np.pi)/len(Agents_Sorted[idx]))*External_Idx),
                                        Circumference_Radius*np.cos(((2*np.pi)/len(Agents_Sorted[idx]))*External_Idx) ]
            External_Idx += 1
        
        Circumference_Radius = Circumference_Radius - Ring_Radius

    # PLOT

    Graph_Figure = plt.figure("Random graph topology", figsize=(8, 8))
    # fig = plt.figure("Random graph topology")
    Graph_Axis = plt.gca()
    # Gcc = Network_Graph.subgraph(sorted(nx.connected_components(Network_Graph), key=len, reverse=True)[0])
    # pos = nx.spring_layout(Network_Graph)

    # plt.autoscale(False)
    plot_colors = dict(Network_Graph.degree())

    Edges_Alpha = np.zeros((NUMBER_OF_ATOMIC_AGENTS*NUMBER_OF_ATOMIC_AGENTS,1))
    external_idx = 0
    for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
        for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
            if i != j:
                if Weighted_Adjacency_Matrix[i,j] != 0:
                    Edges_Alpha[external_idx] = Weighted_Adjacency_Matrix[i,j]
                    external_idx += 1
    non_zero_mask = Edges_Alpha != 0
    Edges_Alpha_Filtered = Edges_Alpha[non_zero_mask]

    # nx.draw_networkx_nodes(Network_Graph, pos, ax=ax0,  node_size=50,cmap= "Blues", node_color=list(plot_colors.values()), edgecolors='cyan', linewidths=0.4)

    radius_circle = 0.5
    number_points_circumference = 100

    Node_Degrees_List = Network_Graph.degree()
    Table_Colors = mcolors.TABLEAU_COLORS

    Color_List_Base = list(Table_Colors.keys())
    Color_List = create_empty_list(NUMBER_OF_ATOMIC_AGENTS)
    external_idx = 0
    for idx in range(NUMBER_OF_ATOMIC_AGENTS):
        Color_List[idx] = Color_List_Base[external_idx]
        external_idx += 1
        if external_idx >= len(Color_List_Base):
            external_idx = 0

    pos = Nodes_Positions
    for idx in range(NUMBER_OF_ATOMIC_AGENTS):

        agent_idx_i = idx
        
        circle = plt.Circle((pos[agent_idx_i][0], pos[agent_idx_i][1]), radius_circle, color='black', alpha=1, zorder=2)
        Graph_Axis.add_patch(circle)
        circle = plt.Circle((pos[agent_idx_i][0], pos[agent_idx_i][1]), radius_circle*0.9, color="white", alpha=1 , zorder=2)
        Graph_Axis.add_patch(circle)
        circle = plt.Circle((pos[agent_idx_i][0], pos[agent_idx_i][1]), radius_circle*0.9, color="Green", alpha= (Node_Degrees_List[idx]/(dmax)) , zorder=2)

        Graph_Axis.add_patch(circle)
        # plt.text(pos[agent_idx_i][0], pos[agent_idx_i][1], str(agent_idx_i), color="black", fontsize=12, horizontalalignment="center", verticalalignment="center")

    for idx in range(NUMBER_OF_ATOMIC_AGENTS):
        agent_idx_i = idx

        for j in range(NUMBER_OF_ATOMIC_AGENTS):
            agent_idx_j = j

            if Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j] != 0:
                plt.plot([pos[agent_idx_i][0], pos[agent_idx_j][0]], [pos[agent_idx_i][1], pos[agent_idx_j][1]], color="cyan", zorder=1,linewidth=2.5*Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j], alpha=0.5)

    # COnvex hull of the control agents, not great with this network representation

    # radius_circle = 0.5
    # number_points_circumference = 100

    # for idx in range(NUMBER_CONTROL_AGENTS):
    #     # if len(Partition[idx]) > 1:
    #         # Points_Convex_Hull = np.zeros((len(Partition[idx]),2))
    #         # for i in range(len(Partition[idx])):
    #         #     agent_idx_i = Partition[idx][i]
    #         #     Points_Convex_Hull[i,0] = pos[agent_idx_i][0]
    #         #     Points_Convex_Hull[i,1] = pos[agent_idx_i][1]
    #     Points_Convex_Hull = np.zeros((number_points_circumference*len(Partition[idx]),2))
    #     global_idx = 0
    #     for i in range(len(Partition[idx])):
    #         agent_idx_i = Partition[idx][i]
    #         Circle_Points = generate_circle_points([pos[agent_idx_i][0],pos[agent_idx_i][1]], radius_circle*1.8, number_points_circumference)
    #         for idx_circle in range(number_points_circumference):
    #             Points_Convex_Hull[global_idx,0] = Circle_Points[0][idx_circle]
    #             Points_Convex_Hull[global_idx,1] = Circle_Points[1][idx_circle]
    #             global_idx+= 1 

    #         Points_Convex_Hull[i,0] = pos[agent_idx_i][0]
    #         Points_Convex_Hull[i,1] = pos[agent_idx_i][1]
            
    #     hull = ConvexHull(Points_Convex_Hull)
    #     plt.fill(Points_Convex_Hull[hull.vertices,0], Points_Convex_Hull[hull.vertices,1], color=Color_List[idx], alpha=0.3, zorder=3)


    plt.tick_params(axis='both',        # Apply to both x and y axes
                    which='both',       # Apply to both major and minor ticks
                    bottom=False,       # Turn off bottom ticks
                    top=False,          # Turn off top ticks
                    left=False,         # Turn off left ticks
                    right=False,        # Turn off right ticks
                    labelbottom=False,  # Turn off bottom labels
                    labelleft=False)    # Turn off left labels


# %% MARK: Agent-based circular representation
if DRAW_PLOTS == True:
    # Number of sections of the external radius
    NUMBER_SECTIONS = NUMBER_CONTROL_AGENTS
    Total_Radius = max(degree_sequence.values()) + 1
    base_radus = 2
    radius_circle = 0.2
    FONT_SIZE_ATOMIC = 8
    # List of the positions of the nodes
    positions_control_agents = create_empty_list(NUMBER_OF_ATOMIC_AGENTS)
    Partition = Partition_Holder.copy()
    MAX_DEGREE = max(degree_sequence.values())
    list_agents = []
    global_idx = 0
    import math as mt


    Agents_Figure = plt.figure("Control agents", figsize=(8, 8))
    Agents_Ax = plt.gca()

    # Edges tickness
    Edges_Alpha = np.zeros((NUMBER_OF_ATOMIC_AGENTS*NUMBER_OF_ATOMIC_AGENTS,1))
    external_idx = 0
    MAX_WEIGHT = np.abs(Weighted_Adjacency_Matrix.max())
    for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
        for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
            if i != j:
                if Weighted_Adjacency_Matrix[i,j] != 0:
                    Edges_Alpha[external_idx] = np.abs(Weighted_Adjacency_Matrix[i,j])/MAX_WEIGHT 
                    external_idx += 1
    non_zero_mask = Edges_Alpha != 0
    Edges_Alpha_Filtered = Edges_Alpha[non_zero_mask]

    # Graph coloring list
    Node_Degrees_List = Network_Graph.degree()
    Table_Colors = mcolors.TABLEAU_COLORS

    Color_List_Base = list(Table_Colors.keys())
    Color_List = create_empty_list(NUMBER_OF_ATOMIC_AGENTS)
    external_idx = 0
    for idx in range(NUMBER_OF_ATOMIC_AGENTS):
        Color_List[idx] = Color_List_Base[external_idx]
        external_idx += 1
        if external_idx >= len(Color_List_Base):
            external_idx = 0
    CUMULATIVE_PERCENTAGE = 0
    for control_agent_idx in range(len(Partition)):
        PERCENTAGE_SECTOR = len(Partition[control_agent_idx])/NUMBER_OF_ATOMIC_AGENTS
        
        for degree_idx in range(1, MAX_DEGREE+1):
            number_agents_degree = 0
            for i in range(len(Partition[control_agent_idx])):
                atomic_agent_idx = Partition[control_agent_idx][i]
                if degree_sequence[atomic_agent_idx] == degree_idx:
                    # print(atomic_agent_idx)
                    # print(degree_sequence[atomic_agent_idx])
                    number_agents_degree += 1
            progressing_item = 0
            if number_agents_degree != 0:
                radius_ring = base_radus*(Total_Radius - degree_idx)
                
                for i in range(len(Partition[control_agent_idx])):
                    

                    atomic_agent_idx = Partition[control_agent_idx][i]
                    if degree_sequence[atomic_agent_idx] == degree_idx:
                        # atomic_agent_idx = Partition[control_agent_idx][i]
                        # positions_control_agents[global_idx] = [radius_ring * mt.cos(((2*mt.pi*control_agent_idx)/(NUMBER_SECTIONS))+((2*mt.pi*progressing_item)/(number_agents_degree))/(NUMBER_SECTIONS)),
                        #                         radius_ring * mt.sin(((2*mt.pi*control_agent_idx)/(NUMBER_SECTIONS))+((2*mt.pi*progressing_item)/(number_agents_degree))/(NUMBER_SECTIONS))]
                        # positions_control_agents[global_idx] = [radius_ring * mt.cos(((2*mt.pi*control_agent_idx)/(NUMBER_SECTIONS))*(1 + (progressing_item/number_agents_degree))),
                        #                         radius_ring * mt.sin(((2*mt.pi*control_agent_idx)/(NUMBER_SECTIONS))*(1 + (progressing_item/number_agents_degree)))]
                        # positions_control_agents[global_idx] = [
                        # radius_ring * mt.cos(2*mt.pi * (control_agent_idx/NUMBER_SECTIONS)+ (1/NUMBER_SECTIONS)*(progressing_item/number_agents_degree)), radius_ring * mt.sin(2*mt.pi * (control_agent_idx/NUMBER_SECTIONS) + (1/NUMBER_SECTIONS)*(progressing_item/number_agents_degree))]
                        positions_control_agents[global_idx] = [
                        radius_ring * mt.cos(2*mt.pi * (CUMULATIVE_PERCENTAGE)+ (2*mt.pi*PERCENTAGE_SECTOR)*(progressing_item/number_agents_degree)), radius_ring * mt.sin(2*mt.pi * (CUMULATIVE_PERCENTAGE) + (2*mt.pi*PERCENTAGE_SECTOR)*(progressing_item/number_agents_degree))]
                        list_agents.append(atomic_agent_idx)
                        global_idx += 1
                        progressing_item += 1
        CUMULATIVE_PERCENTAGE = CUMULATIVE_PERCENTAGE + PERCENTAGE_SECTOR


    # plt.xlim(-5,5)
    # plt.ylim(-5,5)
    # FONT_SIZE_ATOMIC = 10
    # list_agents = Partition[0].copy()
    # for i in range(1, len(Partition)):
    #     for j in range(len(Partition[i])):
    #         list_agents.append(Partition[i][j])

    for idx_figure in range(len(positions_control_agents)):
        if positions_control_agents[idx_figure] != []:
            Control_Agent_Idx = 0    
            for i in range(len(Partition)):
                if list_agents[idx_figure] in Partition[i]:
                    Control_Agent_Idx = i
            # agent_i_position_idx = list_agents.index(idx_figure)  
            circle = plt.Circle((positions_control_agents[idx_figure][0], positions_control_agents[idx_figure][1]), radius_circle, color='black', alpha=1, zorder=2)
            Agents_Ax.add_patch(circle)
            circle = plt.Circle((positions_control_agents[idx_figure][0], positions_control_agents[idx_figure][1]), radius_circle*0.9, color="white", alpha=1 , zorder=2)
            Agents_Ax.add_patch(circle)
            circle = plt.Circle((positions_control_agents[idx_figure][0], positions_control_agents[idx_figure][1]), radius_circle*0.9, color=Color_List[Control_Agent_Idx], alpha=(Node_Degrees_List[list_agents[idx_figure]]/(dmax)) , zorder=2)
            plt.text(positions_control_agents[idx_figure][0], positions_control_agents[idx_figure][1],  f'$\mathcal{{A}}_{{{list_agents[idx_figure]}}}$', color="black", fontsize=FONT_SIZE_ATOMIC, horizontalalignment="center", verticalalignment="center", zorder=4)
            Agents_Ax.add_patch(circle)

    for idx in range(NUMBER_CONTROL_AGENTS):
            for i in range(len(Partition[idx])):
                agent_idx_i = Partition[idx][i]

                for j in range(NUMBER_OF_ATOMIC_AGENTS):
                    agent_idx_j = j
                    if agent_idx_j in Partition[idx]:
                        if agent_idx_i != agent_idx_j:
                            if Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j] != 0:
                                agent_i_position_idx = list_agents.index(agent_idx_i)
                                agent_j_position_idx = list_agents.index(agent_idx_j)
                                plt.plot([positions_control_agents[agent_i_position_idx][0], positions_control_agents[agent_j_position_idx][0]], [positions_control_agents[agent_i_position_idx][1], positions_control_agents[agent_j_position_idx][1]], color=Color_List[idx], zorder=1, linewidth=5*Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j], alpha=0.5)
                    else:
                        if Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j] != 0:
                            agent_i_position_idx = list_agents.index(agent_idx_i)
                            agent_j_position_idx = list_agents.index(agent_idx_j)
                            plt.plot([positions_control_agents[agent_i_position_idx][0], positions_control_agents[agent_j_position_idx][0]], [positions_control_agents[agent_i_position_idx][1], positions_control_agents[agent_j_position_idx][1]], color="black", zorder=1, linewidth=5*Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j], alpha=0.5, linestyle="dashed")

    for idx in range(NUMBER_CONTROL_AGENTS):
        Points_Convex_Hull = np.zeros((number_points_circumference*len(Partition[idx]),2))
        global_idx = 0
        for i in range(len(Partition[idx])):
            agent_idx_i = Partition[idx][i]
            agent_i_position_idx = list_agents.index(agent_idx_i)
            Circle_Points = generate_circle_points([positions_control_agents[agent_i_position_idx][0],positions_control_agents[agent_i_position_idx][1]], radius_circle*1.8, number_points_circumference)
            for idx_circle in range(number_points_circumference):
                Points_Convex_Hull[global_idx,0] = Circle_Points[0][idx_circle]
                Points_Convex_Hull[global_idx,1] = Circle_Points[1][idx_circle]
                global_idx+= 1 

            Points_Convex_Hull[i,0] = positions_control_agents[agent_i_position_idx][0]
            Points_Convex_Hull[i,1] = positions_control_agents[agent_i_position_idx][1]
            
        hull = ConvexHull(Points_Convex_Hull)
        plt.fill(Points_Convex_Hull[hull.vertices,0], Points_Convex_Hull[hull.vertices,1], color=Color_List[idx], alpha=0.3, zorder=3)



    plt.tick_params(axis='both',        # Apply to both x and y axes
                    which='both',       # Apply to both major and minor ticks
                    bottom=False,       # Turn off bottom ticks
                    top=False,          # Turn off top ticks
                    left=False,         # Turn off left ticks
                    right=False,        # Turn off right ticks
                    labelbottom=False,  # Turn off bottom labels
                    labelleft=False)    # Turn off left labels
    plt.show()


    if (DRAW_PLOTS == True) and (EXPORT_PLOTS == True):
        Graph_Figure.savefig(OUTPUT_FOLDER + "/" + Partition_ID + "Graph_Figure.pdf", dpi=300)
        Graph_Figure.savefig(OUTPUT_FOLDER + "/" + Partition_ID + "Graph_Figure.jpeg", dpi=300)
        Agents_Figure.savefig(OUTPUT_FOLDER + "/" + Partition_ID + "Agents_Figure.pdf", dpi=300)
        Agents_Figure.savefig(OUTPUT_FOLDER + "/" + Partition_ID + "Agents_Figure.jpeg", dpi=300)
