# from IPython import get_ipython
# get_ipython().run_line_magic('reset', '-f')

# Optimization toolbox
import cvxpy as cp
import gurobipy as gp
# For matrix operations
import numpy as np
# For file storage
import pickle as pk
# For time tracking
import time as tm
import matplotlib.pyplot as plt
from matplotlib import rc
# Enable LaTeX rendering
rc('text', usetex=True)
rc('font', family='serif')
rc('font', serif='Times New Roman')  # This line might need adjustment based on LaTeX distribution

# Support funciton
from A00_Support_Funcitons import*


# Options
PLOT_FIGURES = True
EXPORT_FIGURES = False
EXPORT_SIMULATION = True
SIMULATION_FOLDER = "A04_Output_Data_DMPC/"

import os
if not os.path.exists(SIMULATION_FOLDER):
    os.mkdir(SIMULATION_FOLDER)

# %%
# MARK: Load system and parameters
NETWORK_FOLDER = "Input_Data/"
NETWORK_NAME = "Modular_64"
with open(NETWORK_FOLDER + NETWORK_NAME, 'rb') as file:
    Network_Dict = pk.load(file)

NUMBER_ATOMIC_AGENTS = Network_Dict['NUMBER_OF_ATOMIC_AGENTS']

# Algorithmic Partitioning 
# PARTITIONING_FOLDER = "A03_Output_Data_Algorithmic_Partitioning_Mixed/"
# PARTITIONING_NAME = "Modular_64_Partition_01_ALPHA_25"
# PARTITIONING_NAME = "Modular_64_Partition_02_ALPHA_1"
# PARTITIONING_NAME = "Modular_64_Partition_03_ALPHA_1e-06"

# Optimization-based Partitioning 
# PARTITIONING_FOLDER = "A09_Output_Data/"
# PARTITIONING_NAME = "Modular_64_Partitioning_04_ALPHA_1"
# PARTITIONING_NAME = "Modular_64_Partitioning_03_ALPHA_100"
# PARTITIONING_NAME = "Modular_64_Partitioning_02_ALPHA_10000"
# PARTITIONING_NAME = "Modular_64_Partitioning_01_ALPHA_1000000"

# Other Partitioning 
PARTITIONING_FOLDER = "A09_Output_Data/"
PARTITIONING_NAME = "Modular_64_Test_1"

Simumation_ID = PARTITIONING_NAME 
with open(PARTITIONING_FOLDER + PARTITIONING_NAME, 'rb') as file:
    Partitioning_Dict = pk.load(file)
# with open('Output_Data/Random_50_0.004', 'rb') as file:
#     Network_Dict = pk.load(file)

A = Network_Dict['A']
B = Network_Dict['B']
Weighted_Adjacency_Matrix = Network_Dict['Weighted_Adjacency_Matrix']
X_INITIAL = Network_Dict["X_INITIAL"]

# List_Atomic_Agents = Partitioning_Dict['List_Atomic_Agents']

# For algorithmic partitioning
# Partition = Partitioning_Dict["Partition"]
# NUMBER_CONTROL_AGENTS = Partitioning_Dict["NUMBER_CONTROL_AGENTS"]
# For optimization-based partitioning
Partition = Partitioning_Dict["Control_Agents_List"]
NUMBER_CONTROL_AGENTS = Partitioning_Dict["Number_Of_Control_Agents"]
# NUMBER_CONTROL_AGENTS = 1
# Partition = create_empty_list(1)
# for i in range(0,NUMBER_ATOMIC_AGENTS):
#     Partition[0].append(i)

PARAMETERS_DICT = simulation_parameters(NUMBER_ATOMIC_AGENTS)

SIMULATION_HORIZON = PARAMETERS_DICT['SIMULATION_HORIZON'] 
PREDICTION_HORIZON = PARAMETERS_DICT['PREDICTION_HORIZON']
NUMBER_ATOMIC_AGENTS = PARAMETERS_DICT['NUMBER_ATOMIC_AGENTS']
MAX_u = PARAMETERS_DICT['MAX_u']
MIN_u = PARAMETERS_DICT['MIN_u']
MAX_x = PARAMETERS_DICT['MAX_x']
MIN_x = PARAMETERS_DICT['MIN_x']
Q = PARAMETERS_DICT['Q']
R = PARAMETERS_DICT['R']

# %% 
# MARK:  Augmented control agents

Augmented_Control_Agents = create_empty_list(NUMBER_CONTROL_AGENTS)

for l in range(0,NUMBER_CONTROL_AGENTS):
    Augmented_Control_Agents[l] = Partition[l].copy()
    Size_Control_Agent = len(Partition[l])
    
    for i in range(Size_Control_Agent):
        Agent_Index = Partition[l][i]
        for j in range(NUMBER_ATOMIC_AGENTS):
            if (A[Agent_Index,j] != 0) and (j not in Partition[l]):
                Augmented_Control_Agents[l].append(j)
        
    Augmented_Control_Agents[l].sort()

# %% 
# MARK: Control agents matrices

Control_Agents_Matrices = create_empty_list(NUMBER_CONTROL_AGENTS)

for l in range(0,NUMBER_CONTROL_AGENTS):
    Size_Augmented_Control_Agent = len(Augmented_Control_Agents[l])
    A_Augmented = np.zeros((Size_Augmented_Control_Agent,Size_Augmented_Control_Agent))
    B_Augmented = np.zeros((Size_Augmented_Control_Agent,Size_Augmented_Control_Agent))
    Q_Augmented = np.zeros((Size_Augmented_Control_Agent,Size_Augmented_Control_Agent))
    R_Augmented = np.zeros((Size_Augmented_Control_Agent,Size_Augmented_Control_Agent))
    for i in range(0, Size_Augmented_Control_Agent):
        Agent_Index_1 = Augmented_Control_Agents[l][i]
        for j in range(0, Size_Augmented_Control_Agent):
            Agent_Index_2 = Augmented_Control_Agents[l][j]
            A_Augmented[i,j] = A[Agent_Index_1, Agent_Index_2]
            B_Augmented[i,j] = B[Agent_Index_1, Agent_Index_2]
            Q_Augmented[i,j] = Q[Agent_Index_1, Agent_Index_2]
            R_Augmented[i,j] = R[Agent_Index_1, Agent_Index_2]
    
    Control_Agents_Matrices[l] = {'A_Augmented': A_Augmented,
                                  'B_Augmented': B_Augmented,
                                  'Q_Augmented': Q_Augmented,
                                  'R_Augmented': R_Augmented}
    
# %%
# MARK: Control agents optimization matrices
Control_Agents_Optimization_Matrices = create_empty_list(NUMBER_CONTROL_AGENTS)

for l in range(0,NUMBER_CONTROL_AGENTS):
    Size_Augmented_Control_Agent = len(Augmented_Control_Agents[l])

    M_Optimization = np.zeros((Size_Augmented_Control_Agent*PREDICTION_HORIZON,Size_Augmented_Control_Agent), dtype=np.float64)
    C_Optimization = np.zeros((Size_Augmented_Control_Agent*PREDICTION_HORIZON,Size_Augmented_Control_Agent*PREDICTION_HORIZON), dtype=np.float64)

    A_Augmented = Control_Agents_Matrices[l]['A_Augmented']
    B_Augmented = Control_Agents_Matrices[l]['B_Augmented']

    Row_Index = 0
    for i in range(1,PREDICTION_HORIZON+1):

        M_Optimization[Row_Index:Row_Index+Size_Augmented_Control_Agent,:] = np.linalg.matrix_power(A_Augmented,i)
        
        Col_Index = 0
        for j in reversed(range(0,i)):
            C_Optimization[Row_Index:Row_Index+Size_Augmented_Control_Agent,Col_Index:Col_Index+Size_Augmented_Control_Agent] = np.linalg.matrix_power(A_Augmented,j) @ B_Augmented
            Col_Index = Col_Index + Size_Augmented_Control_Agent

        Row_Index = Row_Index + Size_Augmented_Control_Agent
    
    
    Q_Augmented = Control_Agents_Matrices[l]['Q_Augmented']
    R_Augmented = Control_Agents_Matrices[l]['R_Augmented']

    Q_Optimization = np.zeros((Size_Augmented_Control_Agent*PREDICTION_HORIZON,Size_Augmented_Control_Agent*PREDICTION_HORIZON), dtype=np.float64)
    R_Optimization = np.zeros((Size_Augmented_Control_Agent*PREDICTION_HORIZON,Size_Augmented_Control_Agent*PREDICTION_HORIZON), dtype=np.float64)

    Pos_Index = 0
    for i in range(0,PREDICTION_HORIZON):
        Q_Optimization[Pos_Index:Pos_Index+Size_Augmented_Control_Agent,Pos_Index:Pos_Index+Size_Augmented_Control_Agent] = Q_Augmented
        R_Optimization[Pos_Index:Pos_Index+Size_Augmented_Control_Agent,Pos_Index:Pos_Index+Size_Augmented_Control_Agent] = R_Augmented
        Pos_Index = Pos_Index + Size_Augmented_Control_Agent
    
    Control_Agents_Optimization_Matrices[l] = {'M_Optimization': M_Optimization,
                                               'C_Optimization': C_Optimization,
                                               'Q_Optimization': Q_Optimization,
                                               'R_Optimization': R_Optimization}
    
# %% MARK: Reference signal

#  Generate sinusoidal reference
Reference_Signal = np.zeros((NUMBER_ATOMIC_AGENTS,SIMULATION_HORIZON+PREDICTION_HORIZON))# Sinusoid 
Number_Cycles = 2.5 # for simulation horizon 50
# Number_Cycles = 10 # for simulation horizon 200 
# Number_Cycles = 5 # for simulation horizon 100 
# Number_Cycles = 0.5 # for simulation horizon 10
# Frequency = Number_Cycles*2*np.pi/(SIMULATION_HORIZON)
# Fixed frequency
Frequency = 2*np.pi/(15)


for i in range(0,len(Reference_Signal)):
    for k in range(0,len(Reference_Signal[i])):
        Reference_Signal[i,k] = np.sin(Frequency * k)

# %%
# MARK: Control simulation
Control_Agents = Partition.copy()

# To activate gurobi
gurobi_model_activate = gp.Model("New_Model")
gurobi_model_activate.dispose()

# Pre-allocation of variables
x_evolution = np.zeros((NUMBER_ATOMIC_AGENTS,SIMULATION_HORIZON+1))
u_evolution = np.zeros((NUMBER_ATOMIC_AGENTS,SIMULATION_HORIZON+1))
Cost_Function = np.zeros((SIMULATION_HORIZON))
# Initial state
# x_initial = np.zeros((NUMBER_ATOMIC_AGENTS,1))
# x_evolution[:,0] = np.ones(NUMBER_ATOMIC_AGENTS)
# x_evolution[:,[0]] = X_INITIAL




max_number_of_iterations = 100
residual_norm = np.zeros((NUMBER_CONTROL_AGENTS))
residual_evolution = np.zeros((max_number_of_iterations))

Total_Core_Seconds = 0
Core_Seconds = create_empty_list(SIMULATION_HORIZON)
for i in range(0, SIMULATION_HORIZON):
    Core_Seconds[i] = np.zeros((NUMBER_CONTROL_AGENTS,max_number_of_iterations))

rho = 0.01################################## BEST RESULTS
# epsilon_convergence = 0.001
################################## COMPROMISE FOR SPEED
epsilon_convergence = 0.01

u_local = create_empty_list(NUMBER_CONTROL_AGENTS)
x_local = create_empty_list(NUMBER_CONTROL_AGENTS)

u_avarage_local = np.zeros((NUMBER_ATOMIC_AGENTS, PREDICTION_HORIZON))
x_avarage_local = np.zeros((NUMBER_ATOMIC_AGENTS, PREDICTION_HORIZON))

consensus_variable = create_empty_list(NUMBER_CONTROL_AGENTS)
avarage_variable = create_empty_list(NUMBER_CONTROL_AGENTS)
avarage_variable_old = create_empty_list(NUMBER_CONTROL_AGENTS)
gamma = create_empty_list(NUMBER_CONTROL_AGENTS)

for l in range(0,NUMBER_CONTROL_AGENTS):
    number_of_agents_l = len(Augmented_Control_Agents[l])
    avarage_variable[l] = np.zeros((2*number_of_agents_l*PREDICTION_HORIZON,1))
    consensus_variable[l] = np.zeros((2*number_of_agents_l*PREDICTION_HORIZON,1))
    gamma[l] = np.zeros((2*number_of_agents_l*PREDICTION_HORIZON,1))


# Simulation dictionary
Simulation_Data = {"Network_Dict": Network_Dict,
                   "Partitioning_Dict": Partitioning_Dict,
                   "Reference_Signal": Reference_Signal,
                   "PARAMETERS_DICT": PARAMETERS_DICT,
                   "x_evolution": x_evolution,
                   "u_evolution": u_evolution,
                   "residual_evolution": residual_evolution,
                   "max_number_of_iterations": max_number_of_iterations,
                   "epsilon_convergence": epsilon_convergence,
                   "rho": rho,
                   "Cost_Function": Cost_Function}


Simulation_Timer_Start = tm.time()

# Simulation 
for k in range(0,SIMULATION_HORIZON):
    
    print("TIME STEP ", k)

    number_of_iterations = 0
    
    norm_codition_1 = False
    norm_codition_2 = False
    Core_Seconds_Control_Agent = np.zeros(NUMBER_CONTROL_AGENTS)
    # while (number_of_iterations < max_number_of_iterations) and (norm_codition_1 == False) and (norm_codition_2 == False):
    while (number_of_iterations < max_number_of_iterations) and (norm_codition_1 == False):
        print("TIME STEP ", k, " - Local iteration ", number_of_iterations + 1)
        
        Problem_Relaxation = False
        
        for l in range(0,len(Augmented_Control_Agents)):
            Start_Time_Holder = tm.time()
            number_of_agents_l = len(Augmented_Control_Agents[l])
            
            #  Avarage variable store for norm evaluation
            avarage_variable_old[l] = avarage_variable[l].copy()

            # Local optimization matrices
            M_Optimization = Control_Agents_Optimization_Matrices[l]['M_Optimization']
            C_Optimization = Control_Agents_Optimization_Matrices[l]['C_Optimization']
            Q_Optimization = Control_Agents_Optimization_Matrices[l]['Q_Optimization'] 
            R_Optimization = Control_Agents_Optimization_Matrices[l]['R_Optimization'] 
            
            #  Model initialization
            gurobi_model = gp.Model("MPC_Local")

            # Bounds
            Lower_Bounds_u = np.full((number_of_agents_l*PREDICTION_HORIZON, 1),MIN_u)
            Upper_Bounds_u = np.full((number_of_agents_l*PREDICTION_HORIZON, 1),MAX_u)

            #  State at k and reference for the agent
            x_k_l = np.zeros((number_of_agents_l,1))
            reference_k_l = np.zeros((number_of_agents_l*PREDICTION_HORIZON,1))
            raw_idx = 0
            for i in range(0,NUMBER_ATOMIC_AGENTS):
                if i in Augmented_Control_Agents[l]:
                    x_k_l[raw_idx,0] = x_evolution[i,k]
                    # reference_k_l[raw_idx,:] = Reference_Signal[i,k:k+PREDICTION_HORIZON+1]
                    raw_idx += 1
            raw_idx = 0
            for j in range(1,PREDICTION_HORIZON+1):
                for i in range(0,NUMBER_ATOMIC_AGENTS):
                    if i in Augmented_Control_Agents[l]:
                        reference_k_l[raw_idx] = Reference_Signal[i,k+j]
                        raw_idx += 1


            # Variables
            u_tilde_k = gurobi_model.addMVar((number_of_agents_l*PREDICTION_HORIZON, 1), lb = Lower_Bounds_u, ub = Upper_Bounds_u)
            x_tilde_k = M_Optimization @ x_k_l + C_Optimization @ u_tilde_k

            # Constraints 
            for i in range(0,number_of_agents_l*PREDICTION_HORIZON):       
                # print("x_tilde_k[i] for i = ", i, ": ", x_tilde_k[i])
                gurobi_model.addConstr(x_tilde_k[i] <= MAX_x)
                gurobi_model.addConstr(x_tilde_k[i] >= MIN_x)

            # Cost function matrices
            # print(C_Optimization.shape)
            # print(Q_Optimization.shape)
            # print(R_Optimization.shape)
            H_Optimization = (np.transpose(C_Optimization) @ Q_Optimization @ C_Optimization) + R_Optimization
            F_Optimization = np.transpose(M_Optimization) @ Q_Optimization @ C_Optimization

            # Cost function 
            # Stage_Cost = (u_tilde_k.T @ H_Optimization @ u_tilde_k) + 2*(np.transpose(x_k_l) @ np.transpose(M_Optimization) - reference_k_l.T) @ Q_Optimization @ C_Optimization @ u_tilde_k  
            # Stage cost including ADMM
            gamma_l = gamma[l].reshape(len(gamma[l]),).copy()
            consensus_variable_l = consensus_variable[l].reshape(len(consensus_variable[l]),).copy()
            avarage_variable_l = avarage_variable[l].reshape(len(avarage_variable[l]),).copy()
            Stage_Cost = (u_tilde_k.T @ H_Optimization @ u_tilde_k) + 2*(np.transpose(x_k_l) @ np.transpose(M_Optimization) - reference_k_l.T) @ Q_Optimization @ C_Optimization @ u_tilde_k + np.transpose(gamma_l)@(consensus_variable_l - avarage_variable_l) + (rho/2)*np.transpose(consensus_variable_l - avarage_variable_l)@(consensus_variable_l - avarage_variable_l)
            # residual_norm[l] = np.linalg.norm(consensus_variable_l - avarage_variable_l, ord=2)
            # residual_norm[l] = np.transpose(gamma_l)@(consensus_variable_l - avarage_variable_l)
            # Optimization variables
            # u_tilde_k = np.zeros((number_of_agents_l, PREDICTION_HORIZON), dtype=object)
            # delta_k = np.zeros((number_of_agents_l, PREDICTION_HORIZON), dtype=object)
            # zeta_k = np.zeros((number_of_agents_l, PREDICTION_HORIZON), dtype=object)

            # for j in range(0, PREDICTION_HORIZON):
            #     for i in range(0,number_of_agents_l):
            #         u_tilde_k[i,j] = gurobi_model.addVar(m_u,M_u)
                # for i in range(0,number_of_agents_l):
                #     delta_k[i,j] = gurobi_model.addVar(vtype=gp.GRB.BINARY) 
                # for i in range(0,number_of_agents_l):
                #     zeta_k[i,j] = gurobi_model.addVar()

            

            #  Local Cost function
            # total_cost = cost_function_local(x_k_l, u_tilde_k, delta_k, zeta_k, reference_k_l, optimization_matrices[l], number_of_agents_l, avarage_variable[l].reshape(len(avarage_variable[l]),), gamma[l].reshape(len(gamma[l]),), l, rho, Augmented_State[l])
            
            # Local Constraints
            # gurobi_model = constraint_function_local(x_k_l, u_tilde_k, delta_k, zeta_k, gurobi_model, optimization_matrices[l], number_of_agents_l)

            # Options
            gurobi_model.setObjective(Stage_Cost, gp.GRB.MINIMIZE) 
            gurobi_model.Params.LogToConsole = 0
            # gurobi_model.Params.FeasibilityTol = 1e-2
            
            # Local Optimization
            gurobi_model.optimize()

            # ceche infeasibility
            if gurobi_model.status == gp.GRB.INFEASIBLE:
                # print("UNFEASIBLE MODEL!!!!")
                # gurobi_model.Params.LogToConsole = 1
                Problem_Relaxation = True
                vars = gurobi_model.getVars()
                # ubpen = [1.0]*gurobi_model.numVars
                gurobi_model.feasRelax(2, False, vars, 0.01*Lower_Bounds_u, 0.01*Upper_Bounds_u, None, None)
                gurobi_model.optimize()

            # Results
            # Local variables
            u_local_k = np.zeros((number_of_agents_l, PREDICTION_HORIZON))
            x_local_k = np.zeros((number_of_agents_l, PREDICTION_HORIZON+1))
            d_local_k = np.zeros((number_of_agents_l, PREDICTION_HORIZON))
            z_local_k = np.zeros((number_of_agents_l, PREDICTION_HORIZON))

            x_local_k[:,0] = x_k_l.reshape(len(x_k_l),)

            var_idx = 0
            for j in range(0, PREDICTION_HORIZON):
                for i in range(0,number_of_agents_l):
                    u_local_k[i,j] = gurobi_model.getVars()[var_idx].x
                    var_idx += 1
                # for i in range(0,number_of_agents_l):
                #     d_local_k[i,j] = gurobi_model.getVars()[var_idx].x
                #     var_idx += 1
                # for i in range(0,number_of_agents_l):
                #     z_local_k[i,j] = gurobi_model.getVars()[var_idx].x
                #     var_idx += 1

            # Local simulation 
            
            A_l_augmented = Control_Agents_Matrices[l]['A_Augmented']
            B_l_u = Control_Agents_Matrices[l]['B_Augmented']
            # B_l_z = optimization_matrices[l][2]

            for k_local in range(0,PREDICTION_HORIZON):
                # x_local_k[:,k_local+1] = A_l_augmented @ x_local_k[:,k_local] + B_l_u @ u_local_k[:,k_local] + B_l_z @ z_local_k[:,k_local]
                x_local_k[:,k_local+1] = A_l_augmented @ x_local_k[:,k_local] + B_l_u @ u_local_k[:,k_local]

            u_local[l] = u_local_k
            x_local[l] = x_local_k[:,1:]

            # Build the consensus variable
            y_consensus = np.zeros((0,0))
            
            for i in range(0,number_of_agents_l):
                y_consensus = np.concatenate((y_consensus,u_local_k[i,:]), axis=None)
            for i in range(0,number_of_agents_l):
                y_consensus = np.concatenate((y_consensus,x_local_k[i,1:]), axis=None)

            consensus_variable[l] = y_consensus.T
            
            #  Global Solution
            for i in range(0,NUMBER_ATOMIC_AGENTS):
                for j in range(0,len(Control_Agents[l])):
                    if i == Control_Agents[l][j]:
                        Agent_Index = Augmented_Control_Agents[l].index(i)
                        # print("Agent: ", i)
                        # print("Selection: ", Augmented_Control_Agents[l][Agent_Index])
                        u_evolution[i,k] = u_local_k[Agent_Index,0]
            
            Total_Time_Holder = tm.time() - Start_Time_Holder
            Core_Seconds[k][l,number_of_iterations] = Core_Seconds[k][l,number_of_iterations] + Total_Time_Holder
                     

        # Consensus Structure

        # Computation of the local avarage variable
        # Computation of the avarage variable for each atomic agent
        u_avarage_local = np.zeros((NUMBER_ATOMIC_AGENTS, PREDICTION_HORIZON))
        x_avarage_local = np.zeros((NUMBER_ATOMIC_AGENTS, PREDICTION_HORIZON))
        total_copies_vector = np.zeros((NUMBER_CONTROL_AGENTS))
        for l in range(0,NUMBER_ATOMIC_AGENTS):
            
            
            total_copies = 0

            for i in range(0,NUMBER_CONTROL_AGENTS):
                Start_Time_Holder = tm.time()
                u_local_k = u_local[i]
                x_local_k = x_local[i]
                for j in range(0,len(Augmented_Control_Agents[i])):
                    if l == Augmented_Control_Agents[i][j]:
                        u_avarage_local[l,:] = u_avarage_local[l,:] + u_local_k[j,:]
                        x_avarage_local[l,:] = x_avarage_local[l,:] + x_local_k[j,:]
                        total_copies += 1
                Total_Time_Holder = tm.time() - Start_Time_Holder
                Core_Seconds[k][i,number_of_iterations] = Core_Seconds[k][i,number_of_iterations] + Total_Time_Holder        
            u_avarage_local[l,:] = u_avarage_local[l,:]/total_copies
            x_avarage_local[l,:] = x_avarage_local[l,:]/total_copies
            # total_copies_vector[l] = total_copies
            

        # Computation of the global avarage variable for each agent
        # Computation of the avarage variable for each control agent
        for l in range(0,NUMBER_CONTROL_AGENTS):
            Start_Time_Holder = tm.time()

            number_of_agents_l = len(Augmented_Control_Agents[l])

            y_avarage = np.zeros((0,0))

            for i in range(0,number_of_agents_l):
                variable_index = Augmented_Control_Agents[l][i]
                y_avarage = np.concatenate((y_avarage,u_avarage_local[variable_index,:]), axis=None)
            for i in range(0,number_of_agents_l):
                variable_index = Augmented_Control_Agents[l][i]
                y_avarage = np.concatenate((y_avarage,x_avarage_local[variable_index,:]), axis=None)
            
            # avarage_variable[l] = y_avarage.T
            # Avarage variable update
            # total_copies = total_copies_vector[l]
            total_neighbors = 1
            if len(Augmented_Control_Agents[l]) - len(Control_Agents[l]) > 0:
                total_neighbors = len(Augmented_Control_Agents[l]) - len(Control_Agents[l])
            gamma_l = gamma[l].reshape(len(gamma[l]),).copy()
            # avarage_variable[l] = y_avarage.T + ((1/(rho*total_copies))*gamma_l )
            avarage_variable[l] = y_avarage.T + ((1/(rho*total_neighbors))*gamma_l )
            # avarage_variable[l] = y_avarage.T + ((1/rho)*gamma_l) 
            Total_Time_Holder = tm.time() - Start_Time_Holder
            Core_Seconds[k][l,number_of_iterations] = Core_Seconds[k][l,number_of_iterations] + Total_Time_Holder
            


        #  Gamma Update
        for l in range(0,len(Augmented_Control_Agents)):
            Start_Time_Holder = tm.time()
            y_consensus = consensus_variable[l].reshape(len(consensus_variable[l]),).copy()
            y_avarage = avarage_variable[l].reshape(len(avarage_variable[l]),).copy()
            gamma_l = gamma[l].reshape(len(gamma[l]),).copy()
            gamma[l] = gamma_l + rho*(y_consensus - y_avarage)
            Total_Time_Holder = tm.time() - Start_Time_Holder
            Core_Seconds[k][l,number_of_iterations] = Core_Seconds[k][l,number_of_iterations] + Total_Time_Holder


        max_norm_error_1 = float('-inf')
        max_norm_error_2 = float('-inf')

        for l in range(0,NUMBER_CONTROL_AGENTS):
            Start_Time_Holder = tm.time()

            y_consensus = consensus_variable[l].copy()
            y_avarage = avarage_variable[l].copy() 
            y_avarage_old = avarage_variable_old[l].copy()

            gamma_l = gamma[l].copy()
            alpha_l = np.linalg.norm(y_consensus - y_avarage, ord = 2)
            beta_l = rho*np.linalg.norm(y_avarage - y_avarage_old, ord = 2)

            epsilon_1 = 0.001*max(np.linalg.norm(y_consensus, ord = 2), np.linalg.norm(y_avarage, ord = 2))
            epsilon_2 = 0.001*np.linalg.norm(gamma_l, ord = 2)
            
            if alpha_l > epsilon_1:
                norm_codition_2 = False
            if alpha_l - epsilon_1> max_norm_error_1:
                max_norm_error_1 = alpha_l - epsilon_1
            
            if beta_l > epsilon_2:
                norm_codition_2 = False
            if beta_l - epsilon_2> max_norm_error_2:
                max_norm_error_2 = beta_l - epsilon_2
            Total_Time_Holder = tm.time() - Start_Time_Holder
            Core_Seconds[k][l,number_of_iterations] = Core_Seconds[k][l,number_of_iterations] + Total_Time_Holder

        for l in range(0,NUMBER_CONTROL_AGENTS):
            gamma_l = gamma[l].reshape(len(gamma[l]),).copy()
            consensus_variable_l = consensus_variable[l].reshape(len(consensus_variable[l]),).copy()
            avarage_variable_l = avarage_variable[l].reshape(len(avarage_variable[l]),).copy()
            residual_norm[l] = np.linalg.norm(consensus_variable_l - avarage_variable_l, ord=2)
        
        
        norm_codition_1 = True
        norm_codition_2 = False
        for l in range(0,NUMBER_CONTROL_AGENTS):
            if residual_norm[l] > epsilon_convergence:
                norm_codition_1 = False


        residual_evolution[number_of_iterations] = residual_evolution[number_of_iterations] + np.max(residual_norm)

        #  Update number of iteration
        number_of_iterations += 1

        #  Print exit conditions
        if norm_codition_1 == True:
            print("---------------> Exit for norm condition 1. Max norm:", np.max(residual_norm))
            if Problem_Relaxation == True:
                print("USING PROBLEM RELAXATION")
        elif norm_codition_1 == False:
            print("---------------> Reiteration. Max norm:", np.max(residual_norm))
            if Problem_Relaxation == True:
                print("USING PROBLEM RELAXATION")
        elif number_of_iterations >= max_number_of_iterations:
            print("---------------> Exit for max iterations")
            if Problem_Relaxation == True:
                print("USING PROBLEM RELAXATION")
        
        print("Max deviation condition 1: ", max_norm_error_1)
        print("Max deviation condition 2: ", max_norm_error_2)


    # State update using PWA dynamics and global solution
    print("Mean evolution error norm = ", (np.linalg.norm(Reference_Signal[:,k] - x_evolution[:,k]))/NUMBER_ATOMIC_AGENTS)
    Simulation_Timer_End = tm.time() - Simulation_Timer_Start
    print("Enlapsed time:", round(Simulation_Timer_End,2), "[s]")
    print("Expected time to finish:", round(((Simulation_Timer_End*SIMULATION_HORIZON)/(k+1)) - Simulation_Timer_End, 2), "[s]")

    # x_evolution[:,k+1] =  state_update_equation_PWA(x_evolution[:,k], u_evolution[:,k])
    x_evolution[:,k+1] =  state_update_equation_linear(A, B, x_evolution[:,k], u_evolution[:,k])

    # Cost function
    if k > 0:
        Cost_Function[k] = Cost_Function[k-1] + ((x_evolution[:,k+1] - Reference_Signal[:,k+1]).T @ Q @ (x_evolution[:,k+1] - Reference_Signal[:,k+1])) + (u_evolution[:,k].T @ R @ u_evolution[:,k])
    else:
        Cost_Function[k] = ((x_evolution[:,k+1] - Reference_Signal[:,k+1]).T @ Q @ (x_evolution[:,k+1] - Reference_Signal[:,k+1])) + (u_evolution[:,k].T @ R @ u_evolution[:,k])

residual_evolution = residual_evolution/SIMULATION_HORIZON
# %% MARK: Evaluation Metrics
print("Number of control agents: ", NUMBER_CONTROL_AGENTS)
print("Total cost: ", Cost_Function[SIMULATION_HORIZON-1])

Total_Core_Seconds = 0
Total_Computation_Time = 0
Total_Number_ADMM_Iterations = 0
Max_Time_Iteration = 0
for k in range(0,SIMULATION_HORIZON):
    for i in range(0, max_number_of_iterations):
        Max_Time_Iteration = np.max(Core_Seconds[k][:,i],0)
        Total_Core_Seconds = Total_Core_Seconds + NUMBER_CONTROL_AGENTS*Max_Time_Iteration
        Total_Computation_Time = Total_Computation_Time + Max_Time_Iteration
        if Max_Time_Iteration > 0:
            Total_Number_ADMM_Iterations += 1
        # print(i)
        # print(Core_Seconds[k][:,i])
        # print(Max_Time_Iteration)

print("Total computation time: ", Total_Computation_Time)
print("Total core seconds: ", Total_Core_Seconds)
print("Total ADMM Iterations: ", Total_Number_ADMM_Iterations)


# %% MARK: Export Results
# EXPORT_DATA = True
# DRAW_PLOTS = True
# EXPORT_PLOTS = True

# OUTPUT_FOLDER = "Output_Data_DMPC"
# import os
# if not os.path.exists(OUTPUT_FOLDER):
#     os.mkdir(OUTPUT_FOLDER)

Simulation_Data["Cost_Function"] = Cost_Function
Simulation_Data["Total_computation_time"] = Total_Computation_Time
Simulation_Data["Total_ADMM_Iterations"] = Total_Number_ADMM_Iterations
Simulation_Data["Total_core_seconds"] = Total_Core_Seconds

if EXPORT_SIMULATION == True:
    with open(SIMULATION_FOLDER + Simumation_ID + "_SimulationData", 'wb') as Write_File:
        pk.dump(Simulation_Data, Write_File)


# %% MARK: Plot Simulations

# State
Fig_Ref = plt.figure(figsize=(8, 6))
plt.title(f'Reference evolution', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'$r(k)$', fontsize=18)
plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON-1)
plt.ylim(1.1*min(Reference_Signal[0,:]),1.1*max(Reference_Signal[0,:]))
plt.plot(Reference_Signal.T,drawstyle='steps-post')
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.show()

# State
Fig_State = plt.figure(figsize=(8, 6))
plt.title(f'State evolution', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'$x(k)$', fontsize=18)
plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON-1)
plt.ylim(MIN_x*1.1,MAX_x*1.1)
plt.plot(x_evolution.T,drawstyle='steps-post')
plt.axhline(y = MIN_x, color='red', linestyle='--', linewidth=2) 
plt.axhline(y = MAX_x, color='red', linestyle='--', linewidth=2) 
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.show()

# Input
Fig_Input = plt.figure(figsize=(8, 6))
plt.title(f'Input evolution', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'$u(k)$', fontsize=18)
plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON-1)
plt.ylim(MIN_u*1.1,MAX_u*1.1)
plt.plot(u_evolution.T,drawstyle='steps-post')
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.axhline(y = MIN_u, color='red', linestyle='--', linewidth=2) 
plt.axhline(y = MAX_u, color='red', linestyle='--', linewidth=2) 
plt.show()

# Cost
Fig_Cost = plt.figure(figsize=(8, 6))
plt.title(f'Cost evolution', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'$J(k)$', fontsize=18)
plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON-1)
plt.ylim(0,1.1*max(Cost_Function))
plt.plot(Cost_Function.T,drawstyle='steps-post')
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.show()

# Residual
Fig_Residual = plt.figure(figsize=(8, 6))
plt.title(f'Residual evolution', fontsize=24)
plt.xlabel(f'Iteration $i$', fontsize=18)
plt.ylabel(f'$\epsilon(i)$', fontsize=18)
plt.tight_layout()
plt.xlim(0,max_number_of_iterations)
plt.ylim(0,1.1*max(residual_evolution))
plt.plot(residual_evolution.T,drawstyle='steps-post')
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.show()



# %% MARK: Export Plots

if (PLOT_FIGURES == True) and (EXPORT_FIGURES == True):
    Fig_State.savefig(SIMULATION_FOLDER + Simumation_ID + "_State.pdf", dpi=300)
    Fig_State.savefig(SIMULATION_FOLDER + Simumation_ID + "_State.jpeg", dpi=300)

    Fig_Input.savefig(SIMULATION_FOLDER + Simumation_ID + "_Input.pdf", dpi=300)
    Fig_Input.savefig(SIMULATION_FOLDER + Simumation_ID + "_Input.jpeg", dpi=300)

    Fig_Cost.savefig(SIMULATION_FOLDER + Simumation_ID + "_Cost.pdf", dpi=300)
    Fig_Cost.savefig(SIMULATION_FOLDER + Simumation_ID + "_Cost.jpeg", dpi=300)

    Fig_Residual.savefig(SIMULATION_FOLDER + Simumation_ID + "_Residual.pdf", dpi=300)
    Fig_Residual.savefig(SIMULATION_FOLDER + Simumation_ID + "_Residual.jpeg", dpi=300)

    Fig_Ref.savefig(SIMULATION_FOLDER + Simumation_ID + "_Ref.pdf", dpi=300)
    Fig_Ref.savefig(SIMULATION_FOLDER + Simumation_ID + "_Ref.jpeg", dpi=300)

