# To add a new cell, type '# %%'
# To add a new markdown cell, type '# %% [markdown]'
# %%
# from IPython import get_ipython

# %% [markdown]
# # Preliminary Operations 
# %% [markdown]
# ## Reset

# %%
# from IPython import get_ipython
# get_ipython().run_line_magic('reset', '-f')

# %% [markdown]
# ## Options and file names

# %%
#  Configuration
ExportData = False
DrawPlots = False
StorePlots = False
BuildSystem = False


# Select the network

BIG_NETWORK = False
SMALL_NETWORK = False
SMALL_NETWORK_FIVE = False
RANDOM_NETWORK = False
RANDOM_NETWORK_PARTIAL = False

WARM_START = False

# %%
import numpy as np 
import pandas as pd
import gurobipy as gp

import os
if not os.path.exists("A09_Output_Data"):
    os.mkdir("A09_Output_Data")

OUTPUT_FOLDER = "A09_Output_Data"

import pickle as pk
import math as mt

# For time tracking
import time as tm

# Support functions
from A00_Support_Funcitons import * 

# %%
# Support functions
def create_empty_list(N):
    return [[] for _ in range(N)]

def create_empty_list_2D(x, y):
    empty_list = [[] for _ in range(x)]
    for i in range(0,x):
        empty_list[i] = [[] for _ in range(y)]
    return empty_list

# %% [markdown]
# # Load Network


# System_File = "Modular_64"
System_File = "Random_50"

with open("Input_Data/" + System_File, 'rb') as f:
    Network_Dict = pk.load(f)

A = Network_Dict['A']
B = Network_Dict['B']
WEIGHTED_ADJACENCY_MATRIX = Network_Dict['Weighted_Adjacency_Matrix']
NUMBER_ATOMIC_AGENTS = Network_Dict['NUMBER_OF_ATOMIC_AGENTS']

# %% [markdown]
# # Partitioning Integrated

# %%
#  Model initialization
Gurobi_Model_Partitioning = gp.Model("Partitioning_Problem")
delta_Partitioning = np.zeros((NUMBER_ATOMIC_AGENTS, NUMBER_ATOMIC_AGENTS), dtype=object)


for i in range(0,NUMBER_ATOMIC_AGENTS):
    for j in range(0,NUMBER_ATOMIC_AGENTS):
        delta_Partitioning[i,j] = Gurobi_Model_Partitioning.addVar(vtype=gp.GRB.BINARY)


# %%
# Non-overlap constraints for control agents

Total_Number_Assignments = 0
Cost_Partitioning = 0
Total_Constraint = 0
Weight_Constraint = 100
Number_of_Sums = 0
# Small network
# WEIGHT_SMALLEST_ELEMENT = 0.01
# BIg Netwrok
WEIGHT_SMALLEST_ELEMENT = float('inf')
for i in range(0,NUMBER_ATOMIC_AGENTS):
    for j in range(0,NUMBER_ATOMIC_AGENTS):
        if (WEIGHTED_ADJACENCY_MATRIX[i,j] < WEIGHT_SMALLEST_ELEMENT) and (WEIGHTED_ADJACENCY_MATRIX[i,j] > 0):
            WEIGHT_SMALLEST_ELEMENT = WEIGHTED_ADJACENCY_MATRIX[i,j]

NUMERATOR_WEIGHT = 1
DENOMINATOR_WEIGHT = 1


for i in range(0,NUMBER_ATOMIC_AGENTS):
    ROW_SUM = 0
    for j in range(0,NUMBER_ATOMIC_AGENTS):
        ROW_SUM = ROW_SUM + delta_Partitioning[i,j]
    Gurobi_Model_Partitioning.addConstr(ROW_SUM == 1)

# %% [markdown]
# ## Optimization Problem Solving

# %%
Cost_Partitioning = 0

Intra_Cost = 0
Extra_Cost = 0
Total_Agent_Size = 0

# Random 50
# ALPHA = 1*((1/WEIGHT_SMALLEST_ELEMENT)**2)
# Partition_ID = "_Partitioning_01"
# ALPHA = 0.01*((1/WEIGHT_SMALLEST_ELEMENT)**2)
# Partition_ID = "_Partitioning_02"
# ALPHA = 0.0001*((1/WEIGHT_SMALLEST_ELEMENT)**2)
# Partition_ID = "_Partitioning_03"
# ALPHA = 0.000001*((1/WEIGHT_SMALLEST_ELEMENT)**2)
# Partition_ID = "_Partitioning_04"
# Additional
# ALPHA = 0.001*((1/WEIGHT_SMALLEST_ELEMENT)**2)
# Partition_ID = "_Partitioning_05"
# ALPHA = 0.1*((1/WEIGHT_SMALLEST_ELEMENT)**2)
# Partition_ID = "_Partitioning_06"
ALPHA = 10*((1/WEIGHT_SMALLEST_ELEMENT)**2)
Partition_ID = "_Partitioning_07"

# Modular network 64
# ALPHA = ((1/WEIGHT_SMALLEST_ELEMENT)**2)
# Partition_ID = "_Partitioning_01"
# ALPHA = 0.01*((1/WEIGHT_SMALLEST_ELEMENT)**2)
# Partition_ID = "_Partitioning_02"
# ALPHA = 0.0001*((1/WEIGHT_SMALLEST_ELEMENT)**2)
# Partition_ID = "_Partitioning_03"
# ALPHA = 0.000001*((1/WEIGHT_SMALLEST_ELEMENT)**2)
# Partition_ID = "_Partitioning_04"
# ALPHA = 0.1*((1/WEIGHT_SMALLEST_ELEMENT)**2)
# Partition_ID = "_Partitioning_02_Retest1"
# ALPHA = 0.001*((1/WEIGHT_SMALLEST_ELEMENT)**2)
# Partition_ID = "_Partitioning_02_Retest2"


START_TIME = tm.time()
print("Start construction optimization problem")
print("####################################################")

# quicksum( for for for for)
for i in range(0, NUMBER_ATOMIC_AGENTS):
    for col_i in range(0,NUMBER_ATOMIC_AGENTS):
        for j in range(0, NUMBER_ATOMIC_AGENTS):
            
            Intra_Cost = gp.quicksum([Intra_Cost, ((1/WEIGHT_SMALLEST_ELEMENT)**2)*(delta_Partitioning[i,col_i]*delta_Partitioning[j,col_i])*((+np.absolute(WEIGHTED_ADJACENCY_MATRIX[i,j])+np.absolute(WEIGHTED_ADJACENCY_MATRIX[j,i])))])

            for col_j in range(0,NUMBER_ATOMIC_AGENTS):
                if (i != j) and (col_i != col_j):
                    Extra_Cost = gp.quicksum([Extra_Cost , ((1/WEIGHT_SMALLEST_ELEMENT)**2)*(delta_Partitioning[i,col_i]*delta_Partitioning[j,col_j])*((np.absolute(WEIGHTED_ADJACENCY_MATRIX[i,j])+np.absolute(WEIGHTED_ADJACENCY_MATRIX[j,i])))])


    # print("Agent ", i, " complete")
    END_TIME_ITERATION = tm.time() - START_TIME
    print("Computation percentage:", round(((i+1)/NUMBER_ATOMIC_AGENTS)*100, 2), "%")
    print("Enlapsed time:", round(END_TIME_ITERATION,2), "[s]")
    print("Expected time to finish:", round(((END_TIME_ITERATION * NUMBER_ATOMIC_AGENTS)/(i + 1)) - END_TIME_ITERATION, 2), "[s]")
    print("####################################################")

for col_i in range(0, NUMBER_ATOMIC_AGENTS):
    Agent_Size = 0
    for i in range(0,NUMBER_ATOMIC_AGENTS):
        Agent_Size = Agent_Size + delta_Partitioning[i,col_i]
    Total_Agent_Size = Total_Agent_Size + (Agent_Size*Agent_Size)





Cost_Partitioning = Extra_Cost + (-Intra_Cost) + ALPHA*(Total_Agent_Size)

# Options
# Most interacting topology
Gurobi_Model_Partitioning.setObjective(Cost_Partitioning, gp.GRB.MINIMIZE) 

END_TIME = tm.time()

TIME_MODEL_CONSTRUCTION = END_TIME - START_TIME
print("#################################################### Model construction duration: ", round(TIME_MODEL_CONSTRUCTION, 2), " seconds")

# Warm start 
if WARM_START == True:
    Gurobi_Model_Partitioning.NumStart = 1
    Gurobi_Model_Partitioning.update()
    Gurobi_Model_Partitioning.params.StartNumber = 0
    
    var_idx = 0
    for i in range(0, NUMBER_ATOMIC_AGENTS):
        for j in range(0,NUMBER_ATOMIC_AGENTS):
            var = Gurobi_Model_Partitioning.getVars()[var_idx]
            if i == j:            
                var.Start = 1
            else:
                var.Start = 0
            var_idx += 1



# %%
START_TIME = tm.time()
Gurobi_Model_Partitioning.Params.TimeLimit  = 60*60*12
Gurobi_Model_Partitioning.Params.Threads = 8
# Gurobi_Model_Partitioning.Params.Heuristics = 0.15
# Gurobi_Model_Partitioning.Params.MIPFocus = 3
Gurobi_Model_Partitioning.optimize()
END_TIME = tm.time()

TIME_OPTIMIZATION = END_TIME - START_TIME
print("#################################################### Optimization duration: ", round(TIME_OPTIMIZATION, 2), " seconds")


# %% Solutions
#
Delta_Partitioning = np.zeros((NUMBER_ATOMIC_AGENTS, NUMBER_ATOMIC_AGENTS))

var_idx = 0
for i in range(0,NUMBER_ATOMIC_AGENTS):
    for j in range(0,NUMBER_ATOMIC_AGENTS):
        Delta_Partitioning[i,j] = Gurobi_Model_Partitioning.getVars()[var_idx].x
        var_idx += 1


Number_Of_Assignments = 0

Control_Agents_List = create_empty_list(NUMBER_ATOMIC_AGENTS)
for i in range(0, NUMBER_ATOMIC_AGENTS):
    for j in range(0,NUMBER_ATOMIC_AGENTS):
        if Delta_Partitioning[i,j] == 1:
            Control_Agents_List[j].append(i)

# Reorganize LIst

Control_Agents_List = [Control_Agent for Control_Agent in Control_Agents_List if Control_Agent != []]
Number_Of_Control_Agents = len(Control_Agents_List)


# %%
TIME_PARTITIONING = TIME_OPTIMIZATION+TIME_MODEL_CONSTRUCTION
Partitioning_List = {"Control_Agents_List": Control_Agents_List,
                     "Number_Of_Control_Agents": Number_Of_Control_Agents,
                     "ALPHA": ALPHA,
                     "TIME_OPTIMIZATION": TIME_OPTIMIZATION,
                     "TIME_MODEL_CONSTRUCTION": TIME_MODEL_CONSTRUCTION,
                     "TIME_PARTITIONING": TIME_PARTITIONING,
                     "A": A, 
                     "B": B, 
                     "WEIGHTED_ADJACENCY_MATRIX": WEIGHTED_ADJACENCY_MATRIX, 
                     "NUMBER_ATOMIC_AGENTS": NUMBER_ATOMIC_AGENTS }
# Partition_ID = "_Partitioning_01"
with open(OUTPUT_FOLDER + "/" + System_File + "_ALPHA_" + str(ALPHA), 'wb') as Write_File:
    pk.dump(Partitioning_List, Write_File)




