# %% 
# MARK: Reset Operations
# from IPython import get_ipython
# get_ipython().run_line_magic('reset', '-f')
# print("\n" * 100)

# %%
# MARK: Import libraries


# For matrix operations
import numpy as np
# For file storage
import pickle as pk
# For time tracking
import time as tm
# For data location
INPUT_FOLDER = "A09_Output_Data"
OUTPUT_FOLDER = "A09_Output_Data"
import os
if not os.path.exists(OUTPUT_FOLDER):
    os.mkdir(OUTPUT_FOLDER)


import networkx as nx
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib import rc
# Enable LaTeX rendering
rc('text', usetex=True)

# Set the font to Times New Roman or the closest available font in LaTeX
rc('font', family='serif')
rc('font', serif='Times New Roman')  # This line might need adjustment based on LaTeX distribution

from scipy.spatial import ConvexHull
import matplotlib.colors as mcolors
import math as mt
# Support functions
from A00_Support_Funcitons import *
from A02_Weighting_Function import*

# Settings
EXPORT_DATA = True
DRAW_PLOTS = True
EXPORT_PLOTS = True


# %%
# MARK: Select network
# Network_File = "Modular_64_Partitioning_01_ALPHA_1000000"
# Network_File = "Modular_64_Partitioning_02_ALPHA_10000"
# Network_File = "Modular_64_Partitioning_03_ALPHA_100"
# Network_File = "Modular_64_Partitioning_04_ALPHA_1"
Network_File = "Modular_64_Test_1"

with open("A09_Output_Data/Modular_64/" + Network_File, 'rb') as f:
    Network_Dict = pk.load(f)

A = Network_Dict['A']
B = Network_Dict['B']
Weighted_Adjacency_Matrix = Network_Dict['WEIGHTED_ADJACENCY_MATRIX']
NUMBER_ATOMIC_AGENTS = Network_Dict['NUMBER_ATOMIC_AGENTS']
NUMBER_CONTROL_AGENTS = Network_Dict['Number_Of_Control_Agents']
ALPHA = Network_Dict['ALPHA'] 
Control_Agents_List = Network_Dict['Control_Agents_List'] 
Partition = Control_Agents_List.copy()
List_Atomic_Agents = []
for i in range(NUMBER_ATOMIC_AGENTS):
    List_Atomic_Agents.append(i)
# print("test")

# with open("Output_Data/" + Network_File, 'rb') as f:
#     Network_Dict = pk.load(f)

Partition_ID = Network_File + "_ALPHA_" + str(ALPHA)

# %% MARK: Node positions

Node_Positions = create_empty_list(NUMBER_ATOMIC_AGENTS)

Node_Positions[0] = [0,0] 
Node_Positions[1] = [0,1]
Node_Positions[2] = [1,0]  
Node_Positions[3] = [1,1]  

Node_Positions[4] = [0,2] 
Node_Positions[5] = [0,3]
Node_Positions[6] = [1,2]  
Node_Positions[7] = [1,3] 

Node_Positions[8] = [2,0] 
Node_Positions[9] = [2,1]
Node_Positions[10] = [3,0]  
Node_Positions[11] = [3,1] 

Node_Positions[12] = [2,2] 
Node_Positions[13] = [2,3]
Node_Positions[14] = [3,2]  
Node_Positions[15] = [3,3]

Node_Positions[16] = [0,4] 
Node_Positions[17] = [0,5] 
Node_Positions[18] = [1,4] 
Node_Positions[19] = [1,5] 

Node_Positions[20] = [0,6] 
Node_Positions[21] = [0,7] 
Node_Positions[22] = [1,6] 
Node_Positions[23] = [1,7] 

Node_Positions[24] = [2,4] 
Node_Positions[25] = [2,5] 
Node_Positions[26] = [3,4] 
Node_Positions[27] = [3,5] 

Node_Positions[28] = [2,6] 
Node_Positions[29] = [2,7] 
Node_Positions[30] = [3,6] 
Node_Positions[31] = [3,7] 

Node_Positions[32] = [4,0] 
Node_Positions[33] = [4,1]
Node_Positions[34] = [5,0]  
Node_Positions[35] = [5,1]  

Node_Positions[36] = [4,2] 
Node_Positions[37] = [4,3]
Node_Positions[38] = [5,2]  
Node_Positions[39] = [5,3] 

Node_Positions[40] = [6,0] 
Node_Positions[41] = [6,1]
Node_Positions[42] = [7,0]  
Node_Positions[43] = [7,1] 

Node_Positions[44] = [6,2] 
Node_Positions[45] = [6,3]
Node_Positions[46] = [7,2]  
Node_Positions[47] = [7,3]

Node_Positions[48] = [4,4] 
Node_Positions[49] = [4,5] 
Node_Positions[50] = [5,4] 
Node_Positions[51] = [5,5] 

Node_Positions[52] = [4,6] 
Node_Positions[53] = [4,7] 
Node_Positions[54] = [5,6] 
Node_Positions[55] = [5,7] 

Node_Positions[56] = [6,4] 
Node_Positions[57] = [6,5] 
Node_Positions[58] = [7,4] 
Node_Positions[59] = [7,5] 

Node_Positions[60] = [6,6] 
Node_Positions[61] = [6,7] 
Node_Positions[62] = [7,6] 
Node_Positions[63] = [7,7] 

# %% MARK: Partitioned graph representation

Agents_Figure = plt.figure("Control agents", figsize=(8, 8))
Agents_Ax = plt.gca()

base_radus = 1
radius_circle = 0.2
FONT_SIZE_ATOMIC = 8

for idx_figure in range(0, NUMBER_ATOMIC_AGENTS):
    circle = plt.Circle((Node_Positions[idx_figure][0], Node_Positions[idx_figure][1]), radius_circle, color='black', alpha=1, zorder=3)
    Agents_Ax.add_patch(circle)
    circle = plt.Circle((Node_Positions[idx_figure][0], Node_Positions[idx_figure][1]), radius_circle*0.9, color="white", alpha=1 , zorder=3)
    Agents_Ax.add_patch(circle)
    circle = plt.Circle((Node_Positions[idx_figure][0], Node_Positions[idx_figure][1]), radius_circle*0.9, color='cyan', alpha=0.5 , zorder=3)
    plt.text(Node_Positions[idx_figure][0], Node_Positions[idx_figure][1],  f'$\mathcal{{A}}_{{{idx_figure+1}}}$', color="black", fontsize=FONT_SIZE_ATOMIC, horizontalalignment="center", verticalalignment="center", zorder=4)
    Agents_Ax.add_patch(circle)

for idx in range(NUMBER_CONTROL_AGENTS):
        for i in range(len(Partition[idx])):
            agent_idx_i = Partition[idx][i]

            for j in range(NUMBER_ATOMIC_AGENTS):
                agent_idx_j = j
                if agent_idx_j in Partition[idx]:
                    if agent_idx_i != agent_idx_j:
                        if Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j] != 0:
                            agent_i_position = Node_Positions[agent_idx_i]
                            agent_j_position = Node_Positions[agent_idx_j]
                            plt.plot([agent_i_position[0], agent_j_position[0]], [agent_i_position[1], agent_j_position[1]], color='black', zorder=2, linewidth=25/mt.log(Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j]), alpha=0.5)
                else:
                    if Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j] != 0:
                        agent_i_position = Node_Positions[agent_idx_i]
                        agent_j_position = Node_Positions[agent_idx_j]
                        plt.plot([agent_i_position[0], agent_j_position[0]], [agent_i_position[1], agent_j_position[1]], color="red", zorder=2, linewidth=25/mt.log(Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j]), alpha=0.5)

# Convex hull
number_points_circumference = 1000
for idx in range(NUMBER_CONTROL_AGENTS):
    Points_Convex_Hull = np.zeros((number_points_circumference*len(Partition[idx]),2))
    global_idx = 0
    for i in range(len(Partition[idx])):
        agent_idx_i = Partition[idx][i]
        agent_i_position = Node_Positions[agent_idx_i]
        Circle_Points = generate_circle_points([agent_i_position[0],agent_i_position[1]], radius_circle*1.8, number_points_circumference)
        for idx_circle in range(number_points_circumference):
            Points_Convex_Hull[global_idx,0] = Circle_Points[0][idx_circle]
            Points_Convex_Hull[global_idx,1] = Circle_Points[1][idx_circle]
            global_idx+= 1 

        Points_Convex_Hull[i,0] = agent_i_position[0]
        Points_Convex_Hull[i,1] = agent_i_position[1]
        
    hull = ConvexHull(Points_Convex_Hull)
    plt.fill(Points_Convex_Hull[hull.vertices,0], Points_Convex_Hull[hull.vertices,1], color='#B5E6A2', alpha=0.3, zorder=1)

plt.xlim(-1, 8)
plt.ylim(-1, 8)
plt.axis("equal")
plt.axis('off') 

fig = plt.gcf()

fig.patch.set_linewidth(1)
fig.patch.set_edgecolor('black')
Agents_Figure.tight_layout()  
plt.tick_params(axis='both',        # Apply to both x and y axes
                which='both',       # Apply to both major and minor ticks
                bottom=False,       # Turn off bottom ticks
                top=False,          # Turn off top ticks
                left=False,         # Turn off left ticks
                right=False,        # Turn off right ticks
                labelbottom=False,  # Turn off bottom labels
                labelleft=False)    # Turn off left labels
plt.show()

# %% MARK: Export graphics
# 
if EXPORT_PLOTS == True:
    Agents_Figure.savefig("A09_Output_Data/" + Network_File + ".png", dpi=300, bbox_inches="tight") 

# %% MARK: Modular graph representation

Agents_Figure = plt.figure("Atomic agents", figsize=(8, 8))
Agents_Ax = plt.gca()

base_radus = 1
radius_circle = 0.2
FONT_SIZE_ATOMIC = 8

for idx_figure in range(0, NUMBER_ATOMIC_AGENTS):
    circle = plt.Circle((Node_Positions[idx_figure][0], Node_Positions[idx_figure][1]), radius_circle, color='black', alpha=1, zorder=3)
    Agents_Ax.add_patch(circle)
    circle = plt.Circle((Node_Positions[idx_figure][0], Node_Positions[idx_figure][1]), radius_circle*0.9, color="white", alpha=1 , zorder=3)
    Agents_Ax.add_patch(circle)
    circle = plt.Circle((Node_Positions[idx_figure][0], Node_Positions[idx_figure][1]), radius_circle*0.9, color='cyan', alpha=0.5 , zorder=3)
    plt.text(Node_Positions[idx_figure][0], Node_Positions[idx_figure][1],  f'$\mathcal{{A}}_{{{idx_figure+1}}}$', color="black", fontsize=FONT_SIZE_ATOMIC, horizontalalignment="center", verticalalignment="center", zorder=4)
    Agents_Ax.add_patch(circle)

for idx in range(NUMBER_CONTROL_AGENTS):
        for i in range(len(Partition[idx])):
            agent_idx_i = Partition[idx][i]

            for j in range(NUMBER_ATOMIC_AGENTS):
                agent_idx_j = j
                if Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j] != 0:
                    agent_i_position = Node_Positions[agent_idx_i]
                    agent_j_position = Node_Positions[agent_idx_j]
                    plt.plot([agent_i_position[0], agent_j_position[0]], [agent_i_position[1], agent_j_position[1]], color="#40e0d0", zorder=2, linewidth=25/mt.log(Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j]), alpha=0.5)
                    

plt.xlim(-1, 8)
plt.ylim(-1, 8)
plt.axis("equal")
plt.axis('off') 


Agents_Figure.patch.set_linewidth(1)
Agents_Figure.patch.set_edgecolor('black')
Agents_Figure.tight_layout()  
plt.tick_params(axis='both',        # Apply to both x and y axes
                which='both',       # Apply to both major and minor ticks
                bottom=False,       # Turn off bottom ticks
                top=False,          # Turn off top ticks
                left=False,         # Turn off left ticks
                right=False,        # Turn off right ticks
                labelbottom=False,  # Turn off bottom labels
                labelleft=False)    # Turn off left labels

plt.show()

# %% MARK: Export graphics
# 
if EXPORT_PLOTS == True:
    
    Agents_Figure.savefig("A09_Output_Data/Modular_64"  + ".png", dpi=300, bbox_inches="tight") 

# %%
