# %% 
# MARK: Reset Operations
# from IPython import get_ipython
# get_ipython().run_line_magic('reset', '-f')
# print("\n" * 100)

# %%
# MARK: Import libraries


# For matrix operations
import numpy as np
# For file storage
import pickle as pk
# For time tracking
import time as tm
# For data location
INPUT_FOLDER = "Input_Data"

# import os
# if not os.path.exists(OUTPUT_FOLDER):
#     os.mkdir(OUTPUT_FOLDER)


# import networkx as nx
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib import rc
# Enable LaTeX rendering
rc('text', usetex=True)

# Set the font to Times New Roman or the closest available font in LaTeX
rc('font', family='serif')
rc('font', serif='Times New Roman')  # This line might need adjustment based on LaTeX distribution

from scipy.spatial import ConvexHull
import matplotlib.colors as mcolors
import networkx as nx
import pandas as pd

# Support functions
from A00_Support_Funcitons import *
from A02_Weighting_Function import*

# Settings
EXPORT_DATA = False
DRAW_PLOTS = True
EXPORT_PLOTS = True


# %%
# MARK: Select network

Network_File = "Random_50"
# Network_File = "Modular_64"


with open("Input_Data/" + Network_File, 'rb') as f:
    Network_Dict = pk.load(f)

A = Network_Dict['A']
B = Network_Dict['B']
Weighted_Adjacency_Matrix = Network_Dict['Weighted_Adjacency_Matrix']
NUMBER_OF_ATOMIC_AGENTS = Network_Dict['NUMBER_OF_ATOMIC_AGENTS']
List_Atomic_Agents = []
for i in range(NUMBER_OF_ATOMIC_AGENTS):
    List_Atomic_Agents.append(i)


#  %% MARK: Load Partition
# Optimization-based Partitioning 
PARTITIONING_FOLDER = "A09_Output_Data/Random 50/"
OUTPUT_FOLDER = "A09_Output_Data/Random 50/"
PARTITIONING_NAME = "Random_50_ALPHA_4304310.87298206" # Partition 4 - 1 agent
Partition_Number = "Partition_7"
# PARTITIONING_NAME = "Random_50_ALPHA_430431.08729820594" # Partition 1 - 41 agents
# Partition_Number = "Partition_1"
# PARTITIONING_NAME = "Random_50_ALPHA_43043.1087298206" # Partition 6 - 15 agents
# Partition_Number = "Partition_6"
# PARTITIONING_NAME = "Random_50_ALPHA_4304.31087298206" # Partition 2 - 6 agents
# Partition_Number = "Partition_2"
# PARTITIONING_NAME = "Random_50_ALPHA_430.43108729820597" # Partition 5 - 2 agents
# Partition_Number = "Partition_5"
# PARTITIONING_NAME = "Random_50_ALPHA_43.043108729820595" # Partition 3 - 1 agent
# Partition_Number = "Partition_3"
# PARTITIONING_NAME = "Random_50_ALPHA_0.4304310872982059" # Partition 4 - 1 agent

# Algorithmic Partitioning
# PARTITIONING_FOLDER = "A03_Output_Data_Algorithmic_Partitioning_Mixed/"
# OUTPUT_FOLDER = "A03_Output_Data_Algorithmic_Partitioning_Mixed/"
# PARTITIONING_NAME = "Random_50_Partition_01_ALPHA_100"
# Partition_Number = "Partition_1"
# PARTITIONING_NAME = "Random_50_Partition_02_ALPHA_30"
# Partition_Number = "Partition_2"
# PARTITIONING_NAME = "Random_50_Partition_03_ALPHA_20"
# Partition_Number = "Partition_3"
# PARTITIONING_NAME = "Random_50_Partition_04_ALPHA_10"
# Partition_Number = "Partition_4"
# PARTITIONING_NAME = "Random_50_Partition_05_ALPHA_1"
# Partition_Number = "Partition_5"
# PARTITIONING_NAME = "Random_50_Partition_06_ALPHA_0.1"
# Partition_Number = "Partition_6"
# PARTITIONING_NAME = "Random_50_Partition_09_ALPHA_1e-13"
# Partition_Number = "Partition_09"

Simumation_ID = PARTITIONING_NAME 

# For algorithmic partitioning
# For optimization-based partitioning
with open(PARTITIONING_FOLDER + PARTITIONING_NAME, 'rb') as file:
    Partitioning_Dict = pk.load(file)
######################################################################################
# For algorithmic partitioning
# Partition = Partitioning_Dict["Partition"]
# NUMBER_CONTROL_AGENTS = Partitioning_Dict["NUMBER_CONTROL_AGENTS"]
# For optimization-based partitioning
Partition = Partitioning_Dict["Control_Agents_List"]
NUMBER_CONTROL_AGENTS = Partitioning_Dict["Number_Of_Control_Agents"]
######################################################################################
# # For GA Partitioning
# # Load the CSV file
# df = pd.read_csv(PARTITIONING_FOLDER + PARTITIONING_NAME, header=None)
# df = df.fillna("")

# # Drop NaN values from the entire DataFrame
# # df_cleaned = df.dropna()

# NUMBER_CONTROL_AGENTS = df.shape[0]


# # Convert the DataFrame to a list of lists (each row is a list)
# Partition_Raw = create_empty_list(NUMBER_CONTROL_AGENTS)
# Partition = create_empty_list(NUMBER_CONTROL_AGENTS)
# for idx in range(NUMBER_CONTROL_AGENTS):
#     Partition_Raw[idx] = df.iloc[idx].tolist()

# for i in range(NUMBER_CONTROL_AGENTS):
#     for j in range(len(Partition_Raw[i])):
#         if Partition_Raw[i][j] != '':
#             Partition[i].append(int(Partition_Raw[i][j]))
######################################################################################
ALPHA = Partitioning_Dict['ALPHA']
Partition_ID = Network_File + "_" + Partition_Number + "_ALPHA_" + str(ALPHA)
Partition_Holder = Partition.copy()


# %% MARK: Agent-based circular representation
if DRAW_PLOTS == True:
    number_points_circumference = 100
    # Build system graph
    Network_Graph = nx.Graph()
    for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
        for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
            if i != j:
                if Weighted_Adjacency_Matrix[i,j] != 0:
                    Network_Graph.add_edge(i, j, weight=Weighted_Adjacency_Matrix[i,j])

    degree_sequence = sorted((d for n, d in Network_Graph.degree()), reverse=True)
    degree_sequence = dict(Network_Graph.degree())
    dmax = max(degree_sequence)

    # Number of sections of the external radius
    NUMBER_SECTIONS = NUMBER_CONTROL_AGENTS
    Total_Radius = max(degree_sequence.values()) + 1
    base_radus = 2
    radius_circle = 0.35
    FONT_SIZE_ATOMIC = 8
    # List of the positions of the nodes
    positions_control_agents = create_empty_list(NUMBER_OF_ATOMIC_AGENTS)
    Partition = Partition_Holder.copy()
    MAX_DEGREE = max(degree_sequence.values())
    list_agents = []
    global_idx = 0
    import math as mt


    Agents_Figure = plt.figure("Control agents", figsize=(8, 8))
    Agents_Ax = plt.gca()

    # Edges tickness
    Edges_Alpha = np.zeros((NUMBER_OF_ATOMIC_AGENTS*NUMBER_OF_ATOMIC_AGENTS,1))
    external_idx = 0
    MAX_WEIGHT = np.abs(Weighted_Adjacency_Matrix.max())
    for i in range(0,NUMBER_OF_ATOMIC_AGENTS):
        for j in range(0,NUMBER_OF_ATOMIC_AGENTS):
            if i != j:
                if Weighted_Adjacency_Matrix[i,j] != 0:
                    Edges_Alpha[external_idx] = np.abs(Weighted_Adjacency_Matrix[i,j])/MAX_WEIGHT 
                    external_idx += 1
    non_zero_mask = Edges_Alpha != 0
    Edges_Alpha_Filtered = Edges_Alpha[non_zero_mask]

    # Graph coloring list
    Node_Degrees_List = Network_Graph.degree()
    Table_Colors = mcolors.TABLEAU_COLORS

    Color_List_Base = list(Table_Colors.keys())
    Color_List = create_empty_list(NUMBER_OF_ATOMIC_AGENTS)
    external_idx = 0
    for idx in range(NUMBER_OF_ATOMIC_AGENTS):
        Color_List[idx] = Color_List_Base[external_idx]
        external_idx += 1
        if external_idx >= len(Color_List_Base):
            external_idx = 0
    CUMULATIVE_PERCENTAGE = 0
    for control_agent_idx in range(len(Partition)):
        PERCENTAGE_SECTOR = len(Partition[control_agent_idx])/NUMBER_OF_ATOMIC_AGENTS
        
        for degree_idx in range(1, MAX_DEGREE+1):
            number_agents_degree = 0
            for i in range(len(Partition[control_agent_idx])):
                atomic_agent_idx = Partition[control_agent_idx][i]
                if degree_sequence[atomic_agent_idx] == degree_idx:
                    # print(atomic_agent_idx)
                    # print(degree_sequence[atomic_agent_idx])
                    number_agents_degree += 1
            progressing_item = 0
            if number_agents_degree != 0:
                radius_ring = base_radus*(Total_Radius - degree_idx)
                
                for i in range(len(Partition[control_agent_idx])):
                    

                    atomic_agent_idx = Partition[control_agent_idx][i]
                    if degree_sequence[atomic_agent_idx] == degree_idx:
                        # atomic_agent_idx = Partition[control_agent_idx][i]
                        # positions_control_agents[global_idx] = [radius_ring * mt.cos(((2*mt.pi*control_agent_idx)/(NUMBER_SECTIONS))+((2*mt.pi*progressing_item)/(number_agents_degree))/(NUMBER_SECTIONS)),
                        #                         radius_ring * mt.sin(((2*mt.pi*control_agent_idx)/(NUMBER_SECTIONS))+((2*mt.pi*progressing_item)/(number_agents_degree))/(NUMBER_SECTIONS))]
                        # positions_control_agents[global_idx] = [radius_ring * mt.cos(((2*mt.pi*control_agent_idx)/(NUMBER_SECTIONS))*(1 + (progressing_item/number_agents_degree))),
                        #                         radius_ring * mt.sin(((2*mt.pi*control_agent_idx)/(NUMBER_SECTIONS))*(1 + (progressing_item/number_agents_degree)))]
                        # positions_control_agents[global_idx] = [
                        # radius_ring * mt.cos(2*mt.pi * (control_agent_idx/NUMBER_SECTIONS)+ (1/NUMBER_SECTIONS)*(progressing_item/number_agents_degree)), radius_ring * mt.sin(2*mt.pi * (control_agent_idx/NUMBER_SECTIONS) + (1/NUMBER_SECTIONS)*(progressing_item/number_agents_degree))]
                        positions_control_agents[global_idx] = [
                        radius_ring * mt.cos(2*mt.pi * (CUMULATIVE_PERCENTAGE)+ (2*mt.pi*PERCENTAGE_SECTOR)*(progressing_item/number_agents_degree)), radius_ring * mt.sin(2*mt.pi * (CUMULATIVE_PERCENTAGE) + (2*mt.pi*PERCENTAGE_SECTOR)*(progressing_item/number_agents_degree))]
                        list_agents.append(atomic_agent_idx)
                        global_idx += 1
                        progressing_item += 1
        CUMULATIVE_PERCENTAGE = CUMULATIVE_PERCENTAGE + PERCENTAGE_SECTOR


    # plt.xlim(-5,5)
    # plt.ylim(-5,5)
    # FONT_SIZE_ATOMIC = 10
    # list_agents = Partition[0].copy()
    # for i in range(1, len(Partition)):
    #     for j in range(len(Partition[i])):
    #         list_agents.append(Partition[i][j])

    for idx_figure in range(len(positions_control_agents)):
        if positions_control_agents[idx_figure] != []:
            Control_Agent_Idx = 0    
            for i in range(len(Partition)):
                if list_agents[idx_figure] in Partition[i]:
                    Control_Agent_Idx = i
            # agent_i_position_idx = list_agents.index(idx_figure)  
            circle = plt.Circle((positions_control_agents[idx_figure][0], positions_control_agents[idx_figure][1]), radius_circle, color='black', alpha=1, zorder=2)
            Agents_Ax.add_patch(circle)
            circle = plt.Circle((positions_control_agents[idx_figure][0], positions_control_agents[idx_figure][1]), radius_circle*0.9, color="white", alpha=1 , zorder=2)
            Agents_Ax.add_patch(circle)
            circle = plt.Circle((positions_control_agents[idx_figure][0], positions_control_agents[idx_figure][1]), radius_circle*0.9, color=Color_List[Control_Agent_Idx], alpha=(Node_Degrees_List[list_agents[idx_figure]]/(dmax)) , zorder=2)
            plt.text(positions_control_agents[idx_figure][0], positions_control_agents[idx_figure][1],  f'$\mathcal{{A}}_{{{list_agents[idx_figure]}}}$', color="black", fontsize=FONT_SIZE_ATOMIC, horizontalalignment="center", verticalalignment="center", zorder=4)
            Agents_Ax.add_patch(circle)

    for idx in range(NUMBER_CONTROL_AGENTS):
            for i in range(len(Partition[idx])):
                agent_idx_i = Partition[idx][i]

                for j in range(NUMBER_OF_ATOMIC_AGENTS):
                    agent_idx_j = j
                    if agent_idx_j in Partition[idx]:
                        if agent_idx_i != agent_idx_j:
                            if Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j] != 0:
                                agent_i_position_idx = list_agents.index(agent_idx_i)
                                agent_j_position_idx = list_agents.index(agent_idx_j)
                                plt.plot([positions_control_agents[agent_i_position_idx][0], positions_control_agents[agent_j_position_idx][0]], [positions_control_agents[agent_i_position_idx][1], positions_control_agents[agent_j_position_idx][1]], color=Color_List[idx], zorder=1, linewidth=5*Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j], alpha=0.5)
                    else:
                        if Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j] != 0:
                            agent_i_position_idx = list_agents.index(agent_idx_i)
                            agent_j_position_idx = list_agents.index(agent_idx_j)
                            plt.plot([positions_control_agents[agent_i_position_idx][0], positions_control_agents[agent_j_position_idx][0]], [positions_control_agents[agent_i_position_idx][1], positions_control_agents[agent_j_position_idx][1]], color="black", zorder=1, linewidth=5*Weighted_Adjacency_Matrix[agent_idx_i,agent_idx_j], alpha=0.5, linestyle="dashed")

    for idx in range(NUMBER_CONTROL_AGENTS):
        Points_Convex_Hull = np.zeros((number_points_circumference*len(Partition[idx]),2))
        global_idx = 0
        for i in range(len(Partition[idx])):
            agent_idx_i = Partition[idx][i]
            agent_i_position_idx = list_agents.index(agent_idx_i)
            Circle_Points = generate_circle_points([positions_control_agents[agent_i_position_idx][0],positions_control_agents[agent_i_position_idx][1]], radius_circle*1.8, number_points_circumference)
            for idx_circle in range(number_points_circumference):
                Points_Convex_Hull[global_idx,0] = Circle_Points[0][idx_circle]
                Points_Convex_Hull[global_idx,1] = Circle_Points[1][idx_circle]
                global_idx+= 1 

            Points_Convex_Hull[i,0] = positions_control_agents[agent_i_position_idx][0]
            Points_Convex_Hull[i,1] = positions_control_agents[agent_i_position_idx][1]
            
        hull = ConvexHull(Points_Convex_Hull)
        plt.fill(Points_Convex_Hull[hull.vertices,0], Points_Convex_Hull[hull.vertices,1], color=Color_List[idx], alpha=0.3, zorder=3)

    plt.xlim(-1, 8)
    plt.ylim(-1, 8)
    plt.axis("equal")
    plt.axis('off') 
    Agents_Figure.patch.set_linewidth(1)
    Agents_Figure.patch.set_edgecolor('black')
    Agents_Figure.tight_layout()  

    plt.tick_params(axis='both',        # Apply to both x and y axes
                    which='both',       # Apply to both major and minor ticks
                    bottom=False,       # Turn off bottom ticks
                    top=False,          # Turn off top ticks
                    left=False,         # Turn off left ticks
                    right=False,        # Turn off right ticks
                    labelbottom=False,  # Turn off bottom labels
                    labelleft=False)    # Turn off left labels
    plt.show()


    if (DRAW_PLOTS == True) and (EXPORT_PLOTS == True):
        # Graph_Figure.savefig(OUTPUT_FOLDER + "/" + Partition_ID + "Graph_Figure.pdf", dpi=300)
        # Graph_Figure.savefig(OUTPUT_FOLDER + "/" + Partition_ID + "Graph_Figure.jpeg", dpi=300)
        Agents_Figure.savefig(OUTPUT_FOLDER + "/" + Partition_ID + "Agents_Figure.pdf", dpi=300, bbox_inches=None)
        Agents_Figure.savefig(OUTPUT_FOLDER + "/" + Partition_ID + "Agents_Figure.jpeg", dpi=300, bbox_inches=None)
