# from IPython import get_ipython
# get_ipython().run_line_magic('reset', '-f')

# Optimization toolbox
import cvxpy as cp
import gurobipy as gp
# For matrix operations
import numpy as np
# For file storage
import pickle as pk
# For time tracking
import time as tm
import matplotlib.pyplot as plt
from matplotlib import rc
# Enable LaTeX rendering
rc('text', usetex=True)
rc('font', family='serif')
rc('font', serif='Times New Roman')  # This line might need adjustment based on LaTeX distribution

# Support funciton
from A00_Support_Funcitons import*


# Options
PLOT_FIGURES = True
EXPORT_FIGURES = True
EXPORT_SIMULATION = False


SIMULATION_FOLDER = "A05_Output_Data_DMPC_Hybrid/Random_50_Simulations/"
SIMULATION_FILE = "Random_50_ALPHA_43.043108729820595_SimulationData"

# SIMULATION_FOLDER = "A04_Output_Data_DMPC/"
# SIMULATION_FILE = "Modular_64_Partitioning_04_ALPHA_1_SimulationData"
# SIMULATION_FILE = "Modular_64_Partitioning_03_ALPHA_100_SimulationData"
# SIMULATION_FILE = "Modular_64_Partitioning_02_ALPHA_10000_SimulationData"
# SIMULATION_FILE = "Modular_64_Partitioning_01_ALPHA_1000000_SimulationData"

Simumation_ID = SIMULATION_FILE

with open(SIMULATION_FOLDER + SIMULATION_FILE, 'rb') as file:
    Simulation_Data = pk.load(file)

Network_Dict = Simulation_Data['Network_Dict']
Partitioning_Dict = Simulation_Data['Partitioning_Dict']
Reference_Signal = Simulation_Data['Reference_Signal']
x_evolution = Simulation_Data['x_evolution']
x_evolution_PWA = Simulation_Data['x_evolution_PWA']
u_evolution = Simulation_Data['u_evolution']
z_evolution = Simulation_Data['z_evolution']
d_evolution = Simulation_Data['d_evolution']
Cost_Function = Simulation_Data['Cost_Function']
residual_evolution = Simulation_Data['residual_evolution']
epsilon_convergence = Simulation_Data['epsilon_convergence']
# Total_Computation_Time = Simulation_Data['Total_computation_time']
# Total_Core_Seconds = Simulation_Data['Total_core_seconds']
# Total_Number_ADMM_Iterations = Simulation_Data['Total_ADMM_Iterations']
max_number_of_iterations = Simulation_Data['max_number_of_iterations']
PARAMETERS_DICT = Simulation_Data['PARAMETERS_DICT']

SIMULATION_HORIZON = PARAMETERS_DICT['SIMULATION_HORIZON'] 
PREDICTION_HORIZON = PARAMETERS_DICT['PREDICTION_HORIZON']
NUMBER_ATOMIC_AGENTS = PARAMETERS_DICT['NUMBER_ATOMIC_AGENTS']
NUMBER_CONTROL_AGENTS = Partitioning_Dict["Number_Of_Control_Agents"]
MAX_u = PARAMETERS_DICT['MAX_u']
MIN_u = PARAMETERS_DICT['MIN_u']
MAX_x = PARAMETERS_DICT['MAX_x']
MIN_x = PARAMETERS_DICT['MIN_x']

# %% MARK: Evaluation metrics
print("Number of control agents: ", NUMBER_CONTROL_AGENTS)
# print("Total cost: ", Cost_Function[SIMULATION_HORIZON-1])
# print("Total computation time: ", Total_Computation_Time)
# print("Total core seconds: ", Total_Core_Seconds)
# print("Total ADMM Iterations: ", Total_Number_ADMM_Iterations)

# %% MARK: Plot Simulations

# Reference
Fig_Ref = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'Reference evolution', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'$r(k)$', fontsize=18)
plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON-1)
plt.ylim(1.1*min(Reference_Signal[0,:]),1.1*max(Reference_Signal[0,:]))
plt.plot(Reference_Signal.T,drawstyle='steps-post')
plt.grid(color='gray', linestyle='--', linewidth=0.5)
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

# State
Fig_State = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'State evolution', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'$x(k)$', fontsize=18)
plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON-1)
plt.ylim(MIN_x*1.1,MAX_x*1.1)
plt.plot(x_evolution.T,drawstyle='steps-post')
plt.axhline(y = MIN_x, color='red', linestyle='--', linewidth=2) 
plt.axhline(y = MAX_x, color='red', linestyle='--', linewidth=2) 
plt.grid(color='gray', linestyle='--', linewidth=0.5)
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

# State PWA
Fig_State_PWA = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'PWA state evolution', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'$x(k)$', fontsize=18)
plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON-1)
plt.ylim(MIN_x*1.1,MAX_x*1.1)
plt.plot(x_evolution_PWA.T,drawstyle='steps-post')
plt.axhline(y = MIN_x, color='red', linestyle='--', linewidth=2) 
plt.axhline(y = MAX_x, color='red', linestyle='--', linewidth=2) 
plt.grid(color='gray', linestyle='--', linewidth=0.5)
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

# Input
Fig_Input = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'Input evolution', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'$u(k)$', fontsize=18)
plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON-1)
plt.ylim(MIN_u*1.1,MAX_u*1.1)
plt.plot(u_evolution.T,drawstyle='steps-post')
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.axhline(y = MIN_u, color='red', linestyle='--', linewidth=2) 
plt.axhline(y = MAX_u, color='red', linestyle='--', linewidth=2) 
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

#  Z evolution
Fig_State_Aux = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'Auxiliary variable evolution', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'$x(k)$', fontsize=18)
plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON-1)
plt.ylim(0,1.1)
plt.plot(z_evolution.T,drawstyle='steps-post')
plt.plot(d_evolution.T,drawstyle='steps-post')
# plt.axhline(y = MIN_x, color='red', linestyle='--', linewidth=2) 
plt.axhline(y = MAX_x, color='red', linestyle='--', linewidth=2) 
plt.grid(color='gray', linestyle='--', linewidth=0.5)
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

# # Residual
Fig_Residual = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'Residual evolution', fontsize=24)
plt.xlabel(f'Iteration $i$', fontsize=18)
plt.ylabel(f'$\epsilon(i)$', fontsize=18)
plt.tight_layout()
plt.xlim(0,max_number_of_iterations)
plt.ylim(0,1.1*max(residual_evolution))
plt.plot(residual_evolution.T,drawstyle='steps-post')
plt.grid(color='gray', linestyle='--', linewidth=0.5)
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

# # Cost
Fig_Cost = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'Cost evolution', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'$J(k)$', fontsize=18)
plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON-1)
plt.ylim(0,1.1*max(Cost_Function[0]))
plt.plot(Cost_Function[0].T,drawstyle='steps-post')
plt.grid(color='gray', linestyle='--', linewidth=0.5)
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

if (PLOT_FIGURES == True) and (EXPORT_FIGURES == True):
    Fig_State.savefig(SIMULATION_FOLDER + Simumation_ID + "_State.pdf", dpi=300, bbox_inches=None)
    Fig_State.savefig(SIMULATION_FOLDER + Simumation_ID + "_State.jpeg", dpi=300, bbox_inches=None)

    Fig_State_PWA.savefig(SIMULATION_FOLDER + Simumation_ID + "_StatePWA.pdf", dpi=300, bbox_inches=None)
    Fig_State_PWA.savefig(SIMULATION_FOLDER + Simumation_ID + "_StatePWA.jpeg", dpi=300, bbox_inches=None)

    Fig_State_Aux.savefig(SIMULATION_FOLDER + Simumation_ID + "_StateAux.pdf", dpi=300, bbox_inches=None)
    Fig_State_Aux.savefig(SIMULATION_FOLDER + Simumation_ID + "_StateAux.jpeg", dpi=300, bbox_inches=None)

    Fig_Input.savefig(SIMULATION_FOLDER + Simumation_ID + "_Input.pdf", dpi=300, bbox_inches=None)
    Fig_Input.savefig(SIMULATION_FOLDER + Simumation_ID + "_Input.jpeg", dpi=300, bbox_inches=None)

    Fig_Cost.savefig(SIMULATION_FOLDER + Simumation_ID + "_Cost.pdf", dpi=300, bbox_inches=None)
    Fig_Cost.savefig(SIMULATION_FOLDER + Simumation_ID + "_Cost.jpeg", dpi=300, bbox_inches=None)

    Fig_Residual.savefig(SIMULATION_FOLDER + Simumation_ID + "_Residual.pdf", dpi=300, bbox_inches=None)
    Fig_Residual.savefig(SIMULATION_FOLDER + Simumation_ID + "_Residual.jpeg", dpi=300, bbox_inches=None)

    Fig_Ref.savefig(SIMULATION_FOLDER + Simumation_ID + "_Ref.pdf", dpi=300, bbox_inches=None)
    Fig_Ref.savefig(SIMULATION_FOLDER + Simumation_ID + "_Ref.jpeg", dpi=300, bbox_inches=None)