# from IPython import get_ipython
# get_ipython().run_line_magic('reset', '-f')

# Optimization toolbox
import cvxpy as cp
import gurobipy as gp
# For matrix operations
import numpy as np
# For file storage
import pickle as pk
# For time tracking
import time as tm
import matplotlib.pyplot as plt
from matplotlib import rc
# Enable LaTeX rendering
rc('text', usetex=True)
rc('font', family='serif')
rc('font', serif='Times New Roman')  # This line might need adjustment based on LaTeX distribution

# Support funciton
from A00_Support_Funcitons import*


# Options
PLOT_FIGURES = True
EXPORT_FIGURES = True
EXPORT_SIMULATION = True

# SIMULATION_FOLDER = "A04_Output_Data_DMPC/"
# SIMULATION_FILE_4 = "Modular_64_Partitioning_04_ALPHA_1_SimulationData"
# SIMULATION_FILE_3 = "Modular_64_Partitioning_03_ALPHA_100_SimulationData"
# SIMULATION_FILE_2 = "Modular_64_Partitioning_02_ALPHA_10000_SimulationData"
# SIMULATION_FILE_1 = "Modular_64_Partitioning_01_ALPHA_1000000_SimulationData"

SIMULATION_FOLDER = "A05_Output_Data_DMPC_Hybrid/Modular_64_Simulations/"
SIMULATION_FILE_4 = "Modular_64_Partitioning_04_ALPHA_1_SimulationData"
SIMULATION_FILE_3 = "Modular_64_Partitioning_03_ALPHA_100_SimulationData"
SIMULATION_FILE_2 = "Modular_64_Partitioning_02_ALPHA_10000_SimulationData"
SIMULATION_FILE_1 = "Modular_64_Partitioning_01_ALPHA_1000000_SimulationData"

# %% MARK: Loading solutions

# Load Solution 1

Simumation_ID = SIMULATION_FILE_1

with open(SIMULATION_FOLDER + SIMULATION_FILE_1, 'rb') as file:
    Simulation_Data_1 = pk.load(file)

Network_Dict_1 = Simulation_Data_1['Network_Dict']
Partitioning_Dict_1 = Simulation_Data_1['Partitioning_Dict']
Reference_Signal_1 = Simulation_Data_1['Reference_Signal']
x_evolution_1 = Simulation_Data_1['x_evolution']
u_evolution_1 = Simulation_Data_1['u_evolution']
Cost_Function_1 = Simulation_Data_1['Cost_Function']
residual_evolution_1 = Simulation_Data_1['residual_evolution']
epsilon_convergence_1 = Simulation_Data_1['epsilon_convergence']
Total_Computation_Time_1 = Simulation_Data_1['Total_computation_time']
Total_Core_Seconds_1 = Simulation_Data_1['Total_core_seconds']
Total_Number_ADMM_Iterations_1 = Simulation_Data_1['Total_ADMM_Iterations']
max_number_of_iterations_1 = Simulation_Data_1['max_number_of_iterations']
PARAMETERS_DICT_1 = Simulation_Data_1['PARAMETERS_DICT']

SIMULATION_HORIZON_1 = PARAMETERS_DICT_1['SIMULATION_HORIZON'] 
PREDICTION_HORIZON_1 = PARAMETERS_DICT_1['PREDICTION_HORIZON']
NUMBER_ATOMIC_AGENTS_1 = PARAMETERS_DICT_1['NUMBER_ATOMIC_AGENTS']
NUMBER_CONTROL_AGENTS_1 = Partitioning_Dict_1["Number_Of_Control_Agents"]
MAX_u_1 = PARAMETERS_DICT_1['MAX_u']
MIN_u_1 = PARAMETERS_DICT_1['MIN_u']
MAX_x_1 = PARAMETERS_DICT_1['MAX_x']
MIN_x_1 = PARAMETERS_DICT_1['MIN_x']

# Load Solution 2

Simumation_ID = SIMULATION_FILE_2

with open(SIMULATION_FOLDER + SIMULATION_FILE_2, 'rb') as file:
    Simulation_Data_2 = pk.load(file)

Network_Dict_2 = Simulation_Data_2['Network_Dict']
Partitioning_Dict_2 = Simulation_Data_2['Partitioning_Dict']
Reference_Signal_2 = Simulation_Data_2['Reference_Signal']
x_evolution_2 = Simulation_Data_2['x_evolution']
u_evolution_2 = Simulation_Data_2['u_evolution']
Cost_Function_2 = Simulation_Data_2['Cost_Function']
residual_evolution_2 = Simulation_Data_2['residual_evolution']
epsilon_convergence_2 = Simulation_Data_2['epsilon_convergence']
Total_Computation_Time_2 = Simulation_Data_2['Total_computation_time']
Total_Core_Seconds_2 = Simulation_Data_2['Total_core_seconds']
Total_Number_ADMM_Iterations_2 = Simulation_Data_2['Total_ADMM_Iterations']
max_number_of_iterations_2 = Simulation_Data_2['max_number_of_iterations']
PARAMETERS_DICT_2 = Simulation_Data_2['PARAMETERS_DICT']

SIMULATION_HORIZON_2 = PARAMETERS_DICT_2['SIMULATION_HORIZON'] 
PREDICTION_HORIZON_2 = PARAMETERS_DICT_2['PREDICTION_HORIZON']
NUMBER_ATOMIC_AGENTS_2 = PARAMETERS_DICT_2['NUMBER_ATOMIC_AGENTS']
NUMBER_CONTROL_AGENTS_2 = Partitioning_Dict_2["Number_Of_Control_Agents"]
MAX_u_2 = PARAMETERS_DICT_2['MAX_u']
MIN_u_2 = PARAMETERS_DICT_2['MIN_u']
MAX_x_2 = PARAMETERS_DICT_2['MAX_x']
MIN_x_2 = PARAMETERS_DICT_2['MIN_x']

# Load Solution 3

Simumation_ID = SIMULATION_FILE_3

with open(SIMULATION_FOLDER + SIMULATION_FILE_3, 'rb') as file:
    Simulation_Data_3 = pk.load(file)

Network_Dict_3 = Simulation_Data_3['Network_Dict']
Partitioning_Dict_3 = Simulation_Data_3['Partitioning_Dict']
Reference_Signal_3 = Simulation_Data_3['Reference_Signal']
x_evolution_3 = Simulation_Data_3['x_evolution']
u_evolution_3 = Simulation_Data_3['u_evolution']
Cost_Function_3 = Simulation_Data_3['Cost_Function']
residual_evolution_3 = Simulation_Data_3['residual_evolution']
epsilon_convergence_3 = Simulation_Data_3['epsilon_convergence']
Total_Computation_Time_3 = Simulation_Data_3['Total_computation_time']
Total_Core_Seconds_3 = Simulation_Data_3['Total_core_seconds']
Total_Number_ADMM_Iterations_3 = Simulation_Data_3['Total_ADMM_Iterations']
max_number_of_iterations_3 = Simulation_Data_3['max_number_of_iterations']
PARAMETERS_DICT_3 = Simulation_Data_3['PARAMETERS_DICT']

SIMULATION_HORIZON_3 = PARAMETERS_DICT_3['SIMULATION_HORIZON'] 
PREDICTION_HORIZON_3 = PARAMETERS_DICT_3['PREDICTION_HORIZON']
NUMBER_ATOMIC_AGENTS_3 = PARAMETERS_DICT_3['NUMBER_ATOMIC_AGENTS']
NUMBER_CONTROL_AGENTS_3 = Partitioning_Dict_3["Number_Of_Control_Agents"]
MAX_u_3 = PARAMETERS_DICT_3['MAX_u']
MIN_u_3 = PARAMETERS_DICT_3['MIN_u']
MAX_x_3 = PARAMETERS_DICT_3['MAX_x']
MIN_x_3 = PARAMETERS_DICT_3['MIN_x']

# Load Solution 4

Simumation_ID = SIMULATION_FILE_4

with open(SIMULATION_FOLDER + SIMULATION_FILE_4, 'rb') as file:
    Simulation_Data_4 = pk.load(file)

Network_Dict_4 = Simulation_Data_4['Network_Dict']
Partitioning_Dict_4 = Simulation_Data_4['Partitioning_Dict']
Reference_Signal_4 = Simulation_Data_4['Reference_Signal']
x_evolution_4 = Simulation_Data_4['x_evolution']
u_evolution_4 = Simulation_Data_4['u_evolution']
Cost_Function_4 = Simulation_Data_4['Cost_Function']
residual_evolution_4 = Simulation_Data_4['residual_evolution']
epsilon_convergence_4 = Simulation_Data_4['epsilon_convergence']
Total_Computation_Time_4 = Simulation_Data_4['Total_computation_time']
Total_Core_Seconds_4 = Simulation_Data_4['Total_core_seconds']
Total_Number_ADMM_Iterations_4 = Simulation_Data_4['Total_ADMM_Iterations']
max_number_of_iterations_4 = Simulation_Data_4['max_number_of_iterations']
PARAMETERS_DICT_4 = Simulation_Data_4['PARAMETERS_DICT']

SIMULATION_HORIZON_4 = PARAMETERS_DICT_4['SIMULATION_HORIZON'] 
PREDICTION_HORIZON_4 = PARAMETERS_DICT_4['PREDICTION_HORIZON']
NUMBER_ATOMIC_AGENTS_4 = PARAMETERS_DICT_4['NUMBER_ATOMIC_AGENTS']
NUMBER_CONTROL_AGENTS_4 = Partitioning_Dict_4["Number_Of_Control_Agents"]
MAX_u_4 = PARAMETERS_DICT_4['MAX_u']
MIN_u_4 = PARAMETERS_DICT_4['MIN_u']
MAX_x_4 = PARAMETERS_DICT_4['MAX_x']
MIN_x_4 = PARAMETERS_DICT_4['MIN_x']

# %% MARK: Evaluation metrics
# print("Number of control agents: ", NUMBER_CONTROL_AGENTS_1)
# print("Total cost: ", Cost_Function_1[SIMULATION_HORIZON_1-1])
# print("Total computation time: ", Total_Computation_Time_1)
# print("Total core seconds: ", Total_Core_Seconds_1)
# print("Total ADMM Iterations: ", Total_Number_ADMM_Iterations_1)

# %% MARK: Plot Comparison

# Total Cost comparison linear
# Cost_Function_Vector = np.array([Cost_Function_4[-2], Cost_Function_3[-2], Cost_Function_2[-2], Cost_Function_1[-2]])
Cores_Vector = np.array([1, 4, 16, 64])

# PWA
Cost_Function_Vector = np.array([Cost_Function_4[0][-2], Cost_Function_3[0][-2], Cost_Function_2[0][-2], Cost_Function_1[0][-2]])

Fig_Com = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'Cost functions comparison', fontsize=24)
plt.xlabel(f'Partiton', fontsize=18)
plt.ylabel(f'$J$', fontsize=18)
# plt.tight_layout()
# plt.xlim(0,SIMULATION_HORIZON_1-1)
# plt.ylim(1.1*min(Cost_Function_Vector),1.1*max(Cost_Function_Vector))
plt.yscale("log")
plt.plot(Cost_Function_Vector, marker='s', markersize=10,  
         markerfacecolor='red',         
         markeredgecolor='blue', 
         markeredgewidth=2.5,
         linestyle='--', color='lightgrey', linewidth=2)
ax_fig.set_xticks([0, 1, 2, 3])  # Define positions of y-axis ticks
ax_fig.set_xticklabels([f'$\mathcal{{P}}_4$', f'$\mathcal{{P}}_3$', f'$\mathcal{{P}}_2$', f'$\mathcal{{P}}_1$'], fontsize=18)  # Define custom labels
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.grid(color='gray', linestyle='--', linewidth=0.5)
# plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

# Total time comparison
Total_Computation_Time_Vector = np.array([Total_Computation_Time_4, Total_Computation_Time_3, Total_Computation_Time_2, Total_Computation_Time_1])

Fig_Sec = plt.figure(figsize=(9, 3.1))
ax_fig = plt.gca()
plt.title(f'Computation time comparison', fontsize=24)
plt.xlabel(f'Number of cores', fontsize=18)
plt.ylabel(f'Seconds [s]', fontsize=18)
# plt.tight_layout()
# plt.xlim(0,SIMULATION_HORIZON_1-1)
# plt.ylim(1.1*min(Cost_Function_Vector),1.1*max(Cost_Function_Vector))
# plt.yscale("log")
plt.plot(Cores_Vector, Total_Computation_Time_Vector, marker='s', markersize=10,  
         markerfacecolor='red',         
         markeredgecolor='blue', 
         markeredgewidth=2.5,
         linestyle='--', color='lightgrey', linewidth=2)
ax_fig.set_xticks(Cores_Vector )  # Define positions of y-axis ticks
ax_fig.set_xticklabels(["1\n" f'$\mathcal{{P}}_4$', "4\n"f'$\mathcal{{P}}_3$', "16\n"f'$\mathcal{{P}}_2$', "64\n"f'$\mathcal{{P}}_1$'], fontsize=18)  # Define custom labels
# ax_fig.text(7, 0.8, "Legend Style Annotation\n- Sine (Blue)\n- Cosine (Orange)",
#         fontsize=12, bbox=dict(facecolor='white', edgecolor='black', boxstyle='round,pad=0.5'))

# ax_fig.set_xticklabels([f'$\mathcal{{P}}_4: 1$', f'$\mathcal{{P}}_3: 4$', f'$\mathcal{{P}}_2: 16$', f'$\mathcal{{P}}_1: 64$'], fontsize=18)  # Define custom labels
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.subplots_adjust(top=0.88, bottom= 0.3, right=0.95, left = 0.15)
plt.show()

# Core seconds comparison
Total_Core_Seconds_Vector = np.array([Total_Core_Seconds_4, Total_Core_Seconds_3, Total_Core_Seconds_2, Total_Core_Seconds_1])

Fig_Cor = plt.figure(figsize=(9, 3.1))
ax_fig = plt.gca()
plt.title(f'Computation cost comparison', fontsize=24)
plt.xlabel(f'Number of cores', fontsize=18)
plt.ylabel(f'Core Seconds [s]', fontsize=18)
# plt.tight_layout()
# plt.xlim(0,SIMULATION_HORIZON_1-1)
# plt.ylim(1.1*min(Cost_Function_Vector),1.1*max(Cost_Function_Vector))
# plt.yscale("log")
plt.plot(Cores_Vector , Total_Core_Seconds_Vector, marker='s', markersize=10,  
         markerfacecolor='red',         
         markeredgecolor='blue', 
         markeredgewidth=2.5,
         linestyle='--', color='lightgrey', linewidth=2)
ax_fig.set_xticks(Cores_Vector )  # Define positions of y-axis ticks
ax_fig.set_xticklabels(["1\n" f'$\mathcal{{P}}_4$', "4\n"f'$\mathcal{{P}}_3$', "16\n"f'$\mathcal{{P}}_2$', "64\n"f'$\mathcal{{P}}_1$'], fontsize=18)  # Define custom labels
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.subplots_adjust(top=0.88, bottom= 0.3, right=0.95, left = 0.15)
plt.show()

# Residual evolution
Fig_Res = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'Residual evolution', fontsize=24)
plt.xlabel(f'ADMM iteration $i$', fontsize=18)
plt.ylabel(f'$r(i)$', fontsize=18)
# plt.tight_layout()
plt.xlim(0,25)
plt.ylim(10e-2,10e1)
plt.yscale("log")
plt.plot(residual_evolution_1.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}_1$')
plt.plot(residual_evolution_2.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}_2$')
plt.plot(residual_evolution_3.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}_3$')
# plt.plot(residual_evolution_4.T,drawstyle='steps-post')
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.legend(loc='upper right', fontsize = 18)
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

# %% MARK: Plot CMPC simulation

# State
Fig_Ref = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'Reference evolution', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'$r(k)$', fontsize=18)
# plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON_4-1)
plt.ylim(1.1*min(Reference_Signal_4[0,:]),1.1*max(Reference_Signal_4[0,:]))
plt.plot(Reference_Signal_4.T,drawstyle='steps-post', linewidth=2)
plt.grid(color='gray', linestyle='--', linewidth=0.5)
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

# State
Fig_State = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'State evolution', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'$x(k)$', fontsize=18)
# plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON_4-1)
# plt.ylim(MIN_x_4*1.1,MAX_x_4*1.1)
plt.ylim(1.1*min(Reference_Signal_4[0,:]),1.1*max(Reference_Signal_4[0,:]))
# plt.plot(Reference_Signal_4.T,drawstyle='steps-post', color='red', label = f"r(k)")
plt.plot(x_evolution_2.T,drawstyle='steps-post', color='blue', label = f"x(k)", linewidth=2)
plt.axhline(y = MIN_x_4, color='red', linestyle='--', linewidth=2) 
plt.axhline(y = MAX_x_4, color='red', linestyle='--', linewidth=2) 
plt.grid(color='gray', linestyle='--', linewidth=0.5)
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
# plt.legend(loc='upper right', fontsize = 18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

# Input
Fig_Input = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'Input evolution', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'$u(k)$', fontsize=18)
# plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON_4-1)
plt.ylim(MIN_u_4*1.1,MAX_u_4*1.1)
plt.plot(u_evolution_2.T,drawstyle='steps-post')
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.axhline(y = MIN_u_4, color='red', linestyle='--', linewidth=2) 
plt.axhline(y = MAX_u_4, color='red', linestyle='--', linewidth=2) 
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

# # # Residual
# Fig_Residual = plt.figure(figsize=(9, 3))
# plt.title(f'Residual evolution', fontsize=24)
# plt.xlabel(f'Iteration $i$', fontsize=18)
# plt.ylabel(f'$\epsilon(i)$', fontsize=18)
# plt.tight_layout()
# plt.xlim(0,max_number_of_iterations_4)
# plt.ylim(0,1.1*max(residual_evolution))
# plt.plot(residual_evolution.T,drawstyle='steps-post')
# plt.grid(color='gray', linestyle='--', linewidth=0.5)
# plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

# # Cost linear
Fig_Cost = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'Cost evolution', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'$J(k)$', fontsize=18)
# plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON_4-1)
#  For Linear Systems
# plt.ylim(0,1.1*max(Cost_Function_4))
# plt.plot(Cost_Function_4.T,drawstyle='steps-post', linewidth=2)
#  For hybrid systems
plt.ylim(0,1.1*max(Cost_Function_4[0]))
plt.plot(Cost_Function_2[0].T,drawstyle='steps-post', linewidth=2)
plt.grid(color='gray', linestyle='--', linewidth=0.5)
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

if (PLOT_FIGURES == True) and (EXPORT_FIGURES == True):
    Fig_State.savefig(SIMULATION_FOLDER + Simumation_ID + "_State.pdf", dpi=300, bbox_inches=None)
    Fig_State.savefig(SIMULATION_FOLDER + Simumation_ID + "_State.jpeg", dpi=300, bbox_inches=None)

    Fig_Input.savefig(SIMULATION_FOLDER + Simumation_ID + "_Input.pdf", dpi=300, bbox_inches=None)
    Fig_Input.savefig(SIMULATION_FOLDER + Simumation_ID + "_Input.jpeg", dpi=300, bbox_inches=None)

    Fig_Cost.savefig(SIMULATION_FOLDER + Simumation_ID + "_Cost.pdf", dpi=300, bbox_inches=None)
    Fig_Cost.savefig(SIMULATION_FOLDER + Simumation_ID + "_Cost.jpeg", dpi=300, bbox_inches=None)

    Fig_Com.savefig(SIMULATION_FOLDER + Simumation_ID + "_Comparison.pdf", dpi=300, bbox_inches=None)
    Fig_Com.savefig(SIMULATION_FOLDER + Simumation_ID + "_Comparison.jpeg", dpi=300, bbox_inches=None)
    Fig_Sec.savefig(SIMULATION_FOLDER + Simumation_ID + "_ComputationTime.pdf", dpi=300, bbox_inches=None)
    Fig_Sec.savefig(SIMULATION_FOLDER + Simumation_ID + "_ComputationTime.jpeg", dpi=300, bbox_inches=None)
    Fig_Cor.savefig(SIMULATION_FOLDER + Simumation_ID + "_CoreSeconds.pdf", dpi=300, bbox_inches=None)
    Fig_Cor.savefig(SIMULATION_FOLDER + Simumation_ID + "_CoreSeconds.jpeg", dpi=300, bbox_inches=None)
    Fig_Res.savefig(SIMULATION_FOLDER + Simumation_ID + "_ResidualEvolution.pdf", dpi=300, bbox_inches=None)
    Fig_Res.savefig(SIMULATION_FOLDER + Simumation_ID + "_ResidualEvolution.jpeg", dpi=300, bbox_inches=None)

    # Fig_Residual.savefig(SIMULATION_FOLDER + Simumation_ID + "_Residual.pdf", dpi=300)
    # Fig_Residual.savefig(SIMULATION_FOLDER + Simumation_ID + "_Residual.jpeg", dpi=300)

    Fig_Ref.savefig(SIMULATION_FOLDER + Simumation_ID + "_Ref.pdf", dpi=300, bbox_inches=None)
    Fig_Ref.savefig(SIMULATION_FOLDER + Simumation_ID + "_Ref.jpeg", dpi=300, bbox_inches=None)