# %% 
# MARK: Reset Operations
from IPython import get_ipython
get_ipython().run_line_magic('reset', '-f')
# print("\n" * 100)

# %%
# MARK: Import libraries


# For matrix operations
import numpy as np
# For file storage
import pickle as pk
# For time tracking
import time as tm
# For data location
INPUT_FOLDER = "Input_Data"
OUTPUT_FOLDER = "A09_Output_Data/Modular_64"
import os
if not os.path.exists(OUTPUT_FOLDER):
    os.mkdir(OUTPUT_FOLDER)


import networkx as nx
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib import rc
# Enable LaTeX rendering
rc('text', usetex=True)

# Set the font to Times New Roman or the closest available font in LaTeX
rc('font', family='serif')
rc('font', serif='Times New Roman')  # This line might need adjustment based on LaTeX distribution

from scipy.spatial import ConvexHull
import matplotlib.colors as mcolors

# Support functions
from A00_Support_Funcitons import *
from A02_Weighting_Function import*

# Settings
EXPORT_DATA = True
DRAW_PLOTS = True
EXPORT_PLOTS = False


# %%
# MARK: Select network

# Network_Dict = modular64()
# Network_File = "Random_100"
# Network_File = "Random_50"
Network_File = "Modular_64"
# Network_File = "Random_10"
with open("Input_Data/" + Network_File, 'rb') as f:
    Network_Dict = pk.load(f)

A = Network_Dict['A']
B = Network_Dict['B']
Weighted_Adjacency_Matrix = Network_Dict['Weighted_Adjacency_Matrix']
NUMBER_OF_ATOMIC_AGENTS = Network_Dict['NUMBER_OF_ATOMIC_AGENTS']
List_Atomic_Agents = []
for i in range(NUMBER_OF_ATOMIC_AGENTS):
    List_Atomic_Agents.append(i)
# print("test")

Partition_Name = "Test_1"

Partition_ID = Network_File + "_" + Partition_Name


# %% MARK: Partition refinement function

def Partition_Refinement(Network_Dict, List_Atomic_Agents, Partition, ALPHA):
    

    W_Partition = weighting_function(Network_Dict, List_Atomic_Agents, Partition, ALPHA)

    Delta_Partition_Max = 0

    while Delta_Partition_Max >= 0:

        Next_Agent = None
        Next_Set = None
        Prev_Set = None
        for i in range(len(Partition)):
            for j in range(len(Partition[i])):
                Test_Agent = Partition[i][j]
                Partition[i].remove(Test_Agent)
                for k in range(len(Partition)):
                    if k != i:
                        Partition[k].append(Test_Agent)
                        W_Partition_Test = weighting_function(Network_Dict, List_Atomic_Agents, Partition, ALPHA)
                        Delta_Partition =  W_Partition_Test - W_Partition 
                        if Delta_Partition > Delta_Partition_Max:
                            Prev_Set = i
                            Next_Agent = Test_Agent
                            Next_Set = k
                        Partition[k].remove(Test_Agent)
                Partition[i].append(Test_Agent)
        
        if Next_Agent != None:
            Partition[Prev_Set].remove(Next_Agent)
            Partition[Next_Set].append(Next_Agent)
            W_Partition_New = weighting_function(Network_Dict, List_Atomic_Agents, Partition, ALPHA)
            Delta_Partition_Print = W_Partition_New - W_Partition
            W_Partition = W_Partition_New
            print("Local Partition refinement")
            print("Assign agent ", Next_Agent, " to set ", Next_Set)
            print("Delta Partition:", Delta_Partition_Print )
            print("Partition weight ", W_Partition)           
        else:
            Delta_Partition_Max = -1       
    
    return Partition

# %%
# MARK: Partitioning algorithm

Agents_Left = List_Atomic_Agents.copy()

# Allocation of the empty partitioning
Partition = create_empty_list(NUMBER_OF_ATOMIC_AGENTS)

# %% MARK: Manual definition of the partition

Partition[0] = [1, 2, 3]
Partition[1] = [4, 7, 10, 13]
Partition[2] = [5, 6, 8]
Partition[3] = [9, 11, 12]
Partition[4] = [14, 15]
Partition[5] = [16, 27, 38, 49]
Partition[6] = [25, 28]
Partition[7] = [17, 18, 19]
Partition[8] = [21, 22, 24]
Partition[9] = [20, 23, 26, 29]
Partition[10] = [30, 31, 32]
Partition[11] = [37, 40]
Partition[12] = [33, 34, 35]
Partition[13] = [36, 39, 42, 45]
Partition[14] = [41, 43, 44]
Partition[15] = [46, 47, 48]
Partition[16] = [50, 51]
Partition[17] = [52, 55, 58, 61]
Partition[18] = [53, 54, 56]
Partition[19] = [62, 63, 64]
Partition[20] = [57, 59, 60]

# Remove empty sets
Partition = [item for item in Partition if item != []]
NUMBER_CONTROL_AGENTS = len(Partition)

# Sort list
for idx in range(NUMBER_CONTROL_AGENTS):
    Partition[idx].sort()

for idx_i in range(NUMBER_CONTROL_AGENTS):
    for idx_j in range(len(Partition[idx_i])):
        Partition[idx_i][idx_j] = Partition[idx_i][idx_j] - 1 

# %% Display result

print("Number atomic control agents: ", NUMBER_OF_ATOMIC_AGENTS)
print("Number atomic control agents: ", NUMBER_CONTROL_AGENTS)
for idx in range(NUMBER_CONTROL_AGENTS):
    print("Control agent ", idx, ": ", Partition[idx])

# %%
# MARK: Store results
Computation_Time_Total = None
ALPHA = 0

Output_Dict = {"A": A, 
               "B": B, 
               "WEIGHTED_ADJACENCY_MATRIX": Weighted_Adjacency_Matrix,
               "ALPHA": ALPHA, 
               "NUMBER_ATOMIC_AGENTS": NUMBER_OF_ATOMIC_AGENTS,
               "List_Atomic_Agents": List_Atomic_Agents,
               "Control_Agents_List": Partition,
               "Number_Of_Control_Agents": NUMBER_CONTROL_AGENTS,
               "Computation_Time_Total": Computation_Time_Total}

if EXPORT_DATA == True:
    with open(OUTPUT_FOLDER + "/" + Partition_ID , 'wb') as Write_File:
        pk.dump(Output_Dict, Write_File)
print(Partition)
Partition_Holder = Partition.copy()

