# from IPython import get_ipython
# get_ipython().run_line_magic('reset', '-f')

# Optimization toolbox
import cvxpy as cp
import gurobipy as gp
# For matrix operations
import numpy as np
# For file storage
import pickle as pk
# For time tracking
import time as tm
import matplotlib.pyplot as plt
from matplotlib import rc
# Enable LaTeX rendering
rc('text', usetex=True)
rc('font', family='serif')
rc('font', serif='Times New Roman')  # This line might need adjustment based on LaTeX distribution

# Support funciton
from A00_Support_Funcitons import*


# Options
PLOT_FIGURES = True
EXPORT_FIGURES = True
EXPORT_SIMULATION = True

SIMULATION_FOLDER = "A05_Output_Data_DMPC_Hybrid/Random_50_Simulations/"
SIMULATION_FILE_1 = "Random_50_ALPHA_4304310.87298206_SimulationData" # Partition 7 - 50 agents
SIMULATION_FILE_2 = "Random_50_ALPHA_43043.1087298206_SimulationData" # Partition 6 - 15 agents
SIMULATION_FILE_3 = "Random_50_ALPHA_4304.31087298206_SimulationData" # Partition 2 - 6 agents
SIMULATION_FILE_4 = "Random_50_ALPHA_430.43108729820597_SimulationData" # Partition 5 - 2 agents
SIMULATION_FILE_5 = "Random_50_ALPHA_43.043108729820595_SimulationData" # Partition 3 - 1 agent
SIMULATION_FILE_6 = "Random_50_Partition_02_ALPHA_30_SimulationData" # Partition 2 ALG - 11 agents
SIMULATION_FILE_7 = "Random_50_Partition_03_ALPHA_20_SimulationData" # Partition 3 ALG - 9 agents
SIMULATION_FILE_8 = "Random_50_Partition_04_ALPHA_10_SimulationData" # Partition 4 - 5 agents
SIMULATION_FILE_9 = "Random_50_Partition_05_ALPHA_1_SimulationData" # Partition 5 - 4 agents


# %% MARK: Loading solutions

# Load Solution_1

Simumation_ID = SIMULATION_FILE_1

with open(SIMULATION_FOLDER + SIMULATION_FILE_1, 'rb') as file:
    Simulation_Data_1 = pk.load(file)

Network_Dict_1 = Simulation_Data_1['Network_Dict']
Partitioning_Dict_1 = Simulation_Data_1['Partitioning_Dict']
Reference_Signal_1 = Simulation_Data_1['Reference_Signal']
x_evolution_1 = Simulation_Data_1['x_evolution']
u_evolution_1 = Simulation_Data_1['u_evolution']
Cost_Function_1 = Simulation_Data_1['Cost_Function']
residual_evolution_1 = Simulation_Data_1['residual_evolution']
epsilon_convergence_1 = Simulation_Data_1['epsilon_convergence']
Total_Computation_Time_1 = Simulation_Data_1['Total_computation_time']
Total_Core_Seconds_1 = Simulation_Data_1['Total_core_seconds']
Total_Number_ADMM_Iterations_1 = Simulation_Data_1['Total_ADMM_Iterations']
max_number_of_iterations_1 = Simulation_Data_1['max_number_of_iterations']
PARAMETERS_DICT_1 = Simulation_Data_1['PARAMETERS_DICT']

SIMULATION_HORIZON_1 = PARAMETERS_DICT_1['SIMULATION_HORIZON'] 
PREDICTION_HORIZON_1 = PARAMETERS_DICT_1['PREDICTION_HORIZON']
NUMBER_ATOMIC_AGENTS_1 = PARAMETERS_DICT_1['NUMBER_ATOMIC_AGENTS']
NUMBER_CONTROL_AGENTS_1 = Partitioning_Dict_1["Number_Of_Control_Agents"]
MAX_u_1 = PARAMETERS_DICT_1['MAX_u']
MIN_u_1 = PARAMETERS_DICT_1['MIN_u']
MAX_x_1 = PARAMETERS_DICT_1['MAX_x']
MIN_x_1 = PARAMETERS_DICT_1['MIN_x']

# Load Solution 2

Simumation_ID = SIMULATION_FILE_2

with open(SIMULATION_FOLDER + SIMULATION_FILE_2, 'rb') as file:
    Simulation_Data_2 = pk.load(file)

Network_Dict_2 = Simulation_Data_2['Network_Dict']
Partitioning_Dict_2 = Simulation_Data_2['Partitioning_Dict']
Reference_Signal_2 = Simulation_Data_2['Reference_Signal']
x_evolution_2 = Simulation_Data_2['x_evolution']
u_evolution_2 = Simulation_Data_2['u_evolution']
Cost_Function_2 = Simulation_Data_2['Cost_Function']
residual_evolution_2 = Simulation_Data_2['residual_evolution']
epsilon_convergence_2 = Simulation_Data_2['epsilon_convergence']
Total_Computation_Time_2 = Simulation_Data_2['Total_computation_time']
Total_Core_Seconds_2 = Simulation_Data_2['Total_core_seconds']
Total_Number_ADMM_Iterations_2 = Simulation_Data_2['Total_ADMM_Iterations']
max_number_of_iterations_2 = Simulation_Data_2['max_number_of_iterations']
PARAMETERS_DICT_2 = Simulation_Data_2['PARAMETERS_DICT']

SIMULATION_HORIZON_2 = PARAMETERS_DICT_2['SIMULATION_HORIZON'] 
PREDICTION_HORIZON_2 = PARAMETERS_DICT_2['PREDICTION_HORIZON']
NUMBER_ATOMIC_AGENTS_2 = PARAMETERS_DICT_2['NUMBER_ATOMIC_AGENTS']
NUMBER_CONTROL_AGENTS_2 = Partitioning_Dict_2["Number_Of_Control_Agents"]
MAX_u_2 = PARAMETERS_DICT_2['MAX_u']
MIN_u_2 = PARAMETERS_DICT_2['MIN_u']
MAX_x_2 = PARAMETERS_DICT_2['MAX_x']
MIN_x_2 = PARAMETERS_DICT_2['MIN_x']

# Load Solution 3

Simumation_ID = SIMULATION_FILE_3

with open(SIMULATION_FOLDER + SIMULATION_FILE_3, 'rb') as file:
    Simulation_Data_3 = pk.load(file)

Network_Dict_3 = Simulation_Data_3['Network_Dict']
Partitioning_Dict_3 = Simulation_Data_3['Partitioning_Dict']
Reference_Signal_3 = Simulation_Data_3['Reference_Signal']
x_evolution_3 = Simulation_Data_3['x_evolution']
u_evolution_3 = Simulation_Data_3['u_evolution']
Cost_Function_3 = Simulation_Data_3['Cost_Function']
residual_evolution_3 = Simulation_Data_3['residual_evolution']
epsilon_convergence_3 = Simulation_Data_3['epsilon_convergence']
Total_Computation_Time_3 = Simulation_Data_3['Total_computation_time']
Total_Core_Seconds_3 = Simulation_Data_3['Total_core_seconds']
Total_Number_ADMM_Iterations_3 = Simulation_Data_3['Total_ADMM_Iterations']
max_number_of_iterations_3 = Simulation_Data_3['max_number_of_iterations']
PARAMETERS_DICT_3 = Simulation_Data_3['PARAMETERS_DICT']

SIMULATION_HORIZON_3 = PARAMETERS_DICT_3['SIMULATION_HORIZON'] 
PREDICTION_HORIZON_3 = PARAMETERS_DICT_3['PREDICTION_HORIZON']
NUMBER_ATOMIC_AGENTS_3 = PARAMETERS_DICT_3['NUMBER_ATOMIC_AGENTS']
NUMBER_CONTROL_AGENTS_3 = Partitioning_Dict_3["Number_Of_Control_Agents"]
MAX_u_3 = PARAMETERS_DICT_3['MAX_u']
MIN_u_3 = PARAMETERS_DICT_3['MIN_u']
MAX_x_3 = PARAMETERS_DICT_3['MAX_x']
MIN_x_3 = PARAMETERS_DICT_3['MIN_x']

# Load Solution 4

Simumation_ID = SIMULATION_FILE_4

with open(SIMULATION_FOLDER + SIMULATION_FILE_4, 'rb') as file:
    Simulation_Data_4 = pk.load(file)

Network_Dict_4 = Simulation_Data_4['Network_Dict']
Partitioning_Dict_4 = Simulation_Data_4['Partitioning_Dict']
Reference_Signal_4 = Simulation_Data_4['Reference_Signal']
x_evolution_4 = Simulation_Data_4['x_evolution']
u_evolution_4 = Simulation_Data_4['u_evolution']
Cost_Function_4 = Simulation_Data_4['Cost_Function']
residual_evolution_4 = Simulation_Data_4['residual_evolution']
epsilon_convergence_4 = Simulation_Data_4['epsilon_convergence']
Total_Computation_Time_4 = Simulation_Data_4['Total_computation_time']
Total_Core_Seconds_4 = Simulation_Data_4['Total_core_seconds']
Total_Number_ADMM_Iterations_4 = Simulation_Data_4['Total_ADMM_Iterations']
max_number_of_iterations_4 = Simulation_Data_4['max_number_of_iterations']
PARAMETERS_DICT_4 = Simulation_Data_4['PARAMETERS_DICT']

SIMULATION_HORIZON_4 = PARAMETERS_DICT_4['SIMULATION_HORIZON'] 
PREDICTION_HORIZON_4 = PARAMETERS_DICT_4['PREDICTION_HORIZON']
NUMBER_ATOMIC_AGENTS_4 = PARAMETERS_DICT_4['NUMBER_ATOMIC_AGENTS']
NUMBER_CONTROL_AGENTS_4 = Partitioning_Dict_4["Number_Of_Control_Agents"]
MAX_u_4 = PARAMETERS_DICT_4['MAX_u']
MIN_u_4 = PARAMETERS_DICT_4['MIN_u']
MAX_x_4 = PARAMETERS_DICT_4['MAX_x']
MIN_x_4 = PARAMETERS_DICT_4['MIN_x']


# Load Solution_5

Simumation_ID = SIMULATION_FILE_5

with open(SIMULATION_FOLDER + SIMULATION_FILE_5, 'rb') as file:
    Simulation_Data_5 = pk.load(file)

Network_Dict_5 = Simulation_Data_5['Network_Dict']
Partitioning_Dict_5 = Simulation_Data_5['Partitioning_Dict']
Reference_Signal_5 = Simulation_Data_5['Reference_Signal']
x_evolution_5 = Simulation_Data_5['x_evolution']
u_evolution_5 = Simulation_Data_5['u_evolution']
Cost_Function_5 = Simulation_Data_5['Cost_Function']
residual_evolution_5 = Simulation_Data_5['residual_evolution']
epsilon_convergence_5 = Simulation_Data_5['epsilon_convergence']
Total_Computation_Time_5 = Simulation_Data_5['Total_computation_time']
Total_Core_Seconds_5 = Simulation_Data_5['Total_core_seconds']
Total_Number_ADMM_Iterations_5 = Simulation_Data_5['Total_ADMM_Iterations']
max_number_of_iterations_5 = Simulation_Data_5['max_number_of_iterations']
PARAMETERS_DICT_5 = Simulation_Data_5['PARAMETERS_DICT']

SIMULATION_HORIZON_5 = PARAMETERS_DICT_5['SIMULATION_HORIZON'] 
PREDICTION_HORIZON_5 = PARAMETERS_DICT_5['PREDICTION_HORIZON']
NUMBER_ATOMIC_AGENTS_5 = PARAMETERS_DICT_5['NUMBER_ATOMIC_AGENTS']
NUMBER_CONTROL_AGENTS_5 = Partitioning_Dict_5["Number_Of_Control_Agents"]
MAX_u_5 = PARAMETERS_DICT_5['MAX_u']
MIN_u_5 = PARAMETERS_DICT_5['MIN_u']
MAX_x_5 = PARAMETERS_DICT_5['MAX_x']
MIN_x_5 = PARAMETERS_DICT_5['MIN_x']

# Load Solution_6

Simumation_ID = SIMULATION_FILE_6

with open(SIMULATION_FOLDER + SIMULATION_FILE_6, 'rb') as file:
    Simulation_Data_6 = pk.load(file)

Network_Dict_6 = Simulation_Data_6['Network_Dict']
Partitioning_Dict_6 = Simulation_Data_6['Partitioning_Dict']
Reference_Signal_6 = Simulation_Data_6['Reference_Signal']
x_evolution_6 = Simulation_Data_6['x_evolution']
u_evolution_6 = Simulation_Data_6['u_evolution']
Cost_Function_6 = Simulation_Data_6['Cost_Function']
residual_evolution_6 = Simulation_Data_6['residual_evolution']
epsilon_convergence_6 = Simulation_Data_6['epsilon_convergence']
Total_Computation_Time_6 = Simulation_Data_6['Total_computation_time']
Total_Core_Seconds_6 = Simulation_Data_6['Total_core_seconds']
Total_Number_ADMM_Iterations_6 = Simulation_Data_6['Total_ADMM_Iterations']
max_number_of_iterations_6 = Simulation_Data_6['max_number_of_iterations']
PARAMETERS_DICT_6 = Simulation_Data_6['PARAMETERS_DICT']

SIMULATION_HORIZON_6 = PARAMETERS_DICT_6['SIMULATION_HORIZON'] 
PREDICTION_HORIZON_6 = PARAMETERS_DICT_6['PREDICTION_HORIZON']
NUMBER_ATOMIC_AGENTS_6 = PARAMETERS_DICT_6['NUMBER_ATOMIC_AGENTS']
NUMBER_CONTROL_AGENTS_6 = Partitioning_Dict_6["NUMBER_CONTROL_AGENTS"]
MAX_u_6 = PARAMETERS_DICT_6['MAX_u']
MIN_u_6 = PARAMETERS_DICT_6['MIN_u']
MAX_x_6 = PARAMETERS_DICT_6['MAX_x']
MIN_x_6 = PARAMETERS_DICT_6['MIN_x']

# Load Solution_7

Simumation_ID = SIMULATION_FILE_7

with open(SIMULATION_FOLDER + SIMULATION_FILE_7, 'rb') as file:
    Simulation_Data_7 = pk.load(file)

Network_Dict_7 = Simulation_Data_7['Network_Dict']
Partitioning_Dict_7 = Simulation_Data_7['Partitioning_Dict']
Reference_Signal_7 = Simulation_Data_7['Reference_Signal']
x_evolution_7 = Simulation_Data_7['x_evolution']
u_evolution_7 = Simulation_Data_7['u_evolution']
Cost_Function_7 = Simulation_Data_7['Cost_Function']
residual_evolution_7 = Simulation_Data_7['residual_evolution']
epsilon_convergence_7 = Simulation_Data_7['epsilon_convergence']
Total_Computation_Time_7 = Simulation_Data_7['Total_computation_time']
Total_Core_Seconds_7 = Simulation_Data_7['Total_core_seconds']
Total_Number_ADMM_Iterations_7 = Simulation_Data_7['Total_ADMM_Iterations']
max_number_of_iterations_7 = Simulation_Data_7['max_number_of_iterations']
PARAMETERS_DICT_7 = Simulation_Data_7['PARAMETERS_DICT']

SIMULATION_HORIZON_7 = PARAMETERS_DICT_7['SIMULATION_HORIZON'] 
PREDICTION_HORIZON_7 = PARAMETERS_DICT_7['PREDICTION_HORIZON']
NUMBER_ATOMIC_AGENTS_7 = PARAMETERS_DICT_7['NUMBER_ATOMIC_AGENTS']
NUMBER_CONTROL_AGENTS_7 = Partitioning_Dict_7["NUMBER_CONTROL_AGENTS"]
MAX_u_7 = PARAMETERS_DICT_7['MAX_u']
MIN_u_7 = PARAMETERS_DICT_7['MIN_u']
MAX_x_7 = PARAMETERS_DICT_7['MAX_x']
MIN_x_7 = PARAMETERS_DICT_7['MIN_x']

# Load Solution_8

Simumation_ID = SIMULATION_FILE_8

with open(SIMULATION_FOLDER + SIMULATION_FILE_8, 'rb') as file:
    Simulation_Data_8 = pk.load(file)

Network_Dict_8 = Simulation_Data_8['Network_Dict']
Partitioning_Dict_8 = Simulation_Data_8['Partitioning_Dict']
Reference_Signal_8 = Simulation_Data_8['Reference_Signal']
x_evolution_8 = Simulation_Data_8['x_evolution']
u_evolution_8 = Simulation_Data_8['u_evolution']
Cost_Function_8 = Simulation_Data_8['Cost_Function']
residual_evolution_8 = Simulation_Data_8['residual_evolution']
epsilon_convergence_8 = Simulation_Data_8['epsilon_convergence']
Total_Computation_Time_8 = Simulation_Data_8['Total_computation_time']
Total_Core_Seconds_8 = Simulation_Data_8['Total_core_seconds']
Total_Number_ADMM_Iterations_8 = Simulation_Data_8['Total_ADMM_Iterations']
max_number_of_iterations_8 = Simulation_Data_8['max_number_of_iterations']
PARAMETERS_DICT_8 = Simulation_Data_8['PARAMETERS_DICT']

SIMULATION_HORIZON_8 = PARAMETERS_DICT_8['SIMULATION_HORIZON'] 
PREDICTION_HORIZON_8 = PARAMETERS_DICT_8['PREDICTION_HORIZON']
NUMBER_ATOMIC_AGENTS_8 = PARAMETERS_DICT_8['NUMBER_ATOMIC_AGENTS']
NUMBER_CONTROL_AGENTS_8 = Partitioning_Dict_8["NUMBER_CONTROL_AGENTS"]
MAX_u_8 = PARAMETERS_DICT_8['MAX_u']
MIN_u_8 = PARAMETERS_DICT_8['MIN_u']
MAX_x_8 = PARAMETERS_DICT_8['MAX_x']
MIN_x_8 = PARAMETERS_DICT_8['MIN_x']

# Load Solution_9

Simumation_ID = SIMULATION_FILE_9

with open(SIMULATION_FOLDER + SIMULATION_FILE_9, 'rb') as file:
    Simulation_Data_9 = pk.load(file)

Network_Dict_9 = Simulation_Data_9['Network_Dict']
Partitioning_Dict_9 = Simulation_Data_9['Partitioning_Dict']
Reference_Signal_9 = Simulation_Data_9['Reference_Signal']
x_evolution_9 = Simulation_Data_9['x_evolution']
u_evolution_9 = Simulation_Data_9['u_evolution']
Cost_Function_9 = Simulation_Data_9['Cost_Function']
residual_evolution_9 = Simulation_Data_9['residual_evolution']
epsilon_convergence_9 = Simulation_Data_9['epsilon_convergence']
Total_Computation_Time_9 = Simulation_Data_9['Total_computation_time']
Total_Core_Seconds_9 = Simulation_Data_9['Total_core_seconds']
Total_Number_ADMM_Iterations_9 = Simulation_Data_9['Total_ADMM_Iterations']
max_number_of_iterations_9 = Simulation_Data_9['max_number_of_iterations']
PARAMETERS_DICT_9 = Simulation_Data_9['PARAMETERS_DICT']

SIMULATION_HORIZON_9 = PARAMETERS_DICT_9['SIMULATION_HORIZON'] 
PREDICTION_HORIZON_9 = PARAMETERS_DICT_9['PREDICTION_HORIZON']
NUMBER_ATOMIC_AGENTS_9 = PARAMETERS_DICT_9['NUMBER_ATOMIC_AGENTS']
NUMBER_CONTROL_AGENTS_9 = Partitioning_Dict_9["NUMBER_CONTROL_AGENTS"]
MAX_u_9 = PARAMETERS_DICT_9['MAX_u']
MIN_u_9 = PARAMETERS_DICT_9['MIN_u']
MAX_x_9 = PARAMETERS_DICT_9['MAX_x']
MIN_x_9 = PARAMETERS_DICT_9['MIN_x']



# %% MARK: Evaluation metrics
# print("Number of control agents: ", NUMBER_CONTROL_AGENTS_1)
# print("Total cost: ", Cost_Function_1[SIMULATION_HORIZON_1-1])
# print("Total computation time: ", Total_Computation_Time_1)
# print("Total core seconds: ", Total_Core_Seconds_1)
# print("Total ADMM Iterations: ", Total_Number_ADMM_Iterations_1)

# %% MARK: Plot Comparison Cost versus time
# Total time comparison
Total_Computation_Time_Vector_OPT = np.array([Total_Computation_Time_1, Total_Computation_Time_2, Total_Computation_Time_3, Total_Computation_Time_4, Total_Computation_Time_5])
Cost_Function_Vector_OPT = np.array([Cost_Function_1[0][-2], Cost_Function_2[0][-2], Cost_Function_3[0][-2], Cost_Function_4[0][-2], Cost_Function_5[0][-2]])
Total_Computation_Time_Vector_ALG = np.array([Total_Computation_Time_1, Total_Computation_Time_6, Total_Computation_Time_7, Total_Computation_Time_8, Total_Computation_Time_9, Total_Computation_Time_5])
Cost_Function_Vector_ALG = np.array([Cost_Function_1[0][-2], Cost_Function_6[0][-2], Cost_Function_7[0][-2], Cost_Function_8[0][-2], Cost_Function_9[0][-2], Cost_Function_5[0][-2]])

Fig_Sec = plt.figure(figsize=(9, 3.1))
ax_fig = plt.gca()
plt.title(f'Cost function - computation time comparison', fontsize=24)
plt.xlabel(f'Seconds [s]', fontsize=18)
plt.ylabel(f'Value cost function', fontsize=18)
# plt.tight_layout()
# plt.xlim(0,SIMULATION_HORIZON_1-1)
# plt.ylim(1.1*min(Cost_Function_Vector),1.1*max(Cost_Function_Vector))
# plt.yscale("log")

plt.plot(np.flip(Total_Computation_Time_Vector_ALG), np.flip(Cost_Function_Vector_ALG), marker='s', markersize=10,  
         markerfacecolor='green',         
         markeredgecolor='blue', 
         markeredgewidth=2.5,
         linestyle='--', color='lightgrey', linewidth=2)

plt.plot(np.flip(Total_Computation_Time_Vector_OPT), np.flip(Cost_Function_Vector_OPT), marker='s', markersize=10,  
         markerfacecolor='red',         
         markeredgecolor='blue', 
         markeredgewidth=2.5,
         linestyle='--', color='lightgrey', linewidth=2)



# ax_fig.set_xticks(Cores_Vector )  # Define positions of y-axis ticks
# ax_fig.set_xticklabels(["1\n" f'$\mathcal{{P}}_4$', "4\n"f'$\mathcal{{P}}_3$', "16\n"f'$\mathcal{{P}}_2$', "64\n"f'$\mathcal{{P}}_1$'], fontsize=18)  # Define custom labels

# %% MARK: Plot Comparison Cost versus core seconds
Total_Core_Seconds_Vector_OPT = np.array([Total_Core_Seconds_1, Total_Core_Seconds_2, Total_Core_Seconds_3, Total_Core_Seconds_4, Total_Core_Seconds_5])
Cost_Function_Vector_OPT = np.array([Cost_Function_1[0][-2], Cost_Function_2[0][-2], Cost_Function_3[0][-2], Cost_Function_4[0][-2], Cost_Function_5[0][-2]])
Total_Core_Seconds_Vector_ALG = np.array([Total_Core_Seconds_1, Total_Core_Seconds_6, Total_Core_Seconds_7, Total_Core_Seconds_8, Total_Core_Seconds_9, Total_Core_Seconds_5])
Cost_Function_Vector_ALG = np.array([Cost_Function_1[0][-2], Cost_Function_6[0][-2], Cost_Function_7[0][-2], Cost_Function_8[0][-2], Cost_Function_9[0][-2], Cost_Function_5[0][-2]])

Fig_Sec = plt.figure(figsize=(9, 3.1))
ax_fig = plt.gca()
plt.title(f'Cost function - computation time comparison', fontsize=24)
plt.xlabel(f'Core seconds [s]', fontsize=18)
plt.ylabel(f'Value cost function', fontsize=18)
# plt.tight_layout()
# plt.xlim(0,SIMULATION_HORIZON_1-1)
# plt.ylim(1.1*min(Cost_Function_Vector),1.1*max(Cost_Function_Vector))
# plt.yscale("log")

plt.scatter(np.flip(Total_Core_Seconds_Vector_ALG), np.flip(Cost_Function_Vector_ALG), s=100, c="red", edgecolor='blue', linewidths=2)

plt.scatter(np.flip(Total_Core_Seconds_Vector_OPT), np.flip(Cost_Function_Vector_OPT), s=100, c="green", edgecolor='blue', linewidths=2)

# %% MARK: Plot Comparison time versus cores 

Total_Cores_Vector_OPT = np.array([NUMBER_CONTROL_AGENTS_1, NUMBER_CONTROL_AGENTS_2, NUMBER_CONTROL_AGENTS_3, NUMBER_CONTROL_AGENTS_4, NUMBER_CONTROL_AGENTS_5])
Total_Cores_Vector_ALG = np.array([NUMBER_CONTROL_AGENTS_1, NUMBER_CONTROL_AGENTS_6, NUMBER_CONTROL_AGENTS_7, NUMBER_CONTROL_AGENTS_8, NUMBER_CONTROL_AGENTS_9, NUMBER_CONTROL_AGENTS_5])


Fig_Sec = plt.figure(figsize=(9, 3.1))
ax_fig = plt.gca()
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.title(f'Computation time comparison', fontsize=24)
plt.xlabel(f'Number of cores', fontsize=18)
plt.ylabel(f'Seconds [s]', fontsize=18)
# plt.tight_layout()
# plt.xlim(0,SIMULATION_HORIZON_1-1)
# plt.ylim(1.1*min(Cost_Function_Vector),1.1*max(Cost_Function_Vector))
# plt.yscale("log")

# plt.plot(np.flip(Total_Cores_Vector_ALG), np.flip(Total_Computation_Time_Vector_ALG), marker='s', markersize=10,  
#          markerfacecolor='green',         
#          markeredgecolor='blue', 
#          markeredgewidth=2.5,
#          linestyle='--', color='lightgrey', linewidth=2)

plt.scatter(np.flip(Total_Cores_Vector_ALG), np.flip(Total_Computation_Time_Vector_ALG), s=100, c="green", edgecolor='blue', linewidths=2, zorder = 3)

plt.plot(np.flip(Total_Cores_Vector_OPT), np.flip(Total_Computation_Time_Vector_OPT), marker='s', markersize=15,  
         markerfacecolor='red',         
         markeredgecolor='blue', 
         markeredgewidth=2.5,
         linestyle='--', color='lightgrey', linewidth=2, zorder = 2)





# ax_fig.set_xticks(Cores_Vector )  # Define positions of y-axis ticks
# ax_fig.set_xticklabels(["1\n" f'$\mathcal{{P}}_4$', "4\n"f'$\mathcal{{P}}_3$', "16\n"f'$\mathcal{{P}}_2$', "64\n"f'$\mathcal{{P}}_1$'], fontsize=18)  # Define custom labels
ax_fig.tick_params(axis='both', which='major', labelsize=18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
Fig_Sec.savefig(SIMULATION_FOLDER + Simumation_ID + "_Computation_time_comparison.pdf", dpi=300, bbox_inches=None)
Fig_Sec.savefig(SIMULATION_FOLDER + Simumation_ID + "_Computation_time_comparison.jpeg", dpi=300, bbox_inches=None)

# %% MARK: Plot Comparison core seconds versus cores 

Total_Cores_Vector_OPT = np.array([NUMBER_CONTROL_AGENTS_1, NUMBER_CONTROL_AGENTS_2, NUMBER_CONTROL_AGENTS_3, NUMBER_CONTROL_AGENTS_4, NUMBER_CONTROL_AGENTS_5])
Total_Cores_Vector_ALG = np.array([NUMBER_CONTROL_AGENTS_1, NUMBER_CONTROL_AGENTS_6, NUMBER_CONTROL_AGENTS_7, NUMBER_CONTROL_AGENTS_8, NUMBER_CONTROL_AGENTS_9, NUMBER_CONTROL_AGENTS_5])


Fig_Sec = plt.figure(figsize=(9, 3.1))
ax_fig = plt.gca()
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.title(f'Computation cost comparison', fontsize=24)
plt.xlabel(f'Number of cores', fontsize=18)
plt.ylabel(f'Core seconds [s]', fontsize=18)
# plt.tight_layout()
# plt.xlim(0,SIMULATION_HORIZON_1-1)
# plt.ylim(1.1*min(Cost_Function_Vector),1.1*max(Cost_Function_Vector))
# plt.yscale("log")

# plt.plot(np.flip(Total_Cores_Vector_ALG), np.flip(Total_Computation_Time_Vector_ALG), marker='s', markersize=10,  
#          markerfacecolor='green',         
#          markeredgecolor='blue', 
#          markeredgewidth=2.5,
#          linestyle='--', color='lightgrey', linewidth=2)

plt.scatter(np.flip(Total_Cores_Vector_ALG), np.flip(Total_Core_Seconds_Vector_ALG), s=100, c="green", edgecolor='blue', linewidths=2, zorder = 3)

plt.plot(np.flip(Total_Cores_Vector_OPT), np.flip(Total_Core_Seconds_Vector_OPT), marker='s', markersize=15,  
         markerfacecolor='red',         
         markeredgecolor='blue', 
         markeredgewidth=2.5,
         linestyle='--', color='lightgrey', linewidth=2, zorder = 2)





# ax_fig.set_xticks(Cores_Vector )  # Define positions of y-axis ticks
# ax_fig.set_xticklabels(fontsize=18)  # Define custom labels

ax_fig.tick_params(axis='both', which='major', labelsize=18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)

Fig_Sec.savefig(SIMULATION_FOLDER + Simumation_ID + "_Computation_cost_comparison.pdf", dpi=300, bbox_inches=None)
Fig_Sec.savefig(SIMULATION_FOLDER + Simumation_ID + "_Computation_cost_comparison.jpeg", dpi=300, bbox_inches=None)

# %% MARK: Plot Comparison time versus number of cores
# Total time comparison

Fig_Sec = plt.figure(figsize=(9, 3.1))
ax_fig = plt.gca()
plt.title(f'Cost function - computation time comparison', fontsize=24)
plt.xlabel(f'Core seconds [s]', fontsize=18)
plt.ylabel(f'Seconds [s]', fontsize=18)
# plt.tight_layout()
# plt.xlim(0,SIMULATION_HORIZON_1-1)
# plt.ylim(1.1*min(Cost_Function_Vector),1.1*max(Cost_Function_Vector))
# plt.yscale("log")

plt.plot(np.flip(Total_Core_Seconds_Vector_ALG), np.flip(Total_Computation_Time_Vector_ALG), marker='s', markersize=10,  
         markerfacecolor='green',         
         markeredgecolor='blue', 
         markeredgewidth=2.5,
         linestyle='--', color='lightgrey', linewidth=2)

plt.plot(np.flip(Total_Core_Seconds_Vector_OPT), np.flip(Total_Computation_Time_Vector_OPT), marker='s', markersize=10,  
         markerfacecolor='red',         
         markeredgecolor='blue', 
         markeredgewidth=2.5,
         linestyle='--', color='lightgrey', linewidth=2)



# ax_fig.set_xticks(Cores_Vector )  # Define positions of y-axis ticks
# ax_fig.set_xticklabels(["1\n" f'$\mathcal{{P}}_4$', "4\n"f'$\mathcal{{P}}_3$', "16\n"f'$\mathcal{{P}}_2$', "64\n"f'$\mathcal{{P}}_1$'], fontsize=18)  # Define custom labels

# %% MARK: Plot Error comparison

Error_Evolution_1 = np.zeros((NUMBER_ATOMIC_AGENTS_1,SIMULATION_HORIZON_1))
Max_Error_Evolution_1 = np.zeros((1,SIMULATION_HORIZON_1))

for idx in range(0,SIMULATION_HORIZON_1):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_1):
        Error_Evolution_1[agent_idx,idx] = Reference_Signal_1[agent_idx,idx] - x_evolution_1[agent_idx,idx]
    Max_Error_Evolution_1[0,idx] = np.max(Error_Evolution_1[:,idx])

Error_Evolution_2 = np.zeros((NUMBER_ATOMIC_AGENTS_2,SIMULATION_HORIZON_2))
Max_Error_Evolution_2 = np.zeros((1,SIMULATION_HORIZON_2))

for idx in range(0,SIMULATION_HORIZON_2):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_2):
        Error_Evolution_2[agent_idx,idx] = Reference_Signal_2[agent_idx,idx] - x_evolution_2[agent_idx,idx]
    Max_Error_Evolution_2[0,idx] = np.max(Error_Evolution_2[:,idx])

Error_Evolution_3 = np.zeros((NUMBER_ATOMIC_AGENTS_3,SIMULATION_HORIZON_3))
Max_Error_Evolution_3 = np.zeros((1,SIMULATION_HORIZON_3))

for idx in range(0,SIMULATION_HORIZON_3):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_3):
        Error_Evolution_3[agent_idx,idx] = Reference_Signal_3[agent_idx,idx] - x_evolution_3[agent_idx,idx]
    Max_Error_Evolution_3[0,idx] = np.max(Error_Evolution_3[:,idx])

Error_Evolution_4 = np.zeros((NUMBER_ATOMIC_AGENTS_4,SIMULATION_HORIZON_4))
Max_Error_Evolution_4 = np.zeros((1,SIMULATION_HORIZON_4))

for idx in range(0,SIMULATION_HORIZON_4):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_4):
        Error_Evolution_4[agent_idx,idx] = Reference_Signal_4[agent_idx,idx] - x_evolution_4[agent_idx,idx]
    Max_Error_Evolution_4[0,idx] = np.max(Error_Evolution_4[:,idx])

Error_Evolution_5 = np.zeros((NUMBER_ATOMIC_AGENTS_5,SIMULATION_HORIZON_5))
Max_Error_Evolution_5 = np.zeros((1,SIMULATION_HORIZON_5))

for idx in range(0,SIMULATION_HORIZON_5):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_5):
        Error_Evolution_5[agent_idx,idx] = Reference_Signal_5[agent_idx,idx] - x_evolution_5[agent_idx,idx]
    Max_Error_Evolution_5[0,idx] = np.max(Error_Evolution_5[:,idx])

Error_Evolution_6 = np.zeros((NUMBER_ATOMIC_AGENTS_6,SIMULATION_HORIZON_6))
Max_Error_Evolution_6 = np.zeros((1,SIMULATION_HORIZON_6))

for idx in range(0,SIMULATION_HORIZON_6):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_6):
        Error_Evolution_6[agent_idx,idx] = Reference_Signal_6[agent_idx,idx] - x_evolution_6[agent_idx,idx]
    Max_Error_Evolution_6[0,idx] = np.max(Error_Evolution_6[:,idx])

Error_Evolution_7 = np.zeros((NUMBER_ATOMIC_AGENTS_7,SIMULATION_HORIZON_7))
Max_Error_Evolution_7 = np.zeros((1,SIMULATION_HORIZON_7))

for idx in range(0,SIMULATION_HORIZON_7):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_7):
        Error_Evolution_7[agent_idx,idx] = Reference_Signal_7[agent_idx,idx] - x_evolution_7[agent_idx,idx]
    Max_Error_Evolution_7[0,idx] = np.max(Error_Evolution_7[:,idx])

Error_Evolution_8 = np.zeros((NUMBER_ATOMIC_AGENTS_8,SIMULATION_HORIZON_8))
Max_Error_Evolution_8 = np.zeros((1,SIMULATION_HORIZON_8))

for idx in range(0,SIMULATION_HORIZON_8):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_8):
        Error_Evolution_8[agent_idx,idx] = Reference_Signal_8[agent_idx,idx] - x_evolution_8[agent_idx,idx]
    Max_Error_Evolution_8[0,idx] = np.max(Error_Evolution_8[:,idx])

Error_Evolution_9 = np.zeros((NUMBER_ATOMIC_AGENTS_9,SIMULATION_HORIZON_9))
Max_Error_Evolution_9 = np.zeros((1,SIMULATION_HORIZON_9))

for idx in range(0,SIMULATION_HORIZON_9):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_9):
        Error_Evolution_9[agent_idx,idx] = Reference_Signal_9[agent_idx,idx] - x_evolution_9[agent_idx,idx]
    Max_Error_Evolution_9[0,idx] = np.max(Error_Evolution_9[:,idx])


# %%
# # Cost linear
Fig_Cost = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'Maximum error evolution', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'$e(k)$', fontsize=18)
# plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON_4-1)
#  For Linear Systems
# plt.ylim(0,1.1*max(Cost_Function_4))
# plt.plot(Cost_Function_4.T,drawstyle='steps-post', linewidth=2)
#  For hybrid systems
# plt.ylim(0,1.1*max(Max_Error_Evolution_1[0,:]))
plt.plot(Max_Error_Evolution_1.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{ADMM}}$')
plt.plot(Max_Error_Evolution_2.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Opt}}_{{1}}$')
plt.plot(Max_Error_Evolution_3.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Opt}}_{{2}}$')
plt.plot(Max_Error_Evolution_4.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Opt}}_{{3}}$')
plt.plot(Max_Error_Evolution_5.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{CMPC}}$')
plt.plot(Max_Error_Evolution_6.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Alg}}_{{1}}$')
plt.plot(Max_Error_Evolution_7.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Alg}}_{{2}}$')
plt.plot(Max_Error_Evolution_8.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Alg}}_{{3}}$')
plt.plot(Max_Error_Evolution_9.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Alg}}_{{4}}$')
plt.legend(loc='upper right', fontsize = 10, ncol=3)
plt.grid(color='gray', linestyle='--', linewidth=0.5)
# plt.axhline(y = MAX_x_1, color='red', linestyle='--', linewidth=2) 
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()
Fig_Cost.savefig(SIMULATION_FOLDER + Simumation_ID + "_Error.pdf", dpi=300, bbox_inches=None)
Fig_Cost.savefig(SIMULATION_FOLDER + Simumation_ID + "_Error.jpeg", dpi=300, bbox_inches=None)
# %% MARK: Maximum bound distance state

Bound_Evolution_1 = np.zeros((NUMBER_ATOMIC_AGENTS_1,SIMULATION_HORIZON_1))
Min_Bound_Evolution_1 = np.zeros((1,SIMULATION_HORIZON_1))

for idx in range(0,SIMULATION_HORIZON_1):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_1):
        Bound_Evolution_1[agent_idx,idx] = np.abs(x_evolution_1[agent_idx,idx])
    Min_Bound_Evolution_1[0,idx] = np.max(Bound_Evolution_1[:,idx])

Bound_Evolution_2 = np.zeros((NUMBER_ATOMIC_AGENTS_2,SIMULATION_HORIZON_2))
Min_Bound_Evolution_2 = np.zeros((1,SIMULATION_HORIZON_2))

for idx in range(0,SIMULATION_HORIZON_2):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_2):
        Bound_Evolution_2[agent_idx,idx] = np.abs(x_evolution_2[agent_idx,idx])
    Min_Bound_Evolution_2[0,idx] = np.max(Bound_Evolution_2[:,idx])

Bound_Evolution_3 = np.zeros((NUMBER_ATOMIC_AGENTS_3,SIMULATION_HORIZON_3))
Min_Bound_Evolution_3 = np.zeros((1,SIMULATION_HORIZON_3))

for idx in range(0,SIMULATION_HORIZON_3):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_3):
        Bound_Evolution_3[agent_idx,idx] = np.abs(x_evolution_3[agent_idx,idx])
    Min_Bound_Evolution_3[0,idx] = np.max(Bound_Evolution_3[:,idx])

Bound_Evolution_4 = np.zeros((NUMBER_ATOMIC_AGENTS_4,SIMULATION_HORIZON_4))
Min_Bound_Evolution_4 = np.zeros((1,SIMULATION_HORIZON_4))

for idx in range(0,SIMULATION_HORIZON_4):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_4):
        Bound_Evolution_4[agent_idx,idx] = np.abs(x_evolution_4[agent_idx,idx])
    Min_Bound_Evolution_4[0,idx] = np.max(Bound_Evolution_4[:,idx])

Bound_Evolution_5 = np.zeros((NUMBER_ATOMIC_AGENTS_5,SIMULATION_HORIZON_5))
Min_Bound_Evolution_5 = np.zeros((1,SIMULATION_HORIZON_5))

for idx in range(0,SIMULATION_HORIZON_5):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_5):
        Bound_Evolution_5[agent_idx,idx] = np.abs(x_evolution_5[agent_idx,idx])
    Min_Bound_Evolution_5[0,idx] = np.max(Bound_Evolution_5[:,idx])

Bound_Evolution_6 = np.zeros((NUMBER_ATOMIC_AGENTS_6,SIMULATION_HORIZON_6))
Min_Bound_Evolution_6 = np.zeros((1,SIMULATION_HORIZON_6))

for idx in range(0,SIMULATION_HORIZON_6):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_6):
        Bound_Evolution_6[agent_idx,idx] = np.abs(x_evolution_6[agent_idx,idx])
    Min_Bound_Evolution_6[0,idx] = np.max(Bound_Evolution_6[:,idx])

Bound_Evolution_7 = np.zeros((NUMBER_ATOMIC_AGENTS_7,SIMULATION_HORIZON_7))
Min_Bound_Evolution_7 = np.zeros((1,SIMULATION_HORIZON_7))

for idx in range(0,SIMULATION_HORIZON_7):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_7):
        Bound_Evolution_7[agent_idx,idx] = np.abs(x_evolution_7[agent_idx,idx])
    Min_Bound_Evolution_7[0,idx] = np.max(Bound_Evolution_7[:,idx])

Bound_Evolution_8 = np.zeros((NUMBER_ATOMIC_AGENTS_8,SIMULATION_HORIZON_8))
Min_Bound_Evolution_8 = np.zeros((1,SIMULATION_HORIZON_8))

for idx in range(0,SIMULATION_HORIZON_8):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_8):
        Bound_Evolution_8[agent_idx,idx] = np.abs(x_evolution_8[agent_idx,idx])
    Min_Bound_Evolution_8[0,idx] = np.max(Bound_Evolution_8[:,idx])

Bound_Evolution_9 = np.zeros((NUMBER_ATOMIC_AGENTS_9,SIMULATION_HORIZON_9))
Min_Bound_Evolution_9 = np.zeros((1,SIMULATION_HORIZON_9))

for idx in range(0,SIMULATION_HORIZON_9):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_9):
        Bound_Evolution_9[agent_idx,idx] = np.abs(x_evolution_9[agent_idx,idx])
    Min_Bound_Evolution_9[0,idx] = np.max(Bound_Evolution_9[:,idx])

# %%
# # Cost linear
Fig_Cost = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'Maximum absolute value for the state', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'Max $|x(k)|$', fontsize=18)
# plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON_4-1)
#  For Linear Systems
# plt.ylim(0,1.1*max(Cost_Function_4))
# plt.plot(Cost_Function_4.T,drawstyle='steps-post', linewidth=2)
#  For hybrid systems
# plt.ylim(0,1.1*max(Max_Error_Evolution_1[0,:]))
plt.plot(Min_Bound_Evolution_1.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{ADMM}}$')
plt.plot(Min_Bound_Evolution_2.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Opt}}_{{1}}$')
plt.plot(Min_Bound_Evolution_3.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Opt}}_{{2}}$')
plt.plot(Min_Bound_Evolution_4.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Opt}}_{{3}}$')
plt.plot(Min_Bound_Evolution_5.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{CMPC}}$')
plt.plot(Min_Bound_Evolution_6.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Alg}}_{{1}}$')
plt.plot(Min_Bound_Evolution_7.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Alg}}_{{2}}$')
plt.plot(Min_Bound_Evolution_8.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Alg}}_{{3}}$')
plt.plot(Min_Bound_Evolution_9.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Alg}}_{{4}}$')
plt.legend(loc='lower right', fontsize = 10, ncol=3)
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.axhline(y = MAX_x_1, color='red', linestyle='--', linewidth=2) 
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()
Fig_Cost.savefig(SIMULATION_FOLDER + Simumation_ID + "_MaxState.pdf", dpi=300, bbox_inches=None)
Fig_Cost.savefig(SIMULATION_FOLDER + Simumation_ID + "_MaxState.jpeg", dpi=300, bbox_inches=None)


# %% MARK: Maximum bound distance input

Bound_Evolution_u_1 = np.zeros((NUMBER_ATOMIC_AGENTS_1,SIMULATION_HORIZON_1))
Min_Bound_Evolution_u_1 = np.zeros((1,SIMULATION_HORIZON_1))

for idx in range(0,SIMULATION_HORIZON_1):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_1):
        Bound_Evolution_u_1[agent_idx,idx] = np.abs(u_evolution_1[agent_idx,idx])
    Min_Bound_Evolution_u_1[0,idx] = np.max(Bound_Evolution_u_1[:,idx])

Bound_Evolution_u_2 = np.zeros((NUMBER_ATOMIC_AGENTS_2,SIMULATION_HORIZON_2))
Min_Bound_Evolution_u_2 = np.zeros((1,SIMULATION_HORIZON_2))

for idx in range(0,SIMULATION_HORIZON_2):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_2):
        Bound_Evolution_u_2[agent_idx,idx] = np.abs(u_evolution_2[agent_idx,idx])
    Min_Bound_Evolution_u_2[0,idx] = np.max(Bound_Evolution_u_2[:,idx])

Bound_Evolution_u_3 = np.zeros((NUMBER_ATOMIC_AGENTS_3,SIMULATION_HORIZON_3))
Min_Bound_Evolution_u_3 = np.zeros((1,SIMULATION_HORIZON_3))

for idx in range(0,SIMULATION_HORIZON_3):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_3):
        Bound_Evolution_u_3[agent_idx,idx] = np.abs(u_evolution_3[agent_idx,idx])
    Min_Bound_Evolution_u_3[0,idx] = np.max(Bound_Evolution_u_3[:,idx])

Bound_Evolution_u_4 = np.zeros((NUMBER_ATOMIC_AGENTS_4,SIMULATION_HORIZON_4))
Min_Bound_Evolution_u_4 = np.zeros((1,SIMULATION_HORIZON_4))

for idx in range(0,SIMULATION_HORIZON_4):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_4):
        Bound_Evolution_u_4[agent_idx,idx] = np.abs(u_evolution_4[agent_idx,idx])
    Min_Bound_Evolution_u_4[0,idx] = np.max(Bound_Evolution_u_4[:,idx])

Bound_Evolution_u_5 = np.zeros((NUMBER_ATOMIC_AGENTS_5,SIMULATION_HORIZON_5))
Min_Bound_Evolution_u_5 = np.zeros((1,SIMULATION_HORIZON_5))

for idx in range(0,SIMULATION_HORIZON_5):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_5):
        Bound_Evolution_u_5[agent_idx,idx] = np.abs(u_evolution_5[agent_idx,idx])
    Min_Bound_Evolution_u_5[0,idx] = np.max(Bound_Evolution_u_5[:,idx])

Bound_Evolution_u_6 = np.zeros((NUMBER_ATOMIC_AGENTS_6,SIMULATION_HORIZON_6))
Min_Bound_Evolution_u_6 = np.zeros((1,SIMULATION_HORIZON_6))

for idx in range(0,SIMULATION_HORIZON_6):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_6):
        Bound_Evolution_u_6[agent_idx,idx] = np.abs(u_evolution_6[agent_idx,idx])
    Min_Bound_Evolution_u_6[0,idx] = np.max(Bound_Evolution_u_6[:,idx])

Bound_Evolution_u_7 = np.zeros((NUMBER_ATOMIC_AGENTS_7,SIMULATION_HORIZON_7))
Min_Bound_Evolution_u_7 = np.zeros((1,SIMULATION_HORIZON_7))

for idx in range(0,SIMULATION_HORIZON_7):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_7):
        Bound_Evolution_u_7[agent_idx,idx] = np.abs(u_evolution_7[agent_idx,idx])
    Min_Bound_Evolution_u_7[0,idx] = np.max(Bound_Evolution_u_7[:,idx])

Bound_Evolution_u_8 = np.zeros((NUMBER_ATOMIC_AGENTS_8,SIMULATION_HORIZON_8))
Min_Bound_Evolution_u_8 = np.zeros((1,SIMULATION_HORIZON_8))

for idx in range(0,SIMULATION_HORIZON_8):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_8):
        Bound_Evolution_u_8[agent_idx,idx] = np.abs(u_evolution_8[agent_idx,idx])
    Min_Bound_Evolution_u_8[0,idx] = np.max(Bound_Evolution_u_8[:,idx])

Bound_Evolution_u_9 = np.zeros((NUMBER_ATOMIC_AGENTS_9,SIMULATION_HORIZON_9))
Min_Bound_Evolution_u_9 = np.zeros((1,SIMULATION_HORIZON_9))

for idx in range(0,SIMULATION_HORIZON_9):
    for agent_idx in range(0,NUMBER_ATOMIC_AGENTS_9):
        Bound_Evolution_u_9[agent_idx,idx] = np.abs(u_evolution_9[agent_idx,idx])
    Min_Bound_Evolution_u_9[0,idx] = np.max(Bound_Evolution_u_9[:,idx])

# %%
# # Cost linear
Fig_Cost = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'Maximum absolute value for the input', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'Max $|u(k)|$', fontsize=18)
# plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON_4-1)
#  For Linear Systems
# plt.ylim(0,1.1*max(Cost_Function_4))
# plt.plot(Cost_Function_4.T,drawstyle='steps-post', linewidth=2)
#  For hybrid systems
# plt.ylim(0,1.1*max(Max_Error_Evolution_u_1[0,:]))
plt.plot(Min_Bound_Evolution_u_1.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{ADMM}}$')
plt.plot(Min_Bound_Evolution_u_2.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Opt}}_{{1}}$')
plt.plot(Min_Bound_Evolution_u_3.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Opt}}_{{2}}$')
plt.plot(Min_Bound_Evolution_u_4.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Opt}}_{{3}}$')
plt.plot(Min_Bound_Evolution_u_5.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{CMPC}}$')
plt.plot(Min_Bound_Evolution_u_6.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Alg}}_{{1}}$')
plt.plot(Min_Bound_Evolution_u_7.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Alg}}_{{2}}$')
plt.plot(Min_Bound_Evolution_u_8.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Alg}}_{{3}}$')
plt.plot(Min_Bound_Evolution_u_9.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Alg}}_{{4}}$')
plt.legend(loc='upper left', fontsize = 10, ncol=3)
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.axhline(y = MAX_u_1, color='red', linestyle='--', linewidth=2) 
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

Fig_Cost.savefig(SIMULATION_FOLDER + Simumation_ID + "_MaxInput.pdf", dpi=300, bbox_inches=None)
Fig_Cost.savefig(SIMULATION_FOLDER + Simumation_ID + "_MaxInput.jpeg", dpi=300, bbox_inches=None)

# %% Export the legend
Fig_Cost = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'Maximum absolute value for the input', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'Max $|u(k)|$', fontsize=18)
# plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON_4-1)
#  For Linear Systems
# plt.ylim(0,1.1*max(Cost_Function_4))
# plt.plot(Cost_Function_4.T,drawstyle='steps-post', linewidth=2)
#  For hybrid systems
# plt.ylim(0,1.1*max(Max_Error_Evolution_u_1[0,:]))
plt.plot(Min_Bound_Evolution_u_1.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{ADMM}}$')
plt.plot(Min_Bound_Evolution_u_2.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Opt}}_{{1}}$')
plt.plot(Min_Bound_Evolution_u_3.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Opt}}_{{2}}$')
plt.plot(Min_Bound_Evolution_u_4.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Opt}}_{{3}}$')
plt.plot(Min_Bound_Evolution_u_5.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{CMPC}}$')
plt.plot(Min_Bound_Evolution_u_6.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Alg}}_{{1}}$')
plt.plot(Min_Bound_Evolution_u_7.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Alg}}_{{2}}$')
plt.plot(Min_Bound_Evolution_u_8.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Alg}}_{{3}}$')
plt.plot(Min_Bound_Evolution_u_9.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}^{{Alg}}_{{4}}$')
plt.legend(fontsize = 18, ncol=3)
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.axhline(y = MAX_u_1, color='red', linestyle='--', linewidth=2) 
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
# plt.show()
# Fig_Cost.legend(loc='upper right', fontsize = 18, ncol=3)
# ax_fig.set_visible(False)
ax_fig.axis("off")
# Fig_Cost.show()
plt.show()
# %%
Fig_Cost.savefig(SIMULATION_FOLDER + Simumation_ID + "_Legend.pdf", dpi=300, bbox_inches=None)
Fig_Cost.savefig(SIMULATION_FOLDER + Simumation_ID + "_Legend.jpeg", dpi=300, bbox_inches=None)

# %% MARK: RESULTS COMPARISON

fig_all, axes_all = plt.subplots(1, 3, figsize=(18, 3))

plt.title(f'Bound limit proximit for the input', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'Bound limit', fontsize=18)
# plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON_4-1)

axes_all[0].set_title(f'Maximum error evolution', fontsize=24)
axes_all[0].set_xlabel(f'Time step $k$', fontsize=18)
axes_all[0].set_ylabel(f'Bound limit', fontsize=18)

axes_all[1].set_title(f'Bound limit proximit for the state', fontsize=24)
axes_all[1].set_xlabel(f'Time step $k$', fontsize=18)
axes_all[1].set_ylabel(f'Bound limit', fontsize=18)

axes_all[2].set_title(f'Bound limit proximit for the input', fontsize=24)
axes_all[2].set_xlabel(f'Time step $k$', fontsize=18)
axes_all[2].set_ylabel(f'Bound limit', fontsize=18)




# %% MARK: Plot Comparison

# Total Cost comparison linear
# Cost_Function_Vector = np.array([Cost_Function_4[-2], Cost_Function_3[-2], Cost_Function_2[-2], Cost_Function_1[-2]])
Cores_Vector = np.array([1, 4, 16, 64])

# PWA
Cost_Function_Vector = np.array([Cost_Function_4[0][-2], Cost_Function_3[0][-2], Cost_Function_2[0][-2], Cost_Function_1[0][-2]])

Fig_Com = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'Cost functions comparison', fontsize=24)
plt.xlabel(f'Partiton', fontsize=18)
plt.ylabel(f'$J$', fontsize=18)
# plt.tight_layout()
# plt.xlim(0,SIMULATION_HORIZON_1-1)
# plt.ylim(1.1*min(Cost_Function_Vector),1.1*max(Cost_Function_Vector))
plt.yscale("log")
plt.plot(Cost_Function_Vector, marker='s', markersize=10,  
         markerfacecolor='red',         
         markeredgecolor='blue', 
         markeredgewidth=2.5,
         linestyle='--', color='lightgrey', linewidth=2)
ax_fig.set_xticks([0, 1, 2, 3])  # Define positions of y-axis ticks
ax_fig.set_xticklabels([f'$\mathcal{{P}}_4$', f'$\mathcal{{P}}_3$', f'$\mathcal{{P}}_2$', f'$\mathcal{{P}}_1$'], fontsize=18)  # Define custom labels
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.grid(color='gray', linestyle='--', linewidth=0.5)
# plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

# Total time comparison
Total_Computation_Time_Vector = np.array([Total_Computation_Time_4, Total_Computation_Time_3, Total_Computation_Time_2, Total_Computation_Time_1])

Fig_Sec = plt.figure(figsize=(9, 3.1))
ax_fig = plt.gca()
plt.title(f'Computation time comparison', fontsize=24)
plt.xlabel(f'Number of cores', fontsize=18)
plt.ylabel(f'Seconds [s]', fontsize=18)
# plt.tight_layout()
# plt.xlim(0,SIMULATION_HORIZON_1-1)
# plt.ylim(1.1*min(Cost_Function_Vector),1.1*max(Cost_Function_Vector))
# plt.yscale("log")
plt.plot(Cores_Vector, Total_Computation_Time_Vector, marker='s', markersize=10,  
         markerfacecolor='red',         
         markeredgecolor='blue', 
         markeredgewidth=2.5,
         linestyle='--', color='lightgrey', linewidth=2)
ax_fig.set_xticks(Cores_Vector )  # Define positions of y-axis ticks
ax_fig.set_xticklabels(["1\n" f'$\mathcal{{P}}_4$', "4\n"f'$\mathcal{{P}}_3$', "16\n"f'$\mathcal{{P}}_2$', "64\n"f'$\mathcal{{P}}_1$'], fontsize=18)  # Define custom labels
# ax_fig.text(7, 0.8, "Legend Style Annotation\n- Sine (Blue)\n- Cosine (Orange)",
#         fontsize=12, bbox=dict(facecolor='white', edgecolor='black', boxstyle='round,pad=0.5'))

# ax_fig.set_xticklabels([f'$\mathcal{{P}}_4: 1$', f'$\mathcal{{P}}_3: 4$', f'$\mathcal{{P}}_2: 16$', f'$\mathcal{{P}}_1: 64$'], fontsize=18)  # Define custom labels
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.subplots_adjust(top=0.88, bottom= 0.3, right=0.95, left = 0.15)
plt.show()

# Core seconds comparison
Total_Core_Seconds_Vector = np.array([Total_Core_Seconds_4, Total_Core_Seconds_3, Total_Core_Seconds_2, Total_Core_Seconds_1])

Fig_Cor = plt.figure(figsize=(9, 3.1))
ax_fig = plt.gca()
plt.title(f'Computation cost comparison', fontsize=24)
plt.xlabel(f'Number of cores', fontsize=18)
plt.ylabel(f'Core Seconds [s]', fontsize=18)
# plt.tight_layout()
# plt.xlim(0,SIMULATION_HORIZON_1-1)
# plt.ylim(1.1*min(Cost_Function_Vector),1.1*max(Cost_Function_Vector))
# plt.yscale("log")
plt.plot(Cores_Vector , Total_Core_Seconds_Vector, marker='s', markersize=10,  
         markerfacecolor='red',         
         markeredgecolor='blue', 
         markeredgewidth=2.5,
         linestyle='--', color='lightgrey', linewidth=2)
ax_fig.set_xticks(Cores_Vector )  # Define positions of y-axis ticks
ax_fig.set_xticklabels(["1\n" f'$\mathcal{{P}}_4$', "4\n"f'$\mathcal{{P}}_3$', "16\n"f'$\mathcal{{P}}_2$', "64\n"f'$\mathcal{{P}}_1$'], fontsize=18)  # Define custom labels
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.subplots_adjust(top=0.88, bottom= 0.3, right=0.95, left = 0.15)
plt.show()

# Residual evolution
Fig_Res = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'Residual evolution', fontsize=24)
plt.xlabel(f'ADMM iteration $i$', fontsize=18)
plt.ylabel(f'$r(i)$', fontsize=18)
# plt.tight_layout()
plt.xlim(0,25)
plt.ylim(10e-2,10e1)
plt.yscale("log")
plt.plot(residual_evolution_1.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}_1$')
plt.plot(residual_evolution_2.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}_2$')
plt.plot(residual_evolution_3.T,drawstyle='steps-post', linewidth=2, label=f'$\mathcal{{P}}_3$')
# plt.plot(residual_evolution_4.T,drawstyle='steps-post')
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.legend(loc='upper right', fontsize = 18)
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

# %% MARK: Plot CMPC simulation

# State
Fig_Ref = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'Reference evolution', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'$r(k)$', fontsize=18)
# plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON_4-1)
plt.ylim(1.1*min(Reference_Signal_4[0,:]),1.1*max(Reference_Signal_4[0,:]))
plt.plot(Reference_Signal_4.T,drawstyle='steps-post', linewidth=2)
plt.grid(color='gray', linestyle='--', linewidth=0.5)
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

# State
Fig_State = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'State evolution', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'$x(k)$', fontsize=18)
# plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON_4-1)
# plt.ylim(MIN_x_4*1.1,MAX_x_4*1.1)
plt.ylim(1.1*min(Reference_Signal_4[0,:]),1.1*max(Reference_Signal_4[0,:]))
# plt.plot(Reference_Signal_4.T,drawstyle='steps-post', color='red', label = f"r(k)")
plt.plot(x_evolution_2.T,drawstyle='steps-post', color='blue', label = f"x(k)", linewidth=2)
plt.axhline(y = MIN_x_4, color='red', linestyle='--', linewidth=2) 
plt.axhline(y = MAX_x_4, color='red', linestyle='--', linewidth=2) 
plt.grid(color='gray', linestyle='--', linewidth=0.5)
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
# plt.legend(loc='upper right', fontsize = 18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

# Input
Fig_Input = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'Input evolution', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'$u(k)$', fontsize=18)
# plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON_4-1)
plt.ylim(MIN_u_4*1.1,MAX_u_4*1.1)
plt.plot(u_evolution_2.T,drawstyle='steps-post')
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.axhline(y = MIN_u_4, color='red', linestyle='--', linewidth=2) 
plt.axhline(y = MAX_u_4, color='red', linestyle='--', linewidth=2) 
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

# # # Residual
# Fig_Residual = plt.figure(figsize=(9, 3))
# plt.title(f'Residual evolution', fontsize=24)
# plt.xlabel(f'Iteration $i$', fontsize=18)
# plt.ylabel(f'$\epsilon(i)$', fontsize=18)
# plt.tight_layout()
# plt.xlim(0,max_number_of_iterations_4)
# plt.ylim(0,1.1*max(residual_evolution))
# plt.plot(residual_evolution.T,drawstyle='steps-post')
# plt.grid(color='gray', linestyle='--', linewidth=0.5)
# plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

# # Cost linear
Fig_Cost = plt.figure(figsize=(9, 3))
ax_fig = plt.gca()
plt.title(f'Cost evolution', fontsize=24)
plt.xlabel(f'Time step $k$', fontsize=18)
plt.ylabel(f'$J(k)$', fontsize=18)
# plt.tight_layout()
plt.xlim(0,SIMULATION_HORIZON_4-1)
#  For Linear Systems
# plt.ylim(0,1.1*max(Cost_Function_4))
# plt.plot(Cost_Function_4.T,drawstyle='steps-post', linewidth=2)
#  For hybrid systems
plt.ylim(0,1.1*max(Cost_Function_4[0]))
plt.plot(Cost_Function_2[0].T,drawstyle='steps-post', linewidth=2)
plt.grid(color='gray', linestyle='--', linewidth=0.5)
ax_fig.tick_params(axis='x', labelsize=18)
ax_fig.tick_params(axis='y', labelsize=18)
plt.subplots_adjust(top=0.88, bottom= 0.2, right=0.95, left = 0.15)
plt.show()

# if (PLOT_FIGURES == True) and (EXPORT_FIGURES == True):
#     Fig_State.savefig(SIMULATION_FOLDER + Simumation_ID + "_State.pdf", dpi=300, bbox_inches=None)
#     Fig_State.savefig(SIMULATION_FOLDER + Simumation_ID + "_State.jpeg", dpi=300, bbox_inches=None)

#     Fig_Input.savefig(SIMULATION_FOLDER + Simumation_ID + "_Input.pdf", dpi=300, bbox_inches=None)
#     Fig_Input.savefig(SIMULATION_FOLDER + Simumation_ID + "_Input.jpeg", dpi=300, bbox_inches=None)

#     Fig_Cost.savefig(SIMULATION_FOLDER + Simumation_ID + "_Cost.pdf", dpi=300, bbox_inches=None)
#     Fig_Cost.savefig(SIMULATION_FOLDER + Simumation_ID + "_Cost.jpeg", dpi=300, bbox_inches=None)

#     Fig_Com.savefig(SIMULATION_FOLDER + Simumation_ID + "_Comparison.pdf", dpi=300, bbox_inches=None)
#     Fig_Com.savefig(SIMULATION_FOLDER + Simumation_ID + "_Comparison.jpeg", dpi=300, bbox_inches=None)
#     Fig_Sec.savefig(SIMULATION_FOLDER + Simumation_ID + "_ComputationTime.pdf", dpi=300, bbox_inches=None)
#     Fig_Sec.savefig(SIMULATION_FOLDER + Simumation_ID + "_ComputationTime.jpeg", dpi=300, bbox_inches=None)
#     Fig_Cor.savefig(SIMULATION_FOLDER + Simumation_ID + "_CoreSeconds.pdf", dpi=300, bbox_inches=None)
#     Fig_Cor.savefig(SIMULATION_FOLDER + Simumation_ID + "_CoreSeconds.jpeg", dpi=300, bbox_inches=None)
#     Fig_Res.savefig(SIMULATION_FOLDER + Simumation_ID + "_ResidualEvolution.pdf", dpi=300, bbox_inches=None)
#     Fig_Res.savefig(SIMULATION_FOLDER + Simumation_ID + "_ResidualEvolution.jpeg", dpi=300, bbox_inches=None)

#     # Fig_Residual.savefig(SIMULATION_FOLDER + Simumation_ID + "_Residual.pdf", dpi=300)
#     # Fig_Residual.savefig(SIMULATION_FOLDER + Simumation_ID + "_Residual.jpeg", dpi=300)

#     Fig_Ref.savefig(SIMULATION_FOLDER + Simumation_ID + "_Ref.pdf", dpi=300, bbox_inches=None)
#     Fig_Ref.savefig(SIMULATION_FOLDER + Simumation_ID + "_Ref.jpeg", dpi=300, bbox_inches=None)